/* Generated by Cython 0.29.34 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "/tmp/vendor/include/libavcodec/avcodec.h",
            "/tmp/vendor/include/libavcodec/bsf.h",
            "/tmp/vendor/include/libavdevice/avdevice.h",
            "/tmp/vendor/include/libavfilter/avfilter.h",
            "/tmp/vendor/include/libavfilter/buffersink.h",
            "/tmp/vendor/include/libavfilter/buffersrc.h",
            "/tmp/vendor/include/libavformat/avformat.h",
            "/tmp/vendor/include/libavutil/audio_fifo.h",
            "/tmp/vendor/include/libavutil/avutil.h",
            "/tmp/vendor/include/libavutil/channel_layout.h",
            "/tmp/vendor/include/libavutil/dict.h",
            "/tmp/vendor/include/libavutil/error.h",
            "/tmp/vendor/include/libavutil/frame.h",
            "/tmp/vendor/include/libavutil/imgutils.h",
            "/tmp/vendor/include/libavutil/log.h",
            "/tmp/vendor/include/libavutil/mathematics.h",
            "/tmp/vendor/include/libavutil/motion_vector.h",
            "/tmp/vendor/include/libavutil/opt.h",
            "/tmp/vendor/include/libavutil/pixdesc.h",
            "/tmp/vendor/include/libavutil/rational.h",
            "/tmp/vendor/include/libavutil/samplefmt.h",
            "/tmp/vendor/include/libswresample/swresample.h",
            "/tmp/vendor/include/libswscale/swscale.h"
        ],
        "include_dirs": [
            "/tmp/vendor/include"
        ],
        "libraries": [
            "avformat",
            "avcodec",
            "avdevice",
            "avutil",
            "avfilter",
            "swscale",
            "swresample"
        ],
        "library_dirs": [
            "/tmp/vendor/lib"
        ],
        "name": "av.codec.context",
        "sources": [
            "av/codec/context.pyx"
        ]
    },
    "module_name": "av.codec.context"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_34"
#define CYTHON_HEX_VERSION 0x001D22F0
#define CYTHON_FUTURE_DIVISION 0
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PY_NOGIL)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #ifndef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS (PY_VERSION_HEX < 0x030C00A5)
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 0
  #elif !defined(CYTHON_FAST_THREAD_STATE)
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL (PY_VERSION_HEX < 0x030A0000)
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS ((PY_VERSION_HEX >= 0x030600B1) && (PY_VERSION_HEX < 0x030C00A5))
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 0
  #elif !defined(CYTHON_USE_EXC_INFO_STACK)
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if PY_VERSION_HEX >= 0x030B00A1
    static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *kwds=NULL, *argcount=NULL, *posonlyargcount=NULL, *kwonlyargcount=NULL;
        PyObject *nlocals=NULL, *stacksize=NULL, *flags=NULL, *replace=NULL, *call_result=NULL, *empty=NULL;
        const char *fn_cstr=NULL;
        const char *name_cstr=NULL;
        PyCodeObject* co=NULL;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        if (!(kwds=PyDict_New())) goto end;
        if (!(argcount=PyLong_FromLong(a))) goto end;
        if (PyDict_SetItemString(kwds, "co_argcount", argcount) != 0) goto end;
        if (!(posonlyargcount=PyLong_FromLong(0))) goto end;
        if (PyDict_SetItemString(kwds, "co_posonlyargcount", posonlyargcount) != 0) goto end;
        if (!(kwonlyargcount=PyLong_FromLong(k))) goto end;
        if (PyDict_SetItemString(kwds, "co_kwonlyargcount", kwonlyargcount) != 0) goto end;
        if (!(nlocals=PyLong_FromLong(l))) goto end;
        if (PyDict_SetItemString(kwds, "co_nlocals", nlocals) != 0) goto end;
        if (!(stacksize=PyLong_FromLong(s))) goto end;
        if (PyDict_SetItemString(kwds, "co_stacksize", stacksize) != 0) goto end;
        if (!(flags=PyLong_FromLong(f))) goto end;
        if (PyDict_SetItemString(kwds, "co_flags", flags) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_code", code) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_consts", c) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_names", n) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_varnames", v) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_freevars", fv) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_cellvars", cell) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_linetable", lnos) != 0) goto end;
        if (!(fn_cstr=PyUnicode_AsUTF8AndSize(fn, NULL))) goto end;
        if (!(name_cstr=PyUnicode_AsUTF8AndSize(name, NULL))) goto end;
        if (!(co = PyCode_NewEmpty(fn_cstr, name_cstr, fline))) goto end;
        if (!(replace = PyObject_GetAttrString((PyObject*)co, "replace"))) goto cleanup_code_too;
        if (!(empty = PyTuple_New(0))) goto cleanup_code_too; // unfortunately __pyx_empty_tuple isn't available here
        if (!(call_result = PyObject_Call(replace, empty, kwds))) goto cleanup_code_too;
        Py_XDECREF((PyObject*)co);
        co = (PyCodeObject*)call_result;
        call_result = NULL;
        if (0) {
            cleanup_code_too:
            Py_XDECREF((PyObject*)co);
            co = NULL;
        }
        end:
        Py_XDECREF(kwds);
        Py_XDECREF(argcount);
        Py_XDECREF(posonlyargcount);
        Py_XDECREF(kwonlyargcount);
        Py_XDECREF(nlocals);
        Py_XDECREF(stacksize);
        Py_XDECREF(replace);
        Py_XDECREF(call_result);
        Py_XDECREF(empty);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return co;
    }
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #if !defined(_USE_MATH_DEFINES)
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__av__codec__context
#define __PYX_HAVE_API__av__codec__context
/* Early includes */
#include <stdint.h>
#include "libavutil/mathematics.h"
#include "libavutil/rational.h"
#include "libavutil/avutil.h"
#include "libavutil/pixdesc.h"
#include "libavutil/channel_layout.h"
#include "libavutil/audio_fifo.h"
#include "stdarg.h"
#include "libavutil/opt.h"
#include "libavutil/imgutils.h"
#include "libavutil/log.h"
#include "libavutil/dict.h"
#include "libavutil/error.h"
#include "libavutil/frame.h"
#include "libavutil/samplefmt.h"
#include "libavutil/motion_vector.h"
#include "libavcodec/avcodec.h"
#include "libavcodec/bsf.h"
#include "libavdevice/avdevice.h"
#include "libavformat/avformat.h"
#include "libswresample/swresample.h"
#include "libswscale/swscale.h"
#include "libavfilter/avfilter.h"

    #if (LIBAVFILTER_VERSION_INT >= 525156)
        // avfilter_filter_pad_count is available since version 8.3.100 of libavfilter (FFmpeg 5.0)
        #define _avfilter_get_num_pads(filter, is_output, pads) (avfilter_filter_pad_count(filter, is_output))
    #else
        // avfilter_filter_pad_count has been deprecated as of version 8.3.100 of libavfilter (FFmpeg 5.0)
        #define _avfilter_get_num_pads(filter, is_output, pads) (avfilter_pad_count(pads))
    #endif
    
#include "libavfilter/buffersink.h"
#include "libavfilter/buffersrc.h"
#include "stdio.h"
#include <errno.h>
#include <string.h>
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 1
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING "ascii"
#define __Pyx_PyObject_FromString __Pyx_PyStr_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyStr_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "av/codec/context.pyx",
  "stringsource",
  "av/codec/context.pxd",
  "av/bytesource.pxd",
  "av/codec/codec.pxd",
  "av/buffer.pxd",
  "av/container/pyio.pxd",
  "av/container/streams.pxd",
  "av/dictionary.pxd",
  "av/format.pxd",
  "av/container/core.pxd",
  "av/stream.pxd",
  "av/packet.pxd",
  "av/sidedata/sidedata.pxd",
  "av/frame.pxd",
};
/* NoFastGil.proto */
#define __Pyx_PyGILState_Ensure PyGILState_Ensure
#define __Pyx_PyGILState_Release PyGILState_Release
#define __Pyx_FastGIL_Remember()
#define __Pyx_FastGIL_Forget()
#define __Pyx_FastGilFuncInit()

/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif


/*--- Type declarations ---*/
struct __pyx_obj_2av_10bytesource_ByteSource;
struct __pyx_obj_2av_5codec_5codec_Codec;
struct __pyx_obj_2av_6buffer_Buffer;
struct __pyx_obj_2av_9container_4pyio_PyIOFile;
struct __pyx_obj_2av_9container_7streams_StreamContainer;
struct __pyx_obj_2av_10dictionary__Dictionary;
struct __pyx_obj_2av_6format_ContainerFormat;
struct __pyx_obj_2av_9container_4core_Container;
struct __pyx_obj_2av_6stream_Stream;
struct __pyx_obj_2av_6packet_Packet;
struct __pyx_obj_2av_8sidedata_8sidedata_SideData;
struct __pyx_obj_2av_8sidedata_8sidedata__SideDataContainer;
struct __pyx_obj_2av_5frame_Frame;
struct __pyx_obj_2av_5codec_7context_CodecContext;
struct __pyx_opt_args_2av_10bytesource_bytesource;

/* "av/bytesource.pxd":14
 *     cdef size_t length
 * 
 * cdef ByteSource bytesource(object, bint allow_none=*)             # <<<<<<<<<<<<<<
 */
struct __pyx_opt_args_2av_10bytesource_bytesource {
  int __pyx_n;
  int allow_none;
};
struct __pyx_opt_args_2av_5codec_5codec_5Codec__init;

/* "av/codec/codec.pxd":10
 *     cdef readonly bint is_encoder
 * 
 *     cdef _init(self, name=?)             # <<<<<<<<<<<<<<
 * 
 * 
 */
struct __pyx_opt_args_2av_5codec_5codec_5Codec__init {
  int __pyx_n;
  PyObject *name;
};
struct __pyx_t_2av_9container_4core_timeout_info;
typedef struct __pyx_t_2av_9container_4core_timeout_info __pyx_t_2av_9container_4core_timeout_info;

/* "av/container/core.pxd":11
 * 
 * # Interrupt callback information, times are in seconds.
 * ctypedef struct timeout_info:             # <<<<<<<<<<<<<<
 *     double start_time
 *     double timeout
 */
struct __pyx_t_2av_9container_4core_timeout_info {
  double start_time;
  double timeout;
};
struct __pyx_opt_args_2av_5frame_5Frame__copy_internal_attributes;

/* "av/frame.pxd":19
 *     cdef readonly int index
 * 
 *     cdef _copy_internal_attributes(self, Frame source, bint data_layout=?)             # <<<<<<<<<<<<<<
 * 
 *     cdef _init_user_attributes(self)
 */
struct __pyx_opt_args_2av_5frame_5Frame__copy_internal_attributes {
  int __pyx_n;
  int data_layout;
};
struct __pyx_opt_args_2av_4enum_define_enum;

/* "av/enum.pxd":2
 * cpdef define_enum(
 *     name, module, items,             # <<<<<<<<<<<<<<
 *     bint is_flags=*
 * )
 */
struct __pyx_opt_args_2av_4enum_define_enum {
  int __pyx_n;
  int is_flags;
};
struct __pyx_opt_args_2av_5error_stash_exception;
struct __pyx_opt_args_2av_5error_err_check;
struct __pyx_opt_args_2av_5error_make_error;

/* "av/error.pxd":2
 * 
 * cdef int stash_exception(exc_info=*)             # <<<<<<<<<<<<<<
 * 
 * cpdef int err_check(int res, filename=*) except -1
 */
struct __pyx_opt_args_2av_5error_stash_exception {
  int __pyx_n;
  PyObject *exc_info;
};

/* "av/error.pxd":4
 * cdef int stash_exception(exc_info=*)
 * 
 * cpdef int err_check(int res, filename=*) except -1             # <<<<<<<<<<<<<<
 * cpdef make_error(int res, filename=*, log=*)
 */
struct __pyx_opt_args_2av_5error_err_check {
  int __pyx_n;
  PyObject *filename;
};

/* "av/error.pxd":5
 * 
 * cpdef int err_check(int res, filename=*) except -1
 * cpdef make_error(int res, filename=*, log=*)             # <<<<<<<<<<<<<<
 */
struct __pyx_opt_args_2av_5error_make_error {
  int __pyx_n;
  PyObject *filename;
  PyObject *log;
};
struct __pyx_opt_args_2av_5codec_7context_12CodecContext_open;
struct __pyx_opt_args_2av_5codec_7context_12CodecContext_close;
struct __pyx_opt_args_2av_5codec_7context_12CodecContext_encode;
struct __pyx_opt_args_2av_5codec_7context_12CodecContext_decode;

/* "av/codec/context.pxd":31
 * 
 *     # Public API.
 *     cpdef open(self, bint strict=?)             # <<<<<<<<<<<<<<
 *     cpdef close(self, bint strict=?)
 * 
 */
struct __pyx_opt_args_2av_5codec_7context_12CodecContext_open {
  int __pyx_n;
  int strict;
};

/* "av/codec/context.pxd":32
 *     # Public API.
 *     cpdef open(self, bint strict=?)
 *     cpdef close(self, bint strict=?)             # <<<<<<<<<<<<<<
 * 
 *     cdef _set_default_time_base(self)
 */
struct __pyx_opt_args_2av_5codec_7context_12CodecContext_close {
  int __pyx_n;
  int strict;
};

/* "av/codec/context.pxd":37
 * 
 *     # Wraps both versions of the transcode API, returning lists.
 *     cpdef encode(self, Frame frame=?)             # <<<<<<<<<<<<<<
 *     cpdef decode(self, Packet packet=?)
 * 
 */
struct __pyx_opt_args_2av_5codec_7context_12CodecContext_encode {
  int __pyx_n;
  struct __pyx_obj_2av_5frame_Frame *frame;
};

/* "av/codec/context.pxd":38
 *     # Wraps both versions of the transcode API, returning lists.
 *     cpdef encode(self, Frame frame=?)
 *     cpdef decode(self, Packet packet=?)             # <<<<<<<<<<<<<<
 * 
 *     # Used by both transcode APIs to setup user-land objects.
 */
struct __pyx_opt_args_2av_5codec_7context_12CodecContext_decode {
  int __pyx_n;
  struct __pyx_obj_2av_6packet_Packet *packet;
};

/* "av/bytesource.pxd":4
 * 
 * 
 * cdef class ByteSource(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef object owner
 */
struct __pyx_obj_2av_10bytesource_ByteSource {
  PyObject_HEAD
  PyObject *owner;
  int has_view;
  Py_buffer view;
  unsigned char *ptr;
  size_t length;
};


/* "av/codec/codec.pxd":4
 * 
 * 
 * cdef class Codec(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef const lib.AVCodec *ptr
 */
struct __pyx_obj_2av_5codec_5codec_Codec {
  PyObject_HEAD
  struct __pyx_vtabstruct_2av_5codec_5codec_Codec *__pyx_vtab;
  struct AVCodec const *ptr;
  struct AVCodecDescriptor const *desc;
  int is_encoder;
};


/* "av/buffer.pxd":2
 * 
 * cdef class Buffer(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef size_t _buffer_size(self)
 */
struct __pyx_obj_2av_6buffer_Buffer {
  PyObject_HEAD
  struct __pyx_vtabstruct_2av_6buffer_Buffer *__pyx_vtab;
};


/* "av/container/pyio.pxd":16
 * 
 * 
 * cdef class PyIOFile(object):             # <<<<<<<<<<<<<<
 * 
 *     # File-like source.
 */
struct __pyx_obj_2av_9container_4pyio_PyIOFile {
  PyObject_HEAD
  PyObject *file;
  PyObject *fread;
  PyObject *fwrite;
  PyObject *fseek;
  PyObject *ftell;
  PyObject *fclose;
  struct AVIOContext *iocontext;
  unsigned char *buffer;
  long pos;
  int pos_is_valid;
};


/* "av/container/streams.pxd":4
 * 
 * 
 * cdef class StreamContainer(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef list _streams
 */
struct __pyx_obj_2av_9container_7streams_StreamContainer {
  PyObject_HEAD
  struct __pyx_vtabstruct_2av_9container_7streams_StreamContainer *__pyx_vtab;
  PyObject *_streams;
  PyObject *video;
  PyObject *audio;
  PyObject *subtitles;
  PyObject *data;
  PyObject *other;
};


/* "av/dictionary.pxd":4
 * 
 * 
 * cdef class _Dictionary(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef lib.AVDictionary *ptr
 */
struct __pyx_obj_2av_10dictionary__Dictionary {
  PyObject_HEAD
  struct __pyx_vtabstruct_2av_10dictionary__Dictionary *__pyx_vtab;
  AVDictionary *ptr;
};


/* "av/format.pxd":4
 * 
 * 
 * cdef class ContainerFormat(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef readonly str name
 */
struct __pyx_obj_2av_6format_ContainerFormat {
  PyObject_HEAD
  PyObject *name;
  struct AVInputFormat *iptr;
  struct AVOutputFormat *optr;
};


/* "av/container/core.pxd":16
 * 
 * 
 * cdef class Container(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef readonly bint writeable
 */
struct __pyx_obj_2av_9container_4core_Container {
  PyObject_HEAD
  struct __pyx_vtabstruct_2av_9container_4core_Container *__pyx_vtab;
  int writeable;
  struct AVFormatContext *ptr;
  PyObject *name;
  PyObject *metadata_encoding;
  PyObject *metadata_errors;
  struct __pyx_obj_2av_9container_4pyio_PyIOFile *file;
  int buffer_size;
  int input_was_opened;
  PyObject *io_open;
  PyObject *open_files;
  struct __pyx_obj_2av_6format_ContainerFormat *format;
  PyObject *options;
  PyObject *container_options;
  PyObject *stream_options;
  struct __pyx_obj_2av_9container_7streams_StreamContainer *streams;
  PyObject *metadata;
  PyObject *open_timeout;
  PyObject *read_timeout;
  __pyx_t_2av_9container_4core_timeout_info interrupt_callback_info;
};


/* "av/stream.pxd":10
 * 
 * 
 * cdef class Stream(object):             # <<<<<<<<<<<<<<
 *     cdef lib.AVStream *ptr
 * 
 */
struct __pyx_obj_2av_6stream_Stream {
  PyObject_HEAD
  struct __pyx_vtabstruct_2av_6stream_Stream *__pyx_vtab;
  struct AVStream *ptr;
  struct __pyx_obj_2av_9container_4core_Container *container;
  PyObject *metadata;
  struct __pyx_obj_2av_5codec_7context_CodecContext *codec_context;
};


/* "av/packet.pxd":8
 * 
 * 
 * cdef class Packet(Buffer):             # <<<<<<<<<<<<<<
 * 
 *     cdef lib.AVPacket* ptr
 */
struct __pyx_obj_2av_6packet_Packet {
  struct __pyx_obj_2av_6buffer_Buffer __pyx_base;
  struct AVPacket *ptr;
  struct __pyx_obj_2av_6stream_Stream *_stream;
  AVRational _time_base;
  struct __pyx_obj_2av_10bytesource_ByteSource *source;
};


/* "av/sidedata/sidedata.pxd":9
 * 
 * 
 * cdef class SideData(Buffer):             # <<<<<<<<<<<<<<
 * 
 *     cdef Frame frame
 */
struct __pyx_obj_2av_8sidedata_8sidedata_SideData {
  struct __pyx_obj_2av_6buffer_Buffer __pyx_base;
  struct __pyx_obj_2av_5frame_Frame *frame;
  struct AVFrameSideData *ptr;
  struct __pyx_obj_2av_10dictionary__Dictionary *metadata;
};


/* "av/sidedata/sidedata.pxd":18
 * cdef SideData wrap_side_data(Frame frame, int index)
 * 
 * cdef class _SideDataContainer(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef Frame frame
 */
struct __pyx_obj_2av_8sidedata_8sidedata__SideDataContainer {
  PyObject_HEAD
  struct __pyx_obj_2av_5frame_Frame *frame;
  PyObject *_by_index;
  PyObject *_by_type;
};


/* "av/frame.pxd":7
 * 
 * 
 * cdef class Frame(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef lib.AVFrame *ptr
 */
struct __pyx_obj_2av_5frame_Frame {
  PyObject_HEAD
  struct __pyx_vtabstruct_2av_5frame_Frame *__pyx_vtab;
  struct AVFrame *ptr;
  AVRational _time_base;
  struct __pyx_obj_2av_8sidedata_8sidedata__SideDataContainer *_side_data;
  int index;
};


/* "av/codec/context.pxd":10
 * 
 * 
 * cdef class CodecContext(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef lib.AVCodecContext *ptr
 */
struct __pyx_obj_2av_5codec_7context_CodecContext {
  PyObject_HEAD
  struct __pyx_vtabstruct_2av_5codec_7context_CodecContext *__pyx_vtab;
  struct AVCodecContext *ptr;
  int extradata_set;
  int stream_index;
  struct AVCodecParserContext *parser;
  struct __pyx_obj_2av_5codec_5codec_Codec *codec;
  PyObject *options;
  struct __pyx_obj_2av_5frame_Frame *_next_frame;
};



/* "av/codec/codec.pxd":4
 * 
 * 
 * cdef class Codec(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef const lib.AVCodec *ptr
 */

struct __pyx_vtabstruct_2av_5codec_5codec_Codec {
  PyObject *(*_init)(struct __pyx_obj_2av_5codec_5codec_Codec *, struct __pyx_opt_args_2av_5codec_5codec_5Codec__init *__pyx_optional_args);
};
static struct __pyx_vtabstruct_2av_5codec_5codec_Codec *__pyx_vtabptr_2av_5codec_5codec_Codec;


/* "av/buffer.pxd":2
 * 
 * cdef class Buffer(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef size_t _buffer_size(self)
 */

struct __pyx_vtabstruct_2av_6buffer_Buffer {
  size_t (*_buffer_size)(struct __pyx_obj_2av_6buffer_Buffer *);
  void *(*_buffer_ptr)(struct __pyx_obj_2av_6buffer_Buffer *);
  int (*_buffer_writable)(struct __pyx_obj_2av_6buffer_Buffer *);
};
static struct __pyx_vtabstruct_2av_6buffer_Buffer *__pyx_vtabptr_2av_6buffer_Buffer;


/* "av/container/streams.pxd":4
 * 
 * 
 * cdef class StreamContainer(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef list _streams
 */

struct __pyx_vtabstruct_2av_9container_7streams_StreamContainer {
  PyObject *(*add_stream)(struct __pyx_obj_2av_9container_7streams_StreamContainer *, struct __pyx_obj_2av_6stream_Stream *);
};
static struct __pyx_vtabstruct_2av_9container_7streams_StreamContainer *__pyx_vtabptr_2av_9container_7streams_StreamContainer;


/* "av/dictionary.pxd":4
 * 
 * 
 * cdef class _Dictionary(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef lib.AVDictionary *ptr
 */

struct __pyx_vtabstruct_2av_10dictionary__Dictionary {
  struct __pyx_obj_2av_10dictionary__Dictionary *(*copy)(struct __pyx_obj_2av_10dictionary__Dictionary *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_2av_10dictionary__Dictionary *__pyx_vtabptr_2av_10dictionary__Dictionary;


/* "av/container/core.pxd":16
 * 
 * 
 * cdef class Container(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef readonly bint writeable
 */

struct __pyx_vtabstruct_2av_9container_4core_Container {
  int (*err_check)(struct __pyx_obj_2av_9container_4core_Container *, int);
  PyObject *(*set_timeout)(struct __pyx_obj_2av_9container_4core_Container *, PyObject *);
  PyObject *(*start_timeout)(struct __pyx_obj_2av_9container_4core_Container *);
};
static struct __pyx_vtabstruct_2av_9container_4core_Container *__pyx_vtabptr_2av_9container_4core_Container;


/* "av/stream.pxd":10
 * 
 * 
 * cdef class Stream(object):             # <<<<<<<<<<<<<<
 *     cdef lib.AVStream *ptr
 * 
 */

struct __pyx_vtabstruct_2av_6stream_Stream {
  PyObject *(*_init)(struct __pyx_obj_2av_6stream_Stream *, struct __pyx_obj_2av_9container_4core_Container *, struct AVStream *, struct __pyx_obj_2av_5codec_7context_CodecContext *);
  PyObject *(*_finalize_for_output)(struct __pyx_obj_2av_6stream_Stream *);
  PyObject *(*_set_time_base)(struct __pyx_obj_2av_6stream_Stream *, PyObject *);
  PyObject *(*_set_id)(struct __pyx_obj_2av_6stream_Stream *, PyObject *);
};
static struct __pyx_vtabstruct_2av_6stream_Stream *__pyx_vtabptr_2av_6stream_Stream;


/* "av/packet.pxd":8
 * 
 * 
 * cdef class Packet(Buffer):             # <<<<<<<<<<<<<<
 * 
 *     cdef lib.AVPacket* ptr
 */

struct __pyx_vtabstruct_2av_6packet_Packet {
  struct __pyx_vtabstruct_2av_6buffer_Buffer __pyx_base;
  PyObject *(*_rebase_time)(struct __pyx_obj_2av_6packet_Packet *, AVRational);
};
static struct __pyx_vtabstruct_2av_6packet_Packet *__pyx_vtabptr_2av_6packet_Packet;


/* "av/sidedata/sidedata.pxd":9
 * 
 * 
 * cdef class SideData(Buffer):             # <<<<<<<<<<<<<<
 * 
 *     cdef Frame frame
 */

struct __pyx_vtabstruct_2av_8sidedata_8sidedata_SideData {
  struct __pyx_vtabstruct_2av_6buffer_Buffer __pyx_base;
};
static struct __pyx_vtabstruct_2av_8sidedata_8sidedata_SideData *__pyx_vtabptr_2av_8sidedata_8sidedata_SideData;


/* "av/frame.pxd":7
 * 
 * 
 * cdef class Frame(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef lib.AVFrame *ptr
 */

struct __pyx_vtabstruct_2av_5frame_Frame {
  PyObject *(*_rebase_time)(struct __pyx_obj_2av_5frame_Frame *, AVRational);
  PyObject *(*_copy_internal_attributes)(struct __pyx_obj_2av_5frame_Frame *, struct __pyx_obj_2av_5frame_Frame *, struct __pyx_opt_args_2av_5frame_5Frame__copy_internal_attributes *__pyx_optional_args);
  PyObject *(*_init_user_attributes)(struct __pyx_obj_2av_5frame_Frame *);
};
static struct __pyx_vtabstruct_2av_5frame_Frame *__pyx_vtabptr_2av_5frame_Frame;


/* "av/codec/context.pyx":138
 * 
 * 
 * cdef class CodecContext(object):             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

struct __pyx_vtabstruct_2av_5codec_7context_CodecContext {
  PyObject *(*_init)(struct __pyx_obj_2av_5codec_7context_CodecContext *, struct AVCodecContext *, struct AVCodec const *);
  PyObject *(*open)(struct __pyx_obj_2av_5codec_7context_CodecContext *, int __pyx_skip_dispatch, struct __pyx_opt_args_2av_5codec_7context_12CodecContext_open *__pyx_optional_args);
  PyObject *(*close)(struct __pyx_obj_2av_5codec_7context_CodecContext *, int __pyx_skip_dispatch, struct __pyx_opt_args_2av_5codec_7context_12CodecContext_close *__pyx_optional_args);
  PyObject *(*_set_default_time_base)(struct __pyx_obj_2av_5codec_7context_CodecContext *);
  PyObject *(*encode)(struct __pyx_obj_2av_5codec_7context_CodecContext *, int __pyx_skip_dispatch, struct __pyx_opt_args_2av_5codec_7context_12CodecContext_encode *__pyx_optional_args);
  PyObject *(*decode)(struct __pyx_obj_2av_5codec_7context_CodecContext *, int __pyx_skip_dispatch, struct __pyx_opt_args_2av_5codec_7context_12CodecContext_decode *__pyx_optional_args);
  PyObject *(*_prepare_frames_for_encode)(struct __pyx_obj_2av_5codec_7context_CodecContext *, struct __pyx_obj_2av_5frame_Frame *);
  PyObject *(*_setup_encoded_packet)(struct __pyx_obj_2av_5codec_7context_CodecContext *, struct __pyx_obj_2av_6packet_Packet *);
  PyObject *(*_setup_decoded_frame)(struct __pyx_obj_2av_5codec_7context_CodecContext *, struct __pyx_obj_2av_5frame_Frame *, struct __pyx_obj_2av_6packet_Packet *);
  PyObject *(*_send_frame_and_recv)(struct __pyx_obj_2av_5codec_7context_CodecContext *, struct __pyx_obj_2av_5frame_Frame *);
  PyObject *(*_recv_packet)(struct __pyx_obj_2av_5codec_7context_CodecContext *);
  PyObject *(*_send_packet_and_recv)(struct __pyx_obj_2av_5codec_7context_CodecContext *, struct __pyx_obj_2av_6packet_Packet *);
  PyObject *(*_recv_frame)(struct __pyx_obj_2av_5codec_7context_CodecContext *);
  struct __pyx_obj_2av_5frame_Frame *(*_alloc_next_frame)(struct __pyx_obj_2av_5codec_7context_CodecContext *);
};
static struct __pyx_vtabstruct_2av_5codec_7context_CodecContext *__pyx_vtabptr_2av_5codec_7context_CodecContext;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if CYTHON_FAST_PYCALL
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif // CYTHON_FAST_PYCALL
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
#if __STDC_VERSION__ >= 201112L
#include <stdalign.h>
#endif
#if __STDC_VERSION__ >= 201112L || __cplusplus >= 201103L
#define __PYX_GET_STRUCT_ALIGNMENT(s) alignof(s)
#else
#define __PYX_GET_STRUCT_ALIGNMENT(s) sizeof(void*)
#endif
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyObject *dict);

/* GetNameInClass.proto */
#define __Pyx_GetNameInClass(var, nmspace, name)  (var) = __Pyx__GetNameInClass(nmspace, name)
static PyObject *__Pyx__GetNameInClass(PyObject *nmspace, PyObject *name);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* GCCDiagnostics.proto */
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__AVDiscard(enum AVDiscard value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_int(unsigned int value);

/* CIntFromPy.proto */
static CYTHON_INLINE unsigned int __Pyx_PyInt_As_unsigned_int(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int64_t(int64_t value);

/* CIntFromPy.proto */
static CYTHON_INLINE enum AVDiscard __Pyx_PyInt_As_enum__AVDiscard(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* FunctionExport.proto */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig);

/* FunctionImport.proto */
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static PyObject *__pyx_f_2av_5codec_7context_12CodecContext__init(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, struct AVCodecContext *__pyx_v_ptr, struct AVCodec const *__pyx_v_codec); /* proto*/
static PyObject *__pyx_f_2av_5codec_7context_12CodecContext_open(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, int __pyx_skip_dispatch, struct __pyx_opt_args_2av_5codec_7context_12CodecContext_open *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_2av_5codec_7context_12CodecContext__set_default_time_base(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_2av_5codec_7context_12CodecContext_close(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, int __pyx_skip_dispatch, struct __pyx_opt_args_2av_5codec_7context_12CodecContext_close *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_2av_5codec_7context_12CodecContext__send_frame_and_recv(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, struct __pyx_obj_2av_5frame_Frame *__pyx_v_frame); /* proto*/
static PyObject *__pyx_f_2av_5codec_7context_12CodecContext__send_packet_and_recv(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, struct __pyx_obj_2av_6packet_Packet *__pyx_v_packet); /* proto*/
static PyObject *__pyx_f_2av_5codec_7context_12CodecContext__prepare_frames_for_encode(CYTHON_UNUSED struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, struct __pyx_obj_2av_5frame_Frame *__pyx_v_frame); /* proto*/
static struct __pyx_obj_2av_5frame_Frame *__pyx_f_2av_5codec_7context_12CodecContext__alloc_next_frame(CYTHON_UNUSED struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_2av_5codec_7context_12CodecContext__recv_frame(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_2av_5codec_7context_12CodecContext__recv_packet(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_2av_5codec_7context_12CodecContext_encode(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, int __pyx_skip_dispatch, struct __pyx_opt_args_2av_5codec_7context_12CodecContext_encode *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_2av_5codec_7context_12CodecContext__setup_encoded_packet(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, struct __pyx_obj_2av_6packet_Packet *__pyx_v_packet); /* proto*/
static PyObject *__pyx_f_2av_5codec_7context_12CodecContext_decode(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, int __pyx_skip_dispatch, struct __pyx_opt_args_2av_5codec_7context_12CodecContext_decode *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_2av_5codec_7context_12CodecContext__setup_decoded_frame(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, struct __pyx_obj_2av_5frame_Frame *__pyx_v_frame, struct __pyx_obj_2av_6packet_Packet *__pyx_v_packet); /* proto*/

/* Module declarations from 'libc.stdint' */

/* Module declarations from 'libav' */

/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'av.bytesource' */
static PyTypeObject *__pyx_ptype_2av_10bytesource_ByteSource = 0;
static struct __pyx_obj_2av_10bytesource_ByteSource *(*__pyx_f_2av_10bytesource_bytesource)(PyObject *, struct __pyx_opt_args_2av_10bytesource_bytesource *__pyx_optional_args); /*proto*/

/* Module declarations from 'av.codec.codec' */
static PyTypeObject *__pyx_ptype_2av_5codec_5codec_Codec = 0;
static struct __pyx_obj_2av_5codec_5codec_Codec *(*__pyx_f_2av_5codec_5codec_wrap_codec)(struct AVCodec const *); /*proto*/

/* Module declarations from 'av.buffer' */
static PyTypeObject *__pyx_ptype_2av_6buffer_Buffer = 0;

/* Module declarations from 'av.container.pyio' */
static PyTypeObject *__pyx_ptype_2av_9container_4pyio_PyIOFile = 0;

/* Module declarations from 'av.container.streams' */
static PyTypeObject *__pyx_ptype_2av_9container_7streams_StreamContainer = 0;

/* Module declarations from 'av.dictionary' */
static PyTypeObject *__pyx_ptype_2av_10dictionary__Dictionary = 0;
static struct __pyx_obj_2av_10dictionary__Dictionary *(*__pyx_f_2av_10dictionary_wrap_dictionary)(AVDictionary *); /*proto*/

/* Module declarations from 'av.format' */
static PyTypeObject *__pyx_ptype_2av_6format_ContainerFormat = 0;

/* Module declarations from 'av.container.core' */
static PyTypeObject *__pyx_ptype_2av_9container_4core_Container = 0;

/* Module declarations from 'av.stream' */
static PyTypeObject *__pyx_ptype_2av_6stream_Stream = 0;

/* Module declarations from 'av.packet' */
static PyTypeObject *__pyx_ptype_2av_6packet_Packet = 0;

/* Module declarations from 'av.sidedata.sidedata' */
static PyTypeObject *__pyx_ptype_2av_8sidedata_8sidedata_SideData = 0;
static PyTypeObject *__pyx_ptype_2av_8sidedata_8sidedata__SideDataContainer = 0;

/* Module declarations from 'av.frame' */
static PyTypeObject *__pyx_ptype_2av_5frame_Frame = 0;

/* Module declarations from 'libc.errno' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'av.enum' */
static PyObject *(*__pyx_f_2av_4enum_define_enum)(PyObject *, PyObject *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_2av_4enum_define_enum *__pyx_optional_args); /*proto*/

/* Module declarations from 'av.error' */
static int (*__pyx_f_2av_5error_err_check)(int, int __pyx_skip_dispatch, struct __pyx_opt_args_2av_5error_err_check *__pyx_optional_args); /*proto*/

/* Module declarations from 'av.utils' */
static PyObject *(*__pyx_f_2av_5utils_avrational_to_fraction)(AVRational const *); /*proto*/
static PyObject *(*__pyx_f_2av_5utils_to_avrational)(PyObject *, AVRational *); /*proto*/

/* Module declarations from 'av.codec.context' */
static PyTypeObject *__pyx_ptype_2av_5codec_7context_CodecContext = 0;
static PyObject *__pyx_v_2av_5codec_7context__cinit_sentinel = 0;
static struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_f_2av_5codec_7context_wrap_codec_context(struct AVCodecContext *, struct AVCodec const *); /*proto*/
#define __Pyx_MODULE_NAME "av.codec.context"
extern int __pyx_module_is_main_av__codec__context;
int __pyx_module_is_main_av__codec__context = 0;

/* Implementation of 'av.codec.context' */
static PyObject *__pyx_builtin_object;
static PyObject *__pyx_builtin_staticmethod;
static PyObject *__pyx_builtin_RuntimeError;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_MemoryError;
static PyObject *__pyx_builtin_id;
static PyObject *__pyx_builtin_NotImplementedError;
static PyObject *__pyx_builtin_TypeError;
static const char __pyx_k_id[] = "id";
static const char __pyx_k_4MV[] = "4MV";
static const char __pyx_k_ALL[] = "ALL";
static const char __pyx_k_get[] = "get";
static const char __pyx_k_AUTO[] = "AUTO";
static const char __pyx_k_FAST[] = "FAST";
static const char __pyx_k_GRAY[] = "GRAY";
static const char __pyx_k_NONE[] = "NONE";
static const char __pyx_k_PSNR[] = "PSNR";
static const char __pyx_k_QPEL[] = "QPEL";
static const char __pyx_k_fast[] = "fast";
static const char __pyx_k_gray[] = "gray";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_mode[] = "mode";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_open[] = "open";
static const char __pyx_k_psnr[] = "psnr";
static const char __pyx_k_qpel[] = "qpel";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_type[] = "type";
static const char __pyx_k_warn[] = "warn";
static const char __pyx_k_BIDIR[] = "BIDIR";
static const char __pyx_k_FRAME[] = "FRAME";
static const char __pyx_k_Flags[] = "Flags";
static const char __pyx_k_PASS1[] = "PASS1";
static const char __pyx_k_PASS2[] = "PASS2";
static const char __pyx_k_SLICE[] = "SLICE";
static const char __pyx_k_ascii[] = "ascii";
static const char __pyx_k_c_ctx[] = "c_ctx";
static const char __pyx_k_class[] = "__class__";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_codec[] = "codec";
static const char __pyx_k_flags[] = "flags";
static const char __pyx_k_frame[] = "frame";
static const char __pyx_k_get_2[] = "_get";
static const char __pyx_k_pass1[] = "pass1";
static const char __pyx_k_pass2[] = "pass2";
static const char __pyx_k_value[] = "value";
static const char __pyx_k_CHUNKS[] = "CHUNKS";
static const char __pyx_k_Flags2[] = "Flags2";
static const char __pyx_k_NONKEY[] = "NONKEY";
static const char __pyx_k_NONREF[] = "NONREF";
static const char __pyx_k_QSCALE[] = "QSCALE";
static const char __pyx_k_chunks[] = "chunks";
static const char __pyx_k_create[] = "create";
static const char __pyx_k_decode[] = "decode";
static const char __pyx_k_encode[] = "encode";
static const char __pyx_k_flags2[] = "flags2";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_little[] = "little";
static const char __pyx_k_name_2[] = "name";
static const char __pyx_k_notype[] = "<notype>";
static const char __pyx_k_object[] = "object";
static const char __pyx_k_packet[] = "packet";
static const char __pyx_k_qscale[] = "qscale";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_signed[] = "signed";
static const char __pyx_k_strict[] = "strict";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_AC_PRED[] = "AC_PRED";
static const char __pyx_k_DEFAULT[] = "DEFAULT";
static const char __pyx_k_ac_pred[] = "ac_pred";
static const char __pyx_k_four_mv[] = "four_mv";
static const char __pyx_k_nocodec[] = "<nocodec>";
static const char __pyx_k_BITEXACT[] = "BITEXACT";
static const char __pyx_k_NONINTRA[] = "NONINTRA";
static const char __pyx_k_SHOW_ALL[] = "SHOW_ALL";
static const char __pyx_k_SkipType[] = "SkipType";
static const char __pyx_k_bitexact[] = "bitexact";
static const char __pyx_k_cy_codec[] = "cy_codec";
static const char __pyx_k_encoding[] = "encoding";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_property[] = "property";
static const char __pyx_k_sentinel[] = "sentinel";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_show_all[] = "show_all";
static const char __pyx_k_to_bytes[] = "to_bytes";
static const char __pyx_k_warnings[] = "warnings";
static const char __pyx_k_LOW_DELAY[] = "LOW_DELAY";
static const char __pyx_k_NO_OUTPUT[] = "NO_OUTPUT";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_UNALIGNED[] = "UNALIGNED";
static const char __pyx_k_byteorder[] = "byteorder";
static const char __pyx_k_get_flags[] = "_get_flags";
static const char __pyx_k_low_delay[] = "low_delay";
static const char __pyx_k_no_output[] = "no_output";
static const char __pyx_k_raw_input[] = "raw_input";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_set_flags[] = "_set_flags";
static const char __pyx_k_unaligned[] = "unaligned";
static const char __pyx_k_CLOSED_GOP[] = "CLOSED_GOP";
static const char __pyx_k_Dictionary[] = "Dictionary";
static const char __pyx_k_EXPORT_MVS[] = "EXPORT_MVS";
static const char __pyx_k_ThreadType[] = "ThreadType";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_closed_gop[] = "closed_gop";
static const char __pyx_k_export_mvs[] = "export_mvs";
static const char __pyx_k_from_bytes[] = "from_bytes";
static const char __pyx_k_get_flags2[] = "_get_flags2";
static const char __pyx_k_is_decoder[] = "is_decoder";
static const char __pyx_k_is_encoder[] = "is_encoder";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_set_flags2[] = "_set_flags2";
static const char __pyx_k_DROPCHANGED[] = "DROPCHANGED";
static const char __pyx_k_Discard_all[] = "Discard all";
static const char __pyx_k_IGNORE_CROP[] = "IGNORE_CROP";
static const char __pyx_k_LOOP_FILTER[] = "LOOP_FILTER";
static const char __pyx_k_MemoryError[] = "MemoryError";
static const char __pyx_k_SKIP_MANUAL[] = "SKIP_MANUAL";
static const char __pyx_k_Use_qpel_MC[] = "Use qpel MC.";
static const char __pyx_k_ignore_crop[] = "ignore_crop";
static const char __pyx_k_loop_filter[] = "loop filter.";
static const char __pyx_k_skip_manual[] = "skip_manual";
static const char __pyx_k_CodecContext[] = "CodecContext";
static const char __pyx_k_LOCAL_HEADER[] = "LOCAL_HEADER";
static const char __pyx_k_RuntimeError[] = "RuntimeError";
static const char __pyx_k_drop_changed[] = "drop_changed";
static const char __pyx_k_local_header[] = "local_header";
static const char __pyx_k_staticmethod[] = "staticmethod";
static const char __pyx_k_GLOBAL_HEADER[] = "GLOBAL_HEADER";
static const char __pyx_k_INTERLACED_ME[] = "INTERLACED_ME";
static const char __pyx_k_RO_FLUSH_NOOP[] = "RO_FLUSH_NOOP";
static const char __pyx_k_av_dictionary[] = "av.dictionary";
static const char __pyx_k_flag_property[] = "flag_property";
static const char __pyx_k_global_header[] = "global_header";
static const char __pyx_k_interlaced_me[] = "interlaced_me";
static const char __pyx_k_loop_filter_2[] = "loop_filter";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_ro_flush_noop[] = "ro_flush_noop";
static const char __pyx_k_INTERLACED_DCT[] = "INTERLACED_DCT";
static const char __pyx_k_OUTPUT_CORRUPT[] = "OUTPUT_CORRUPT";
static const char __pyx_k_av_deprecation[] = "av.deprecation";
static const char __pyx_k_interlaced_dct[] = "interlaced_dct";
static const char __pyx_k_output_corrupt[] = "output_corrupt";
static const char __pyx_k_Discard_nothing[] = "Discard nothing";
static const char __pyx_k_Force_low_delay[] = "Force low delay.";
static const char __pyx_k_No_parser_for_s[] = "No parser for %s";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_Use_fixed_qscale[] = "Use fixed qscale.";
static const char __pyx_k_av_codec_context[] = "av.codec.context";
static const char __pyx_k_av_s_s_s_at_0x_x[] = "<av.%s %s/%s at 0x%x>";
static const char __pyx_k_AudioCodecContext[] = "AudioCodecContext";
static const char __pyx_k_VideoCodecContext[] = "VideoCodecContext";
static const char __pyx_k_Use_interlaced_DCT[] = "Use interlaced DCT.";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_NotImplementedError[] = "NotImplementedError";
static const char __pyx_k_AVDeprecationWarning[] = "AVDeprecationWarning";
static const char __pyx_k_SubtitleCodecContext[] = "SubtitleCodecContext";
static const char __pyx_k_av_codec_context_pyx[] = "av/codec/context.pyx";
static const char __pyx_k_av_audio_codeccontext[] = "av.audio.codeccontext";
static const char __pyx_k_av_video_codeccontext[] = "av.video.codeccontext";
static const char __pyx_k_Skip_bitstream_encoding[] = "Skip bitstream encoding.";
static const char __pyx_k_Cannot_allocate_extradata[] = "Cannot allocate extradata";
static const char __pyx_k_Discard_all_non_reference[] = "Discard all non reference";
static const char __pyx_k_av_subtitles_codeccontext[] = "av.subtitles.codeccontext";
static const char __pyx_k_cannot_decode_unknown_codec[] = "cannot decode unknown codec";
static const char __pyx_k_CodecContext_is_already_open[] = "CodecContext is already open.";
static const char __pyx_k_Discard_all_non_intra_frames[] = "Discard all non intra frames";
static const char __pyx_k_Flag_property_of_class_Flags[] = "Flag property of :class:`.Flags`.";
static const char __pyx_k_Interlaced_motion_estimation[] = "Interlaced motion estimation";
static const char __pyx_k_Only_decode_encode_grayscale[] = "Only decode/encode grayscale.";
static const char __pyx_k_Flag_property_of_class_Flags2[] = "Flag property of :class:`.Flags2`.";
static const char __pyx_k_CodecContext_is_already_closed[] = "CodecContext is already closed.";
static const char __pyx_k_Base_CodecContext_cannot_decode[] = "Base CodecContext cannot decode.";
static const char __pyx_k_Cannot_change_thread_type_after[] = "Cannot change thread_type after codec is open.";
static const char __pyx_k_Cannot_instantiate_CodecContext[] = "Cannot instantiate CodecContext";
static const char __pyx_k_4_MV_per_MB_allowed_advanced_pre[] = "4 MV per MB allowed / advanced prediction for H.263.";
static const char __pyx_k_Allow_decoders_to_produce_frames[] = "Allow decoders to produce frames with data planes that are not aligned\n        to CPU requirements (e.g. due to cropping).";
static const char __pyx_k_Allow_non_spec_compliant_speedup[] = "Allow non spec compliant speedup tricks.";
static const char __pyx_k_Can_only_set_extradata_for_decod[] = "Can only set extradata for decoders.";
static const char __pyx_k_Cannot_change_thread_count_after[] = "Cannot change thread_count after codec is open.";
static const char __pyx_k_Codec_tag_should_be_a_4_characte[] = "Codec tag should be a 4 character string.";
static const char __pyx_k_Decode_more_than_one_frame_at_on[] = "Decode more than one frame at once";
static const char __pyx_k_Decode_more_than_one_part_of_a_s[] = "Decode more than one part of a single frame at once";
static const char __pyx_k_Decode_using_both_FRAME_and_SLIC[] = "Decode using both FRAME and SLICE methods.";
static const char __pyx_k_Discard_all_bidirectional_frames[] = "Discard all bidirectional frames";
static const char __pyx_k_Discard_all_frames_except_keyfra[] = "Discard all frames except keyframes";
static const char __pyx_k_Discard_cropping_information_fro[] = "Discard cropping information from SPS.";
static const char __pyx_k_Discard_useless_packets_like_0_s[] = "Discard useless packets like 0 size packets in AVI";
static const char __pyx_k_Do_not_reset_ASS_ReadOrder_field[] = "Do not reset ASS ReadOrder field on flush (subtitles decoding)";
static const char __pyx_k_Do_not_skip_samples_and_export_s[] = "Do not skip samples and export skip information as frame side data";
static const char __pyx_k_Don_t_output_frames_whose_parame[] = "Don't output frames whose parameters differ from first\n        decoded frame in stream.";
static const char __pyx_k_Encoding_is_only_supported_for_a[] = "Encoding is only supported for audio and video.";
static const char __pyx_k_Export_motion_vectors_through_fr[] = "Export motion vectors through frame side data";
static const char __pyx_k_H_263_advanced_intra_coding_MPEG[] = "H.263 advanced intra coding / MPEG-4 AC prediction";
static const char __pyx_k_Input_bitstream_might_be_truncat[] = "Input bitstream might be truncated at a packet boundaries\n        instead of only at frame boundaries.";
static const char __pyx_k_Output_even_those_frames_that_mi[] = "Output even those frames that might be corrupted.";
static const char __pyx_k_Place_global_headers_at_every_ke[] = "Place global headers at every keyframe instead of in extradata.";
static const char __pyx_k_Place_global_headers_in_extradat[] = "Place global headers in extradata instead of every keyframe.";
static const char __pyx_k_Show_all_frames_before_the_first[] = "Show all frames before the first keyframe";
static const char __pyx_k_Use_internal_2pass_ratecontrol_i[] = "Use internal 2pass ratecontrol in first pass mode.";
static const char __pyx_k_Use_only_bitexact_stuff_except_I[] = "Use only bitexact stuff (except (I)DCT).";
static const char __pyx_k_Using_CodecContext_time_base_for[] = "Using CodecContext.time_base for decoders is deprecated.";
static const char __pyx_k_Wrapping_CodecContext_with_misma[] = "Wrapping CodecContext with mismatched codec.";
static const char __pyx_k_error_variables_will_be_set_duri[] = "error[?] variables will be set during encoding.";
static const char __pyx_k_no_default___reduce___due_to_non[] = "no default __reduce__ due to non-trivial __cinit__";
static const char __pyx_k_Use_internal_2pass_ratecontrol_i_2[] = "Use internal 2pass ratecontrol in second pass mode.";
static PyObject *__pyx_kp_s_4MV;
static PyObject *__pyx_kp_s_4_MV_per_MB_allowed_advanced_pre;
static PyObject *__pyx_n_s_AC_PRED;
static PyObject *__pyx_n_s_ALL;
static PyObject *__pyx_n_s_AUTO;
static PyObject *__pyx_n_s_AVDeprecationWarning;
static PyObject *__pyx_kp_s_Allow_decoders_to_produce_frames;
static PyObject *__pyx_kp_s_Allow_non_spec_compliant_speedup;
static PyObject *__pyx_n_s_AudioCodecContext;
static PyObject *__pyx_n_s_BIDIR;
static PyObject *__pyx_n_s_BITEXACT;
static PyObject *__pyx_kp_s_Base_CodecContext_cannot_decode;
static PyObject *__pyx_n_s_CHUNKS;
static PyObject *__pyx_n_s_CLOSED_GOP;
static PyObject *__pyx_kp_s_Can_only_set_extradata_for_decod;
static PyObject *__pyx_kp_s_Cannot_allocate_extradata;
static PyObject *__pyx_kp_s_Cannot_change_thread_count_after;
static PyObject *__pyx_kp_s_Cannot_change_thread_type_after;
static PyObject *__pyx_kp_s_Cannot_instantiate_CodecContext;
static PyObject *__pyx_n_s_CodecContext;
static PyObject *__pyx_kp_s_CodecContext_is_already_closed;
static PyObject *__pyx_kp_s_CodecContext_is_already_open;
static PyObject *__pyx_kp_s_Codec_tag_should_be_a_4_characte;
static PyObject *__pyx_n_s_DEFAULT;
static PyObject *__pyx_n_s_DROPCHANGED;
static PyObject *__pyx_kp_s_Decode_more_than_one_frame_at_on;
static PyObject *__pyx_kp_s_Decode_more_than_one_part_of_a_s;
static PyObject *__pyx_kp_s_Decode_using_both_FRAME_and_SLIC;
static PyObject *__pyx_n_s_Dictionary;
static PyObject *__pyx_kp_s_Discard_all;
static PyObject *__pyx_kp_s_Discard_all_bidirectional_frames;
static PyObject *__pyx_kp_s_Discard_all_frames_except_keyfra;
static PyObject *__pyx_kp_s_Discard_all_non_intra_frames;
static PyObject *__pyx_kp_s_Discard_all_non_reference;
static PyObject *__pyx_kp_s_Discard_cropping_information_fro;
static PyObject *__pyx_kp_s_Discard_nothing;
static PyObject *__pyx_kp_s_Discard_useless_packets_like_0_s;
static PyObject *__pyx_kp_s_Do_not_reset_ASS_ReadOrder_field;
static PyObject *__pyx_kp_s_Do_not_skip_samples_and_export_s;
static PyObject *__pyx_kp_s_Don_t_output_frames_whose_parame;
static PyObject *__pyx_n_s_EXPORT_MVS;
static PyObject *__pyx_kp_s_Encoding_is_only_supported_for_a;
static PyObject *__pyx_kp_s_Export_motion_vectors_through_fr;
static PyObject *__pyx_n_s_FAST;
static PyObject *__pyx_n_s_FRAME;
static PyObject *__pyx_kp_s_Flag_property_of_class_Flags;
static PyObject *__pyx_kp_s_Flag_property_of_class_Flags2;
static PyObject *__pyx_n_s_Flags;
static PyObject *__pyx_n_s_Flags2;
static PyObject *__pyx_kp_s_Force_low_delay;
static PyObject *__pyx_n_s_GLOBAL_HEADER;
static PyObject *__pyx_n_s_GRAY;
static PyObject *__pyx_kp_s_H_263_advanced_intra_coding_MPEG;
static PyObject *__pyx_n_s_IGNORE_CROP;
static PyObject *__pyx_n_s_INTERLACED_DCT;
static PyObject *__pyx_n_s_INTERLACED_ME;
static PyObject *__pyx_kp_s_Input_bitstream_might_be_truncat;
static PyObject *__pyx_kp_s_Interlaced_motion_estimation;
static PyObject *__pyx_n_s_LOCAL_HEADER;
static PyObject *__pyx_n_s_LOOP_FILTER;
static PyObject *__pyx_n_s_LOW_DELAY;
static PyObject *__pyx_n_s_MemoryError;
static PyObject *__pyx_n_s_NONE;
static PyObject *__pyx_n_s_NONINTRA;
static PyObject *__pyx_n_s_NONKEY;
static PyObject *__pyx_n_s_NONREF;
static PyObject *__pyx_n_s_NO_OUTPUT;
static PyObject *__pyx_kp_s_No_parser_for_s;
static PyObject *__pyx_n_s_NotImplementedError;
static PyObject *__pyx_n_s_OUTPUT_CORRUPT;
static PyObject *__pyx_kp_s_Only_decode_encode_grayscale;
static PyObject *__pyx_kp_s_Output_even_those_frames_that_mi;
static PyObject *__pyx_n_s_PASS1;
static PyObject *__pyx_n_s_PASS2;
static PyObject *__pyx_n_s_PSNR;
static PyObject *__pyx_kp_s_Place_global_headers_at_every_ke;
static PyObject *__pyx_kp_s_Place_global_headers_in_extradat;
static PyObject *__pyx_n_s_QPEL;
static PyObject *__pyx_n_s_QSCALE;
static PyObject *__pyx_n_s_RO_FLUSH_NOOP;
static PyObject *__pyx_n_s_RuntimeError;
static PyObject *__pyx_n_s_SHOW_ALL;
static PyObject *__pyx_n_s_SKIP_MANUAL;
static PyObject *__pyx_n_s_SLICE;
static PyObject *__pyx_kp_s_Show_all_frames_before_the_first;
static PyObject *__pyx_n_s_SkipType;
static PyObject *__pyx_kp_s_Skip_bitstream_encoding;
static PyObject *__pyx_n_s_SubtitleCodecContext;
static PyObject *__pyx_n_s_ThreadType;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_n_s_UNALIGNED;
static PyObject *__pyx_kp_s_Use_fixed_qscale;
static PyObject *__pyx_kp_s_Use_interlaced_DCT;
static PyObject *__pyx_kp_s_Use_internal_2pass_ratecontrol_i;
static PyObject *__pyx_kp_s_Use_internal_2pass_ratecontrol_i_2;
static PyObject *__pyx_kp_s_Use_only_bitexact_stuff_except_I;
static PyObject *__pyx_kp_s_Use_qpel_MC;
static PyObject *__pyx_kp_s_Using_CodecContext_time_base_for;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_n_s_VideoCodecContext;
static PyObject *__pyx_kp_s_Wrapping_CodecContext_with_misma;
static PyObject *__pyx_n_s_ac_pred;
static PyObject *__pyx_n_s_ascii;
static PyObject *__pyx_n_s_av_audio_codeccontext;
static PyObject *__pyx_n_s_av_codec_context;
static PyObject *__pyx_kp_s_av_codec_context_pyx;
static PyObject *__pyx_n_s_av_deprecation;
static PyObject *__pyx_n_s_av_dictionary;
static PyObject *__pyx_kp_s_av_s_s_s_at_0x_x;
static PyObject *__pyx_n_s_av_subtitles_codeccontext;
static PyObject *__pyx_n_s_av_video_codeccontext;
static PyObject *__pyx_n_s_bitexact;
static PyObject *__pyx_n_s_byteorder;
static PyObject *__pyx_n_s_c_ctx;
static PyObject *__pyx_kp_s_cannot_decode_unknown_codec;
static PyObject *__pyx_n_s_chunks;
static PyObject *__pyx_n_s_class;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_close;
static PyObject *__pyx_n_s_closed_gop;
static PyObject *__pyx_n_s_codec;
static PyObject *__pyx_n_s_create;
static PyObject *__pyx_n_s_cy_codec;
static PyObject *__pyx_n_s_decode;
static PyObject *__pyx_n_s_drop_changed;
static PyObject *__pyx_n_s_encode;
static PyObject *__pyx_n_s_encoding;
static PyObject *__pyx_kp_s_error_variables_will_be_set_duri;
static PyObject *__pyx_n_s_export_mvs;
static PyObject *__pyx_n_s_fast;
static PyObject *__pyx_n_s_flag_property;
static PyObject *__pyx_n_s_flags;
static PyObject *__pyx_n_s_flags2;
static PyObject *__pyx_n_s_four_mv;
static PyObject *__pyx_n_s_frame;
static PyObject *__pyx_n_s_from_bytes;
static PyObject *__pyx_n_s_get;
static PyObject *__pyx_n_s_get_2;
static PyObject *__pyx_n_s_get_flags;
static PyObject *__pyx_n_s_get_flags2;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_n_s_global_header;
static PyObject *__pyx_n_s_gray;
static PyObject *__pyx_n_s_id;
static PyObject *__pyx_n_s_ignore_crop;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_interlaced_dct;
static PyObject *__pyx_n_s_interlaced_me;
static PyObject *__pyx_n_s_is_decoder;
static PyObject *__pyx_n_s_is_encoder;
static PyObject *__pyx_n_s_little;
static PyObject *__pyx_n_s_local_header;
static PyObject *__pyx_kp_s_loop_filter;
static PyObject *__pyx_n_s_loop_filter_2;
static PyObject *__pyx_n_s_low_delay;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_mode;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_name_2;
static PyObject *__pyx_kp_s_no_default___reduce___due_to_non;
static PyObject *__pyx_n_s_no_output;
static PyObject *__pyx_kp_s_nocodec;
static PyObject *__pyx_kp_s_notype;
static PyObject *__pyx_n_s_object;
static PyObject *__pyx_n_s_open;
static PyObject *__pyx_n_s_output_corrupt;
static PyObject *__pyx_n_s_packet;
static PyObject *__pyx_n_s_pass1;
static PyObject *__pyx_n_s_pass2;
static PyObject *__pyx_n_s_property;
static PyObject *__pyx_n_s_psnr;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_qpel;
static PyObject *__pyx_n_s_qscale;
static PyObject *__pyx_n_s_raw_input;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_ro_flush_noop;
static PyObject *__pyx_n_s_sentinel;
static PyObject *__pyx_n_s_set_flags;
static PyObject *__pyx_n_s_set_flags2;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_show_all;
static PyObject *__pyx_n_s_signed;
static PyObject *__pyx_n_s_skip_manual;
static PyObject *__pyx_n_s_staticmethod;
static PyObject *__pyx_n_s_strict;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_to_bytes;
static PyObject *__pyx_n_s_type;
static PyObject *__pyx_n_s_unaligned;
static PyObject *__pyx_n_s_update;
static PyObject *__pyx_n_s_value;
static PyObject *__pyx_n_s_warn;
static PyObject *__pyx_n_s_warnings;
static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_create(PyObject *__pyx_v_codec, PyObject *__pyx_v_mode); /* proto */
static int __pyx_pf_2av_5codec_7context_12CodecContext_2__cinit__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, PyObject *__pyx_v_sentinel, CYTHON_UNUSED PyObject *__pyx_v_args, CYTHON_UNUSED PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_4_get_flags(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_6_set_flags(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_8_get_flags2(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_10_set_flags2(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_9extradata___get__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self); /* proto */
static int __pyx_pf_2av_5codec_7context_12CodecContext_9extradata_2__set__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, PyObject *__pyx_v_data); /* proto */
static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_14extradata_size___get__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_7is_open___get__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_10is_encoder___get__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_10is_decoder___get__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_12open(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, int __pyx_v_strict); /* proto */
static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_14close(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, int __pyx_v_strict); /* proto */
static void __pyx_pf_2av_5codec_7context_12CodecContext_16__dealloc__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_18__repr__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_20parse(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, PyObject *__pyx_v_raw_input); /* proto */
static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_22encode(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, struct __pyx_obj_2av_5frame_Frame *__pyx_v_frame); /* proto */
static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_24decode(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, struct __pyx_obj_2av_6packet_Packet *__pyx_v_packet); /* proto */
static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_4name___get__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_4type___get__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_7profile___get__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_9time_base___get__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self); /* proto */
static int __pyx_pf_2av_5codec_7context_12CodecContext_9time_base_2__set__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_9codec_tag___get__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self); /* proto */
static int __pyx_pf_2av_5codec_7context_12CodecContext_9codec_tag_2__set__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_15ticks_per_frame___get__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_8bit_rate___get__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self); /* proto */
static int __pyx_pf_2av_5codec_7context_12CodecContext_8bit_rate_2__set__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, int __pyx_v_value); /* proto */
static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_12max_bit_rate___get__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_18bit_rate_tolerance___get__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self); /* proto */
static int __pyx_pf_2av_5codec_7context_12CodecContext_18bit_rate_tolerance_2__set__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, int __pyx_v_value); /* proto */
static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_12thread_count___get__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self); /* proto */
static int __pyx_pf_2av_5codec_7context_12CodecContext_12thread_count_2__set__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, int __pyx_v_value); /* proto */
static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_11thread_type___get__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self); /* proto */
static int __pyx_pf_2av_5codec_7context_12CodecContext_11thread_type_2__set__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_10skip_frame___get__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self); /* proto */
static int __pyx_pf_2av_5codec_7context_12CodecContext_10skip_frame_2__set__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_5codec___get__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_7options___get__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self); /* proto */
static int __pyx_pf_2av_5codec_7context_12CodecContext_7options_2__set__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_2av_5codec_7context_12CodecContext_7options_4__del__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_26__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_28__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_2av_5codec_7context_CodecContext(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_4;
static PyObject *__pyx_int_32;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__13;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_codeobj__19;
/* Late includes */

/* "av/codec/context.pyx":23
 * 
 * 
 * cdef CodecContext wrap_codec_context(lib.AVCodecContext *c_ctx, const lib.AVCodec *c_codec):             # <<<<<<<<<<<<<<
 *     """Build an av.CodecContext for an existing AVCodecContext."""
 * 
 */

static struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_f_2av_5codec_7context_wrap_codec_context(struct AVCodecContext *__pyx_v_c_ctx, struct AVCodec const *__pyx_v_c_codec) {
  struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_py_ctx = 0;
  PyObject *__pyx_v_VideoCodecContext = NULL;
  PyObject *__pyx_v_AudioCodecContext = NULL;
  PyObject *__pyx_v_SubtitleCodecContext = NULL;
  struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("wrap_codec_context", 0);

  /* "av/codec/context.pyx":29
 * 
 *     # TODO: This.
 *     if c_ctx.codec_type == lib.AVMEDIA_TYPE_VIDEO:             # <<<<<<<<<<<<<<
 *         from av.video.codeccontext import VideoCodecContext
 *         py_ctx = VideoCodecContext(_cinit_sentinel)
 */
  switch (__pyx_v_c_ctx->codec_type) {
    case AVMEDIA_TYPE_VIDEO:

    /* "av/codec/context.pyx":30
 *     # TODO: This.
 *     if c_ctx.codec_type == lib.AVMEDIA_TYPE_VIDEO:
 *         from av.video.codeccontext import VideoCodecContext             # <<<<<<<<<<<<<<
 *         py_ctx = VideoCodecContext(_cinit_sentinel)
 *     elif c_ctx.codec_type == lib.AVMEDIA_TYPE_AUDIO:
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 30, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_VideoCodecContext);
    __Pyx_GIVEREF(__pyx_n_s_VideoCodecContext);
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_VideoCodecContext);
    __pyx_t_2 = __Pyx_Import(__pyx_n_s_av_video_codeccontext, __pyx_t_1, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 30, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_VideoCodecContext); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 30, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_VideoCodecContext = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "av/codec/context.pyx":31
 *     if c_ctx.codec_type == lib.AVMEDIA_TYPE_VIDEO:
 *         from av.video.codeccontext import VideoCodecContext
 *         py_ctx = VideoCodecContext(_cinit_sentinel)             # <<<<<<<<<<<<<<
 *     elif c_ctx.codec_type == lib.AVMEDIA_TYPE_AUDIO:
 *         from av.audio.codeccontext import AudioCodecContext
 */
    __Pyx_INCREF(__pyx_v_VideoCodecContext);
    __pyx_t_1 = __pyx_v_VideoCodecContext; __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_v_2av_5codec_7context__cinit_sentinel) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_2av_5codec_7context__cinit_sentinel);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 31, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_2av_5codec_7context_CodecContext))))) __PYX_ERR(0, 31, __pyx_L1_error)
    __pyx_v_py_ctx = ((struct __pyx_obj_2av_5codec_7context_CodecContext *)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "av/codec/context.pyx":29
 * 
 *     # TODO: This.
 *     if c_ctx.codec_type == lib.AVMEDIA_TYPE_VIDEO:             # <<<<<<<<<<<<<<
 *         from av.video.codeccontext import VideoCodecContext
 *         py_ctx = VideoCodecContext(_cinit_sentinel)
 */
    break;
    case AVMEDIA_TYPE_AUDIO:

    /* "av/codec/context.pyx":33
 *         py_ctx = VideoCodecContext(_cinit_sentinel)
 *     elif c_ctx.codec_type == lib.AVMEDIA_TYPE_AUDIO:
 *         from av.audio.codeccontext import AudioCodecContext             # <<<<<<<<<<<<<<
 *         py_ctx = AudioCodecContext(_cinit_sentinel)
 *     elif c_ctx.codec_type == lib.AVMEDIA_TYPE_SUBTITLE:
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_AudioCodecContext);
    __Pyx_GIVEREF(__pyx_n_s_AudioCodecContext);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_AudioCodecContext);
    __pyx_t_1 = __Pyx_Import(__pyx_n_s_av_audio_codeccontext, __pyx_t_2, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_AudioCodecContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v_AudioCodecContext = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "av/codec/context.pyx":34
 *     elif c_ctx.codec_type == lib.AVMEDIA_TYPE_AUDIO:
 *         from av.audio.codeccontext import AudioCodecContext
 *         py_ctx = AudioCodecContext(_cinit_sentinel)             # <<<<<<<<<<<<<<
 *     elif c_ctx.codec_type == lib.AVMEDIA_TYPE_SUBTITLE:
 *         from av.subtitles.codeccontext import SubtitleCodecContext
 */
    __Pyx_INCREF(__pyx_v_AudioCodecContext);
    __pyx_t_2 = __pyx_v_AudioCodecContext; __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_2av_5codec_7context__cinit_sentinel) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_2av_5codec_7context__cinit_sentinel);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_2av_5codec_7context_CodecContext))))) __PYX_ERR(0, 34, __pyx_L1_error)
    __pyx_v_py_ctx = ((struct __pyx_obj_2av_5codec_7context_CodecContext *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "av/codec/context.pyx":32
 *         from av.video.codeccontext import VideoCodecContext
 *         py_ctx = VideoCodecContext(_cinit_sentinel)
 *     elif c_ctx.codec_type == lib.AVMEDIA_TYPE_AUDIO:             # <<<<<<<<<<<<<<
 *         from av.audio.codeccontext import AudioCodecContext
 *         py_ctx = AudioCodecContext(_cinit_sentinel)
 */
    break;
    case AVMEDIA_TYPE_SUBTITLE:

    /* "av/codec/context.pyx":36
 *         py_ctx = AudioCodecContext(_cinit_sentinel)
 *     elif c_ctx.codec_type == lib.AVMEDIA_TYPE_SUBTITLE:
 *         from av.subtitles.codeccontext import SubtitleCodecContext             # <<<<<<<<<<<<<<
 *         py_ctx = SubtitleCodecContext(_cinit_sentinel)
 *     else:
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_SubtitleCodecContext);
    __Pyx_GIVEREF(__pyx_n_s_SubtitleCodecContext);
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_SubtitleCodecContext);
    __pyx_t_2 = __Pyx_Import(__pyx_n_s_av_subtitles_codeccontext, __pyx_t_1, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_SubtitleCodecContext); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_SubtitleCodecContext = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "av/codec/context.pyx":37
 *     elif c_ctx.codec_type == lib.AVMEDIA_TYPE_SUBTITLE:
 *         from av.subtitles.codeccontext import SubtitleCodecContext
 *         py_ctx = SubtitleCodecContext(_cinit_sentinel)             # <<<<<<<<<<<<<<
 *     else:
 *         py_ctx = CodecContext(_cinit_sentinel)
 */
    __Pyx_INCREF(__pyx_v_SubtitleCodecContext);
    __pyx_t_1 = __pyx_v_SubtitleCodecContext; __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_v_2av_5codec_7context__cinit_sentinel) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_2av_5codec_7context__cinit_sentinel);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_2av_5codec_7context_CodecContext))))) __PYX_ERR(0, 37, __pyx_L1_error)
    __pyx_v_py_ctx = ((struct __pyx_obj_2av_5codec_7context_CodecContext *)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "av/codec/context.pyx":35
 *         from av.audio.codeccontext import AudioCodecContext
 *         py_ctx = AudioCodecContext(_cinit_sentinel)
 *     elif c_ctx.codec_type == lib.AVMEDIA_TYPE_SUBTITLE:             # <<<<<<<<<<<<<<
 *         from av.subtitles.codeccontext import SubtitleCodecContext
 *         py_ctx = SubtitleCodecContext(_cinit_sentinel)
 */
    break;
    default:

    /* "av/codec/context.pyx":39
 *         py_ctx = SubtitleCodecContext(_cinit_sentinel)
 *     else:
 *         py_ctx = CodecContext(_cinit_sentinel)             # <<<<<<<<<<<<<<
 * 
 *     py_ctx._init(c_ctx, c_codec)
 */
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext), __pyx_v_2av_5codec_7context__cinit_sentinel); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 39, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_py_ctx = ((struct __pyx_obj_2av_5codec_7context_CodecContext *)__pyx_t_2);
    __pyx_t_2 = 0;
    break;
  }

  /* "av/codec/context.pyx":41
 *         py_ctx = CodecContext(_cinit_sentinel)
 * 
 *     py_ctx._init(c_ctx, c_codec)             # <<<<<<<<<<<<<<
 * 
 *     return py_ctx
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_2av_5codec_7context_CodecContext *)__pyx_v_py_ctx->__pyx_vtab)->_init(__pyx_v_py_ctx, __pyx_v_c_ctx, __pyx_v_c_codec); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "av/codec/context.pyx":43
 *     py_ctx._init(c_ctx, c_codec)
 * 
 *     return py_ctx             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_py_ctx));
  __pyx_r = __pyx_v_py_ctx;
  goto __pyx_L0;

  /* "av/codec/context.pyx":23
 * 
 * 
 * cdef CodecContext wrap_codec_context(lib.AVCodecContext *c_ctx, const lib.AVCodec *c_codec):             # <<<<<<<<<<<<<<
 *     """Build an av.CodecContext for an existing AVCodecContext."""
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("av.codec.context.wrap_codec_context", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_py_ctx);
  __Pyx_XDECREF(__pyx_v_VideoCodecContext);
  __Pyx_XDECREF(__pyx_v_AudioCodecContext);
  __Pyx_XDECREF(__pyx_v_SubtitleCodecContext);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pyx":141
 * 
 *     @staticmethod
 *     def create(codec, mode=None):             # <<<<<<<<<<<<<<
 *         cdef Codec cy_codec = codec if isinstance(codec, Codec) else Codec(codec, mode)
 *         cdef lib.AVCodecContext *c_ctx = lib.avcodec_alloc_context3(cy_codec.ptr)
 */

/* Python wrapper */
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_1create(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_2av_5codec_7context_12CodecContext_create[] = "CodecContext.create(codec, mode=None)";
static PyMethodDef __pyx_mdef_2av_5codec_7context_12CodecContext_1create = {"create", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_2av_5codec_7context_12CodecContext_1create, METH_VARARGS|METH_KEYWORDS, __pyx_doc_2av_5codec_7context_12CodecContext_create};
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_1create(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_codec = 0;
  PyObject *__pyx_v_mode = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_codec,&__pyx_n_s_mode,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_codec)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mode);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "create") < 0)) __PYX_ERR(0, 141, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_codec = values[0];
    __pyx_v_mode = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("create", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 141, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("av.codec.context.CodecContext.create", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_2av_5codec_7context_12CodecContext_create(__pyx_v_codec, __pyx_v_mode);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_create(PyObject *__pyx_v_codec, PyObject *__pyx_v_mode) {
  struct __pyx_obj_2av_5codec_5codec_Codec *__pyx_v_cy_codec = 0;
  struct AVCodecContext *__pyx_v_c_ctx;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("create", 0);

  /* "av/codec/context.pyx":142
 *     @staticmethod
 *     def create(codec, mode=None):
 *         cdef Codec cy_codec = codec if isinstance(codec, Codec) else Codec(codec, mode)             # <<<<<<<<<<<<<<
 *         cdef lib.AVCodecContext *c_ctx = lib.avcodec_alloc_context3(cy_codec.ptr)
 *         return wrap_codec_context(c_ctx, cy_codec.ptr)
 */
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_codec, __pyx_ptype_2av_5codec_5codec_Codec); 
  if ((__pyx_t_2 != 0)) {
    if (!(likely(((__pyx_v_codec) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_codec, __pyx_ptype_2av_5codec_5codec_Codec))))) __PYX_ERR(0, 142, __pyx_L1_error)
    __Pyx_INCREF(__pyx_v_codec);
    __pyx_t_1 = __pyx_v_codec;
  } else {
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 142, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_codec);
    __Pyx_GIVEREF(__pyx_v_codec);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_codec);
    __Pyx_INCREF(__pyx_v_mode);
    __Pyx_GIVEREF(__pyx_v_mode);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_mode);
    __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_2av_5codec_5codec_Codec), __pyx_t_3, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 142, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_1 = __pyx_t_4;
    __pyx_t_4 = 0;
  }
  __pyx_v_cy_codec = ((struct __pyx_obj_2av_5codec_5codec_Codec *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "av/codec/context.pyx":143
 *     def create(codec, mode=None):
 *         cdef Codec cy_codec = codec if isinstance(codec, Codec) else Codec(codec, mode)
 *         cdef lib.AVCodecContext *c_ctx = lib.avcodec_alloc_context3(cy_codec.ptr)             # <<<<<<<<<<<<<<
 *         return wrap_codec_context(c_ctx, cy_codec.ptr)
 * 
 */
  __pyx_v_c_ctx = avcodec_alloc_context3(__pyx_v_cy_codec->ptr);

  /* "av/codec/context.pyx":144
 *         cdef Codec cy_codec = codec if isinstance(codec, Codec) else Codec(codec, mode)
 *         cdef lib.AVCodecContext *c_ctx = lib.avcodec_alloc_context3(cy_codec.ptr)
 *         return wrap_codec_context(c_ctx, cy_codec.ptr)             # <<<<<<<<<<<<<<
 * 
 *     def __cinit__(self, sentinel=None, *args, **kwargs):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_2av_5codec_7context_wrap_codec_context(__pyx_v_c_ctx, __pyx_v_cy_codec->ptr)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "av/codec/context.pyx":141
 * 
 *     @staticmethod
 *     def create(codec, mode=None):             # <<<<<<<<<<<<<<
 *         cdef Codec cy_codec = codec if isinstance(codec, Codec) else Codec(codec, mode)
 *         cdef lib.AVCodecContext *c_ctx = lib.avcodec_alloc_context3(cy_codec.ptr)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("av.codec.context.CodecContext.create", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_cy_codec);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pyx":146
 *         return wrap_codec_context(c_ctx, cy_codec.ptr)
 * 
 *     def __cinit__(self, sentinel=None, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if sentinel is not _cinit_sentinel:
 *             raise RuntimeError('Cannot instantiate CodecContext')
 */

/* Python wrapper */
static int __pyx_pw_2av_5codec_7context_12CodecContext_3__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_2av_5codec_7context_12CodecContext_3__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_sentinel = 0;
  CYTHON_UNUSED PyObject *__pyx_v_args = 0;
  CYTHON_UNUSED PyObject *__pyx_v_kwargs = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return -1;
  __Pyx_GOTREF(__pyx_v_kwargs);
  if (PyTuple_GET_SIZE(__pyx_args) > 1) {
    __pyx_v_args = PyTuple_GetSlice(__pyx_args, 1, PyTuple_GET_SIZE(__pyx_args));
    if (unlikely(!__pyx_v_args)) {
      __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
      __Pyx_RefNannyFinishContext();
      return -1;
    }
    __Pyx_GOTREF(__pyx_v_args);
  } else {
    __pyx_v_args = __pyx_empty_tuple; __Pyx_INCREF(__pyx_empty_tuple);
  }
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_sentinel,0};
    PyObject* values[1] = {0};
    values[0] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        default:
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sentinel);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t used_pos_args = (pos_args < 1) ? pos_args : 1;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, __pyx_v_kwargs, values, used_pos_args, "__cinit__") < 0)) __PYX_ERR(0, 146, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        default:
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
      }
    }
    __pyx_v_sentinel = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_DECREF(__pyx_v_args); __pyx_v_args = 0;
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
  __Pyx_AddTraceback("av.codec.context.CodecContext.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_2av_5codec_7context_12CodecContext_2__cinit__(((struct __pyx_obj_2av_5codec_7context_CodecContext *)__pyx_v_self), __pyx_v_sentinel, __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2av_5codec_7context_12CodecContext_2__cinit__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, PyObject *__pyx_v_sentinel, CYTHON_UNUSED PyObject *__pyx_v_args, CYTHON_UNUSED PyObject *__pyx_v_kwargs) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "av/codec/context.pyx":147
 * 
 *     def __cinit__(self, sentinel=None, *args, **kwargs):
 *         if sentinel is not _cinit_sentinel:             # <<<<<<<<<<<<<<
 *             raise RuntimeError('Cannot instantiate CodecContext')
 * 
 */
  __pyx_t_1 = (__pyx_v_sentinel != __pyx_v_2av_5codec_7context__cinit_sentinel);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (unlikely(__pyx_t_2)) {

    /* "av/codec/context.pyx":148
 *     def __cinit__(self, sentinel=None, *args, **kwargs):
 *         if sentinel is not _cinit_sentinel:
 *             raise RuntimeError('Cannot instantiate CodecContext')             # <<<<<<<<<<<<<<
 * 
 *         self.options = {}
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 148, __pyx_L1_error)

    /* "av/codec/context.pyx":147
 * 
 *     def __cinit__(self, sentinel=None, *args, **kwargs):
 *         if sentinel is not _cinit_sentinel:             # <<<<<<<<<<<<<<
 *             raise RuntimeError('Cannot instantiate CodecContext')
 * 
 */
  }

  /* "av/codec/context.pyx":150
 *             raise RuntimeError('Cannot instantiate CodecContext')
 * 
 *         self.options = {}             # <<<<<<<<<<<<<<
 *         self.stream_index = -1  # This is set by the container immediately.
 * 
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->options);
  __Pyx_DECREF(__pyx_v_self->options);
  __pyx_v_self->options = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "av/codec/context.pyx":151
 * 
 *         self.options = {}
 *         self.stream_index = -1  # This is set by the container immediately.             # <<<<<<<<<<<<<<
 * 
 *     cdef _init(self, lib.AVCodecContext *ptr, const lib.AVCodec *codec):
 */
  __pyx_v_self->stream_index = -1;

  /* "av/codec/context.pyx":146
 *         return wrap_codec_context(c_ctx, cy_codec.ptr)
 * 
 *     def __cinit__(self, sentinel=None, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if sentinel is not _cinit_sentinel:
 *             raise RuntimeError('Cannot instantiate CodecContext')
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("av.codec.context.CodecContext.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pyx":153
 *         self.stream_index = -1  # This is set by the container immediately.
 * 
 *     cdef _init(self, lib.AVCodecContext *ptr, const lib.AVCodec *codec):             # <<<<<<<<<<<<<<
 * 
 *         self.ptr = ptr
 */

static PyObject *__pyx_f_2av_5codec_7context_12CodecContext__init(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, struct AVCodecContext *__pyx_v_ptr, struct AVCodec const *__pyx_v_codec) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  struct AVCodec const *__pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_init", 0);

  /* "av/codec/context.pyx":155
 *     cdef _init(self, lib.AVCodecContext *ptr, const lib.AVCodec *codec):
 * 
 *         self.ptr = ptr             # <<<<<<<<<<<<<<
 *         if self.ptr.codec and codec and self.ptr.codec != codec:
 *             raise RuntimeError('Wrapping CodecContext with mismatched codec.')
 */
  __pyx_v_self->ptr = __pyx_v_ptr;

  /* "av/codec/context.pyx":156
 * 
 *         self.ptr = ptr
 *         if self.ptr.codec and codec and self.ptr.codec != codec:             # <<<<<<<<<<<<<<
 *             raise RuntimeError('Wrapping CodecContext with mismatched codec.')
 *         self.codec = wrap_codec(codec if codec != NULL else self.ptr.codec)
 */
  __pyx_t_2 = (__pyx_v_self->ptr->codec != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_codec != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_self->ptr->codec != __pyx_v_codec) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "av/codec/context.pyx":157
 *         self.ptr = ptr
 *         if self.ptr.codec and codec and self.ptr.codec != codec:
 *             raise RuntimeError('Wrapping CodecContext with mismatched codec.')             # <<<<<<<<<<<<<<
 *         self.codec = wrap_codec(codec if codec != NULL else self.ptr.codec)
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 157, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 157, __pyx_L1_error)

    /* "av/codec/context.pyx":156
 * 
 *         self.ptr = ptr
 *         if self.ptr.codec and codec and self.ptr.codec != codec:             # <<<<<<<<<<<<<<
 *             raise RuntimeError('Wrapping CodecContext with mismatched codec.')
 *         self.codec = wrap_codec(codec if codec != NULL else self.ptr.codec)
 */
  }

  /* "av/codec/context.pyx":158
 *         if self.ptr.codec and codec and self.ptr.codec != codec:
 *             raise RuntimeError('Wrapping CodecContext with mismatched codec.')
 *         self.codec = wrap_codec(codec if codec != NULL else self.ptr.codec)             # <<<<<<<<<<<<<<
 * 
 *         # Set reasonable threading defaults.
 */
  if (((__pyx_v_codec != NULL) != 0)) {
    __pyx_t_4 = __pyx_v_codec;
  } else {
    __pyx_t_4 = __pyx_v_self->ptr->codec;
  }
  __pyx_t_3 = ((PyObject *)__pyx_f_2av_5codec_5codec_wrap_codec(__pyx_t_4)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->codec);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->codec));
  __pyx_v_self->codec = ((struct __pyx_obj_2av_5codec_5codec_Codec *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "av/codec/context.pyx":163
 *         # count == 0 -> use as many threads as there are CPUs.
 *         # type == 2 -> thread within a frame. This does not change the API.
 *         self.ptr.thread_count = 0             # <<<<<<<<<<<<<<
 *         self.ptr.thread_type = 2
 * 
 */
  __pyx_v_self->ptr->thread_count = 0;

  /* "av/codec/context.pyx":164
 *         # type == 2 -> thread within a frame. This does not change the API.
 *         self.ptr.thread_count = 0
 *         self.ptr.thread_type = 2             # <<<<<<<<<<<<<<
 * 
 *     def _get_flags(self):
 */
  __pyx_v_self->ptr->thread_type = 2;

  /* "av/codec/context.pyx":153
 *         self.stream_index = -1  # This is set by the container immediately.
 * 
 *     cdef _init(self, lib.AVCodecContext *ptr, const lib.AVCodec *codec):             # <<<<<<<<<<<<<<
 * 
 *         self.ptr = ptr
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("av.codec.context.CodecContext._init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pyx":166
 *         self.ptr.thread_type = 2
 * 
 *     def _get_flags(self):             # <<<<<<<<<<<<<<
 *         return self.ptr.flags
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_5_get_flags(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_2av_5codec_7context_12CodecContext_4_get_flags[] = "CodecContext._get_flags()";
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_5_get_flags(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_flags (wrapper)", 0);
  __pyx_r = __pyx_pf_2av_5codec_7context_12CodecContext_4_get_flags(((struct __pyx_obj_2av_5codec_7context_CodecContext *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_4_get_flags(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_flags", 0);

  /* "av/codec/context.pyx":167
 * 
 *     def _get_flags(self):
 *         return self.ptr.flags             # <<<<<<<<<<<<<<
 * 
 *     def _set_flags(self, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->ptr->flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "av/codec/context.pyx":166
 *         self.ptr.thread_type = 2
 * 
 *     def _get_flags(self):             # <<<<<<<<<<<<<<
 *         return self.ptr.flags
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("av.codec.context.CodecContext._get_flags", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pyx":169
 *         return self.ptr.flags
 * 
 *     def _set_flags(self, value):             # <<<<<<<<<<<<<<
 *         self.ptr.flags = value
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_7_set_flags(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static char __pyx_doc_2av_5codec_7context_12CodecContext_6_set_flags[] = "CodecContext._set_flags(value)";
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_7_set_flags(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_set_flags (wrapper)", 0);
  __pyx_r = __pyx_pf_2av_5codec_7context_12CodecContext_6_set_flags(((struct __pyx_obj_2av_5codec_7context_CodecContext *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_6_set_flags(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_set_flags", 0);

  /* "av/codec/context.pyx":170
 * 
 *     def _set_flags(self, value):
 *         self.ptr.flags = value             # <<<<<<<<<<<<<<
 * 
 *     flags = Flags.property(
 */
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 170, __pyx_L1_error)
  __pyx_v_self->ptr->flags = __pyx_t_1;

  /* "av/codec/context.pyx":169
 *         return self.ptr.flags
 * 
 *     def _set_flags(self, value):             # <<<<<<<<<<<<<<
 *         self.ptr.flags = value
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("av.codec.context.CodecContext._set_flags", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pyx":197
 *     closed_gop = flags.flag_property('CLOSED_GOP')
 * 
 *     def _get_flags2(self):             # <<<<<<<<<<<<<<
 *         return self.ptr.flags2
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_9_get_flags2(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_2av_5codec_7context_12CodecContext_8_get_flags2[] = "CodecContext._get_flags2()";
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_9_get_flags2(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_flags2 (wrapper)", 0);
  __pyx_r = __pyx_pf_2av_5codec_7context_12CodecContext_8_get_flags2(((struct __pyx_obj_2av_5codec_7context_CodecContext *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_8_get_flags2(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_flags2", 0);

  /* "av/codec/context.pyx":198
 * 
 *     def _get_flags2(self):
 *         return self.ptr.flags2             # <<<<<<<<<<<<<<
 * 
 *     def _set_flags2(self, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->ptr->flags2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "av/codec/context.pyx":197
 *     closed_gop = flags.flag_property('CLOSED_GOP')
 * 
 *     def _get_flags2(self):             # <<<<<<<<<<<<<<
 *         return self.ptr.flags2
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("av.codec.context.CodecContext._get_flags2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pyx":200
 *         return self.ptr.flags2
 * 
 *     def _set_flags2(self, value):             # <<<<<<<<<<<<<<
 *         self.ptr.flags2 = value
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_11_set_flags2(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static char __pyx_doc_2av_5codec_7context_12CodecContext_10_set_flags2[] = "CodecContext._set_flags2(value)";
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_11_set_flags2(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_set_flags2 (wrapper)", 0);
  __pyx_r = __pyx_pf_2av_5codec_7context_12CodecContext_10_set_flags2(((struct __pyx_obj_2av_5codec_7context_CodecContext *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_10_set_flags2(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_set_flags2", 0);

  /* "av/codec/context.pyx":201
 * 
 *     def _set_flags2(self, value):
 *         self.ptr.flags2 = value             # <<<<<<<<<<<<<<
 * 
 *     flags2 = Flags2.property(
 */
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 201, __pyx_L1_error)
  __pyx_v_self->ptr->flags2 = __pyx_t_1;

  /* "av/codec/context.pyx":200
 *         return self.ptr.flags2
 * 
 *     def _set_flags2(self, value):             # <<<<<<<<<<<<<<
 *         self.ptr.flags2 = value
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("av.codec.context.CodecContext._set_flags2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pyx":220
 * 
 *     property extradata:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             if self.ptr.extradata_size > 0:
 *                 return <bytes>(<uint8_t*>self.ptr.extradata)[:self.ptr.extradata_size]
 */

/* Python wrapper */
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_9extradata_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_9extradata_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2av_5codec_7context_12CodecContext_9extradata___get__(((struct __pyx_obj_2av_5codec_7context_CodecContext *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_9extradata___get__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "av/codec/context.pyx":221
 *     property extradata:
 *         def __get__(self):
 *             if self.ptr.extradata_size > 0:             # <<<<<<<<<<<<<<
 *                 return <bytes>(<uint8_t*>self.ptr.extradata)[:self.ptr.extradata_size]
 *             else:
 */
  __pyx_t_1 = ((__pyx_v_self->ptr->extradata_size > 0) != 0);
  if (__pyx_t_1) {

    /* "av/codec/context.pyx":222
 *         def __get__(self):
 *             if self.ptr.extradata_size > 0:
 *                 return <bytes>(<uint8_t*>self.ptr.extradata)[:self.ptr.extradata_size]             # <<<<<<<<<<<<<<
 *             else:
 *                 return None
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyBytes_FromStringAndSize(((const char*)((uint8_t *)__pyx_v_self->ptr->extradata)) + 0, __pyx_v_self->ptr->extradata_size - 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 222, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject*)__pyx_t_2));
    __pyx_r = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "av/codec/context.pyx":221
 *     property extradata:
 *         def __get__(self):
 *             if self.ptr.extradata_size > 0:             # <<<<<<<<<<<<<<
 *                 return <bytes>(<uint8_t*>self.ptr.extradata)[:self.ptr.extradata_size]
 *             else:
 */
  }

  /* "av/codec/context.pyx":224
 *                 return <bytes>(<uint8_t*>self.ptr.extradata)[:self.ptr.extradata_size]
 *             else:
 *                 return None             # <<<<<<<<<<<<<<
 * 
 *         def __set__(self, data):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;
  }

  /* "av/codec/context.pyx":220
 * 
 *     property extradata:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             if self.ptr.extradata_size > 0:
 *                 return <bytes>(<uint8_t*>self.ptr.extradata)[:self.ptr.extradata_size]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("av.codec.context.CodecContext.extradata.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pyx":226
 *                 return None
 * 
 *         def __set__(self, data):             # <<<<<<<<<<<<<<
 *             if not self.is_decoder:
 *                 raise ValueError("Can only set extradata for decoders.")
 */

/* Python wrapper */
static int __pyx_pw_2av_5codec_7context_12CodecContext_9extradata_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_data); /*proto*/
static int __pyx_pw_2av_5codec_7context_12CodecContext_9extradata_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_data) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2av_5codec_7context_12CodecContext_9extradata_2__set__(((struct __pyx_obj_2av_5codec_7context_CodecContext *)__pyx_v_self), ((PyObject *)__pyx_v_data));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2av_5codec_7context_12CodecContext_9extradata_2__set__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, PyObject *__pyx_v_data) {
  struct __pyx_obj_2av_10bytesource_ByteSource *__pyx_v_source = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "av/codec/context.pyx":227
 * 
 *         def __set__(self, data):
 *             if not self.is_decoder:             # <<<<<<<<<<<<<<
 *                 raise ValueError("Can only set extradata for decoders.")
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_is_decoder); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 227, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = ((!__pyx_t_2) != 0);
  if (unlikely(__pyx_t_3)) {

    /* "av/codec/context.pyx":228
 *         def __set__(self, data):
 *             if not self.is_decoder:
 *                 raise ValueError("Can only set extradata for decoders.")             # <<<<<<<<<<<<<<
 * 
 *             if data is None:
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 228, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 228, __pyx_L1_error)

    /* "av/codec/context.pyx":227
 * 
 *         def __set__(self, data):
 *             if not self.is_decoder:             # <<<<<<<<<<<<<<
 *                 raise ValueError("Can only set extradata for decoders.")
 * 
 */
  }

  /* "av/codec/context.pyx":230
 *                 raise ValueError("Can only set extradata for decoders.")
 * 
 *             if data is None:             # <<<<<<<<<<<<<<
 *                 lib.av_freep(&self.ptr.extradata)
 *                 self.ptr.extradata_size = 0
 */
  __pyx_t_3 = (__pyx_v_data == Py_None);
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (__pyx_t_2) {

    /* "av/codec/context.pyx":231
 * 
 *             if data is None:
 *                 lib.av_freep(&self.ptr.extradata)             # <<<<<<<<<<<<<<
 *                 self.ptr.extradata_size = 0
 *             else:
 */
    av_freep((&__pyx_v_self->ptr->extradata));

    /* "av/codec/context.pyx":232
 *             if data is None:
 *                 lib.av_freep(&self.ptr.extradata)
 *                 self.ptr.extradata_size = 0             # <<<<<<<<<<<<<<
 *             else:
 *                 source = bytesource(data)
 */
    __pyx_v_self->ptr->extradata_size = 0;

    /* "av/codec/context.pyx":230
 *                 raise ValueError("Can only set extradata for decoders.")
 * 
 *             if data is None:             # <<<<<<<<<<<<<<
 *                 lib.av_freep(&self.ptr.extradata)
 *                 self.ptr.extradata_size = 0
 */
    goto __pyx_L4;
  }

  /* "av/codec/context.pyx":234
 *                 self.ptr.extradata_size = 0
 *             else:
 *                 source = bytesource(data)             # <<<<<<<<<<<<<<
 *                 self.ptr.extradata = <uint8_t*>lib.av_realloc(self.ptr.extradata, source.length + lib.AV_INPUT_BUFFER_PADDING_SIZE)
 *                 if not self.ptr.extradata:
 */
  /*else*/ {
    __pyx_t_1 = ((PyObject *)__pyx_f_2av_10bytesource_bytesource(__pyx_v_data, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 234, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_source = ((struct __pyx_obj_2av_10bytesource_ByteSource *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "av/codec/context.pyx":235
 *             else:
 *                 source = bytesource(data)
 *                 self.ptr.extradata = <uint8_t*>lib.av_realloc(self.ptr.extradata, source.length + lib.AV_INPUT_BUFFER_PADDING_SIZE)             # <<<<<<<<<<<<<<
 *                 if not self.ptr.extradata:
 *                     raise MemoryError("Cannot allocate extradata")
 */
    __pyx_v_self->ptr->extradata = ((uint8_t *)av_realloc(__pyx_v_self->ptr->extradata, (__pyx_v_source->length + AV_INPUT_BUFFER_PADDING_SIZE)));

    /* "av/codec/context.pyx":236
 *                 source = bytesource(data)
 *                 self.ptr.extradata = <uint8_t*>lib.av_realloc(self.ptr.extradata, source.length + lib.AV_INPUT_BUFFER_PADDING_SIZE)
 *                 if not self.ptr.extradata:             # <<<<<<<<<<<<<<
 *                     raise MemoryError("Cannot allocate extradata")
 *                 memcpy(self.ptr.extradata, source.ptr, source.length)
 */
    __pyx_t_2 = ((!(__pyx_v_self->ptr->extradata != 0)) != 0);
    if (unlikely(__pyx_t_2)) {

      /* "av/codec/context.pyx":237
 *                 self.ptr.extradata = <uint8_t*>lib.av_realloc(self.ptr.extradata, source.length + lib.AV_INPUT_BUFFER_PADDING_SIZE)
 *                 if not self.ptr.extradata:
 *                     raise MemoryError("Cannot allocate extradata")             # <<<<<<<<<<<<<<
 *                 memcpy(self.ptr.extradata, source.ptr, source.length)
 *                 self.ptr.extradata_size = source.length
 */
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_MemoryError, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 237, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 237, __pyx_L1_error)

      /* "av/codec/context.pyx":236
 *                 source = bytesource(data)
 *                 self.ptr.extradata = <uint8_t*>lib.av_realloc(self.ptr.extradata, source.length + lib.AV_INPUT_BUFFER_PADDING_SIZE)
 *                 if not self.ptr.extradata:             # <<<<<<<<<<<<<<
 *                     raise MemoryError("Cannot allocate extradata")
 *                 memcpy(self.ptr.extradata, source.ptr, source.length)
 */
    }

    /* "av/codec/context.pyx":238
 *                 if not self.ptr.extradata:
 *                     raise MemoryError("Cannot allocate extradata")
 *                 memcpy(self.ptr.extradata, source.ptr, source.length)             # <<<<<<<<<<<<<<
 *                 self.ptr.extradata_size = source.length
 *             self.extradata_set = True
 */
    (void)(memcpy(__pyx_v_self->ptr->extradata, __pyx_v_source->ptr, __pyx_v_source->length));

    /* "av/codec/context.pyx":239
 *                     raise MemoryError("Cannot allocate extradata")
 *                 memcpy(self.ptr.extradata, source.ptr, source.length)
 *                 self.ptr.extradata_size = source.length             # <<<<<<<<<<<<<<
 *             self.extradata_set = True
 * 
 */
    __pyx_t_4 = __pyx_v_source->length;
    __pyx_v_self->ptr->extradata_size = __pyx_t_4;
  }
  __pyx_L4:;

  /* "av/codec/context.pyx":240
 *                 memcpy(self.ptr.extradata, source.ptr, source.length)
 *                 self.ptr.extradata_size = source.length
 *             self.extradata_set = True             # <<<<<<<<<<<<<<
 * 
 *     property extradata_size:
 */
  __pyx_v_self->extradata_set = 1;

  /* "av/codec/context.pyx":226
 *                 return None
 * 
 *         def __set__(self, data):             # <<<<<<<<<<<<<<
 *             if not self.is_decoder:
 *                 raise ValueError("Can only set extradata for decoders.")
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("av.codec.context.CodecContext.extradata.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_source);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pyx":243
 * 
 *     property extradata_size:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.ptr.extradata_size
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_14extradata_size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_14extradata_size_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2av_5codec_7context_12CodecContext_14extradata_size___get__(((struct __pyx_obj_2av_5codec_7context_CodecContext *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_14extradata_size___get__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "av/codec/context.pyx":244
 *     property extradata_size:
 *         def __get__(self):
 *             return self.ptr.extradata_size             # <<<<<<<<<<<<<<
 * 
 *     property is_open:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->ptr->extradata_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "av/codec/context.pyx":243
 * 
 *     property extradata_size:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.ptr.extradata_size
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("av.codec.context.CodecContext.extradata_size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pyx":247
 * 
 *     property is_open:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return lib.avcodec_is_open(self.ptr)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_7is_open_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_7is_open_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2av_5codec_7context_12CodecContext_7is_open___get__(((struct __pyx_obj_2av_5codec_7context_CodecContext *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_7is_open___get__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "av/codec/context.pyx":248
 *     property is_open:
 *         def __get__(self):
 *             return lib.avcodec_is_open(self.ptr)             # <<<<<<<<<<<<<<
 * 
 *     property is_encoder:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(avcodec_is_open(__pyx_v_self->ptr)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 248, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "av/codec/context.pyx":247
 * 
 *     property is_open:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return lib.avcodec_is_open(self.ptr)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("av.codec.context.CodecContext.is_open.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pyx":251
 * 
 *     property is_encoder:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return lib.av_codec_is_encoder(self.ptr.codec)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_10is_encoder_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_10is_encoder_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2av_5codec_7context_12CodecContext_10is_encoder___get__(((struct __pyx_obj_2av_5codec_7context_CodecContext *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_10is_encoder___get__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "av/codec/context.pyx":252
 *     property is_encoder:
 *         def __get__(self):
 *             return lib.av_codec_is_encoder(self.ptr.codec)             # <<<<<<<<<<<<<<
 * 
 *     property is_decoder:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(av_codec_is_encoder(__pyx_v_self->ptr->codec)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "av/codec/context.pyx":251
 * 
 *     property is_encoder:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return lib.av_codec_is_encoder(self.ptr.codec)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("av.codec.context.CodecContext.is_encoder.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pyx":255
 * 
 *     property is_decoder:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return lib.av_codec_is_decoder(self.ptr.codec)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_10is_decoder_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_10is_decoder_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2av_5codec_7context_12CodecContext_10is_decoder___get__(((struct __pyx_obj_2av_5codec_7context_CodecContext *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_10is_decoder___get__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "av/codec/context.pyx":256
 *     property is_decoder:
 *         def __get__(self):
 *             return lib.av_codec_is_decoder(self.ptr.codec)             # <<<<<<<<<<<<<<
 * 
 *     cpdef open(self, bint strict=True):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(av_codec_is_decoder(__pyx_v_self->ptr->codec)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "av/codec/context.pyx":255
 * 
 *     property is_decoder:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return lib.av_codec_is_decoder(self.ptr.codec)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("av.codec.context.CodecContext.is_decoder.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pyx":258
 *             return lib.av_codec_is_decoder(self.ptr.codec)
 * 
 *     cpdef open(self, bint strict=True):             # <<<<<<<<<<<<<<
 * 
 *         if lib.avcodec_is_open(self.ptr):
 */

static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_13open(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_2av_5codec_7context_12CodecContext_open(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, int __pyx_skip_dispatch, struct __pyx_opt_args_2av_5codec_7context_12CodecContext_open *__pyx_optional_args) {
  int __pyx_v_strict = ((int)1);
  struct __pyx_obj_2av_10dictionary__Dictionary *__pyx_v_options = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("open", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_strict = __pyx_optional_args->strict;
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_open); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 258, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_2av_5codec_7context_12CodecContext_13open)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_strict); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 258, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 258, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "av/codec/context.pyx":260
 *     cpdef open(self, bint strict=True):
 * 
 *         if lib.avcodec_is_open(self.ptr):             # <<<<<<<<<<<<<<
 *             if strict:
 *                 raise ValueError('CodecContext is already open.')
 */
  __pyx_t_6 = (avcodec_is_open(__pyx_v_self->ptr) != 0);
  if (__pyx_t_6) {

    /* "av/codec/context.pyx":261
 * 
 *         if lib.avcodec_is_open(self.ptr):
 *             if strict:             # <<<<<<<<<<<<<<
 *                 raise ValueError('CodecContext is already open.')
 *             return
 */
    __pyx_t_6 = (__pyx_v_strict != 0);
    if (unlikely(__pyx_t_6)) {

      /* "av/codec/context.pyx":262
 *         if lib.avcodec_is_open(self.ptr):
 *             if strict:
 *                 raise ValueError('CodecContext is already open.')             # <<<<<<<<<<<<<<
 *             return
 * 
 */
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 262, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 262, __pyx_L1_error)

      /* "av/codec/context.pyx":261
 * 
 *         if lib.avcodec_is_open(self.ptr):
 *             if strict:             # <<<<<<<<<<<<<<
 *                 raise ValueError('CodecContext is already open.')
 *             return
 */
    }

    /* "av/codec/context.pyx":263
 *             if strict:
 *                 raise ValueError('CodecContext is already open.')
 *             return             # <<<<<<<<<<<<<<
 * 
 *         # TODO: Do this better.
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "av/codec/context.pyx":260
 *     cpdef open(self, bint strict=True):
 * 
 *         if lib.avcodec_is_open(self.ptr):             # <<<<<<<<<<<<<<
 *             if strict:
 *                 raise ValueError('CodecContext is already open.')
 */
  }

  /* "av/codec/context.pyx":266
 * 
 *         # TODO: Do this better.
 *         cdef _Dictionary options = Dictionary()             # <<<<<<<<<<<<<<
 *         options.update(self.options or {})
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Dictionary); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_2av_10dictionary__Dictionary))))) __PYX_ERR(0, 266, __pyx_L1_error)
  __pyx_v_options = ((struct __pyx_obj_2av_10dictionary__Dictionary *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "av/codec/context.pyx":267
 *         # TODO: Do this better.
 *         cdef _Dictionary options = Dictionary()
 *         options.update(self.options or {})             # <<<<<<<<<<<<<<
 * 
 *         # Assert we have a time_base for encoders.
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_options), __pyx_n_s_update); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 267, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_self->options); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 267, __pyx_L1_error)
  if (!__pyx_t_6) {
  } else {
    __Pyx_INCREF(__pyx_v_self->options);
    __pyx_t_4 = __pyx_v_self->options;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 267, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = __pyx_t_3;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_L5_bool_binop_done:;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 267, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "av/codec/context.pyx":270
 * 
 *         # Assert we have a time_base for encoders.
 *         if not self.ptr.time_base.num and self.is_encoder:             # <<<<<<<<<<<<<<
 *             self._set_default_time_base()
 * 
 */
  __pyx_t_7 = ((!(__pyx_v_self->ptr->time_base.num != 0)) != 0);
  if (__pyx_t_7) {
  } else {
    __pyx_t_6 = __pyx_t_7;
    goto __pyx_L8_bool_binop_done;
  }
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_is_encoder); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 270, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __pyx_t_7;
  __pyx_L8_bool_binop_done:;
  if (__pyx_t_6) {

    /* "av/codec/context.pyx":271
 *         # Assert we have a time_base for encoders.
 *         if not self.ptr.time_base.num and self.is_encoder:
 *             self._set_default_time_base()             # <<<<<<<<<<<<<<
 * 
 *         err_check(lib.avcodec_open2(self.ptr, self.codec.ptr, &options.ptr))
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_2av_5codec_7context_CodecContext *)__pyx_v_self->__pyx_vtab)->_set_default_time_base(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 271, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "av/codec/context.pyx":270
 * 
 *         # Assert we have a time_base for encoders.
 *         if not self.ptr.time_base.num and self.is_encoder:             # <<<<<<<<<<<<<<
 *             self._set_default_time_base()
 * 
 */
  }

  /* "av/codec/context.pyx":273
 *             self._set_default_time_base()
 * 
 *         err_check(lib.avcodec_open2(self.ptr, self.codec.ptr, &options.ptr))             # <<<<<<<<<<<<<<
 * 
 *         self.options = dict(options)
 */
  __pyx_t_8 = __pyx_f_2av_5error_err_check(avcodec_open2(__pyx_v_self->ptr, __pyx_v_self->codec->ptr, (&__pyx_v_options->ptr)), 0, NULL); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 273, __pyx_L1_error)

  /* "av/codec/context.pyx":275
 *         err_check(lib.avcodec_open2(self.ptr, self.codec.ptr, &options.ptr))
 * 
 *         self.options = dict(options)             # <<<<<<<<<<<<<<
 * 
 *     cdef _set_default_time_base(self):
 */
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), ((PyObject *)__pyx_v_options)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->options);
  __Pyx_DECREF(__pyx_v_self->options);
  __pyx_v_self->options = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "av/codec/context.pyx":258
 *             return lib.av_codec_is_decoder(self.ptr.codec)
 * 
 *     cpdef open(self, bint strict=True):             # <<<<<<<<<<<<<<
 * 
 *         if lib.avcodec_is_open(self.ptr):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("av.codec.context.CodecContext.open", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_options);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_13open(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_2av_5codec_7context_12CodecContext_12open[] = "CodecContext.open(bool strict=True)";
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_13open(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_strict;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("open (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_strict,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_strict);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "open") < 0)) __PYX_ERR(0, 258, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_strict = __Pyx_PyObject_IsTrue(values[0]); if (unlikely((__pyx_v_strict == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 258, __pyx_L3_error)
    } else {
      __pyx_v_strict = ((int)1);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("open", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 258, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("av.codec.context.CodecContext.open", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_2av_5codec_7context_12CodecContext_12open(((struct __pyx_obj_2av_5codec_7context_CodecContext *)__pyx_v_self), __pyx_v_strict);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_12open(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, int __pyx_v_strict) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_2av_5codec_7context_12CodecContext_open __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("open", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.strict = __pyx_v_strict;
  __pyx_t_1 = __pyx_vtabptr_2av_5codec_7context_CodecContext->open(__pyx_v_self, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("av.codec.context.CodecContext.open", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pyx":277
 *         self.options = dict(options)
 * 
 *     cdef _set_default_time_base(self):             # <<<<<<<<<<<<<<
 *         self.ptr.time_base.num = 1
 *         self.ptr.time_base.den = lib.AV_TIME_BASE
 */

static PyObject *__pyx_f_2av_5codec_7context_12CodecContext__set_default_time_base(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_set_default_time_base", 0);

  /* "av/codec/context.pyx":278
 * 
 *     cdef _set_default_time_base(self):
 *         self.ptr.time_base.num = 1             # <<<<<<<<<<<<<<
 *         self.ptr.time_base.den = lib.AV_TIME_BASE
 * 
 */
  __pyx_v_self->ptr->time_base.num = 1;

  /* "av/codec/context.pyx":279
 *     cdef _set_default_time_base(self):
 *         self.ptr.time_base.num = 1
 *         self.ptr.time_base.den = lib.AV_TIME_BASE             # <<<<<<<<<<<<<<
 * 
 *     cpdef close(self, bint strict=True):
 */
  __pyx_v_self->ptr->time_base.den = AV_TIME_BASE;

  /* "av/codec/context.pyx":277
 *         self.options = dict(options)
 * 
 *     cdef _set_default_time_base(self):             # <<<<<<<<<<<<<<
 *         self.ptr.time_base.num = 1
 *         self.ptr.time_base.den = lib.AV_TIME_BASE
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pyx":281
 *         self.ptr.time_base.den = lib.AV_TIME_BASE
 * 
 *     cpdef close(self, bint strict=True):             # <<<<<<<<<<<<<<
 *         if not lib.avcodec_is_open(self.ptr):
 *             if strict:
 */

static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_15close(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_2av_5codec_7context_12CodecContext_close(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, int __pyx_skip_dispatch, struct __pyx_opt_args_2av_5codec_7context_12CodecContext_close *__pyx_optional_args) {
  int __pyx_v_strict = ((int)1);
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("close", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_strict = __pyx_optional_args->strict;
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_close); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 281, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_2av_5codec_7context_12CodecContext_15close)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_strict); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 281, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 281, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "av/codec/context.pyx":282
 * 
 *     cpdef close(self, bint strict=True):
 *         if not lib.avcodec_is_open(self.ptr):             # <<<<<<<<<<<<<<
 *             if strict:
 *                 raise ValueError('CodecContext is already closed.')
 */
  __pyx_t_6 = ((!(avcodec_is_open(__pyx_v_self->ptr) != 0)) != 0);
  if (__pyx_t_6) {

    /* "av/codec/context.pyx":283
 *     cpdef close(self, bint strict=True):
 *         if not lib.avcodec_is_open(self.ptr):
 *             if strict:             # <<<<<<<<<<<<<<
 *                 raise ValueError('CodecContext is already closed.')
 *             return
 */
    __pyx_t_6 = (__pyx_v_strict != 0);
    if (unlikely(__pyx_t_6)) {

      /* "av/codec/context.pyx":284
 *         if not lib.avcodec_is_open(self.ptr):
 *             if strict:
 *                 raise ValueError('CodecContext is already closed.')             # <<<<<<<<<<<<<<
 *             return
 *         err_check(lib.avcodec_close(self.ptr))
 */
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 284, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 284, __pyx_L1_error)

      /* "av/codec/context.pyx":283
 *     cpdef close(self, bint strict=True):
 *         if not lib.avcodec_is_open(self.ptr):
 *             if strict:             # <<<<<<<<<<<<<<
 *                 raise ValueError('CodecContext is already closed.')
 *             return
 */
    }

    /* "av/codec/context.pyx":285
 *             if strict:
 *                 raise ValueError('CodecContext is already closed.')
 *             return             # <<<<<<<<<<<<<<
 *         err_check(lib.avcodec_close(self.ptr))
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "av/codec/context.pyx":282
 * 
 *     cpdef close(self, bint strict=True):
 *         if not lib.avcodec_is_open(self.ptr):             # <<<<<<<<<<<<<<
 *             if strict:
 *                 raise ValueError('CodecContext is already closed.')
 */
  }

  /* "av/codec/context.pyx":286
 *                 raise ValueError('CodecContext is already closed.')
 *             return
 *         err_check(lib.avcodec_close(self.ptr))             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __pyx_t_7 = __pyx_f_2av_5error_err_check(avcodec_close(__pyx_v_self->ptr), 0, NULL); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 286, __pyx_L1_error)

  /* "av/codec/context.pyx":281
 *         self.ptr.time_base.den = lib.AV_TIME_BASE
 * 
 *     cpdef close(self, bint strict=True):             # <<<<<<<<<<<<<<
 *         if not lib.avcodec_is_open(self.ptr):
 *             if strict:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("av.codec.context.CodecContext.close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_15close(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_2av_5codec_7context_12CodecContext_14close[] = "CodecContext.close(bool strict=True)";
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_15close(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_strict;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("close (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_strict,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_strict);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "close") < 0)) __PYX_ERR(0, 281, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_strict = __Pyx_PyObject_IsTrue(values[0]); if (unlikely((__pyx_v_strict == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 281, __pyx_L3_error)
    } else {
      __pyx_v_strict = ((int)1);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("close", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 281, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("av.codec.context.CodecContext.close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_2av_5codec_7context_12CodecContext_14close(((struct __pyx_obj_2av_5codec_7context_CodecContext *)__pyx_v_self), __pyx_v_strict);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_14close(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, int __pyx_v_strict) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_2av_5codec_7context_12CodecContext_close __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("close", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.strict = __pyx_v_strict;
  __pyx_t_1 = __pyx_vtabptr_2av_5codec_7context_CodecContext->close(__pyx_v_self, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("av.codec.context.CodecContext.close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pyx":288
 *         err_check(lib.avcodec_close(self.ptr))
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.ptr and self.extradata_set:
 *             lib.av_freep(&self.ptr.extradata)
 */

/* Python wrapper */
static void __pyx_pw_2av_5codec_7context_12CodecContext_17__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_2av_5codec_7context_12CodecContext_17__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_2av_5codec_7context_12CodecContext_16__dealloc__(((struct __pyx_obj_2av_5codec_7context_CodecContext *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_2av_5codec_7context_12CodecContext_16__dealloc__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "av/codec/context.pyx":289
 * 
 *     def __dealloc__(self):
 *         if self.ptr and self.extradata_set:             # <<<<<<<<<<<<<<
 *             lib.av_freep(&self.ptr.extradata)
 *         if self.ptr:
 */
  __pyx_t_2 = (__pyx_v_self->ptr != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->extradata_set != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "av/codec/context.pyx":290
 *     def __dealloc__(self):
 *         if self.ptr and self.extradata_set:
 *             lib.av_freep(&self.ptr.extradata)             # <<<<<<<<<<<<<<
 *         if self.ptr:
 *             lib.avcodec_close(self.ptr)
 */
    av_freep((&__pyx_v_self->ptr->extradata));

    /* "av/codec/context.pyx":289
 * 
 *     def __dealloc__(self):
 *         if self.ptr and self.extradata_set:             # <<<<<<<<<<<<<<
 *             lib.av_freep(&self.ptr.extradata)
 *         if self.ptr:
 */
  }

  /* "av/codec/context.pyx":291
 *         if self.ptr and self.extradata_set:
 *             lib.av_freep(&self.ptr.extradata)
 *         if self.ptr:             # <<<<<<<<<<<<<<
 *             lib.avcodec_close(self.ptr)
 *             lib.avcodec_free_context(&self.ptr)
 */
  __pyx_t_1 = (__pyx_v_self->ptr != 0);
  if (__pyx_t_1) {

    /* "av/codec/context.pyx":292
 *             lib.av_freep(&self.ptr.extradata)
 *         if self.ptr:
 *             lib.avcodec_close(self.ptr)             # <<<<<<<<<<<<<<
 *             lib.avcodec_free_context(&self.ptr)
 *         if self.parser:
 */
    (void)(avcodec_close(__pyx_v_self->ptr));

    /* "av/codec/context.pyx":293
 *         if self.ptr:
 *             lib.avcodec_close(self.ptr)
 *             lib.avcodec_free_context(&self.ptr)             # <<<<<<<<<<<<<<
 *         if self.parser:
 *             lib.av_parser_close(self.parser)
 */
    avcodec_free_context((&__pyx_v_self->ptr));

    /* "av/codec/context.pyx":291
 *         if self.ptr and self.extradata_set:
 *             lib.av_freep(&self.ptr.extradata)
 *         if self.ptr:             # <<<<<<<<<<<<<<
 *             lib.avcodec_close(self.ptr)
 *             lib.avcodec_free_context(&self.ptr)
 */
  }

  /* "av/codec/context.pyx":294
 *             lib.avcodec_close(self.ptr)
 *             lib.avcodec_free_context(&self.ptr)
 *         if self.parser:             # <<<<<<<<<<<<<<
 *             lib.av_parser_close(self.parser)
 * 
 */
  __pyx_t_1 = (__pyx_v_self->parser != 0);
  if (__pyx_t_1) {

    /* "av/codec/context.pyx":295
 *             lib.avcodec_free_context(&self.ptr)
 *         if self.parser:
 *             lib.av_parser_close(self.parser)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    av_parser_close(__pyx_v_self->parser);

    /* "av/codec/context.pyx":294
 *             lib.avcodec_close(self.ptr)
 *             lib.avcodec_free_context(&self.ptr)
 *         if self.parser:             # <<<<<<<<<<<<<<
 *             lib.av_parser_close(self.parser)
 * 
 */
  }

  /* "av/codec/context.pyx":288
 *         err_check(lib.avcodec_close(self.ptr))
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.ptr and self.extradata_set:
 *             lib.av_freep(&self.ptr.extradata)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "av/codec/context.pyx":297
 *             lib.av_parser_close(self.parser)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return '<av.%s %s/%s at 0x%x>' % (
 *             self.__class__.__name__,
 */

/* Python wrapper */
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_19__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_19__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2av_5codec_7context_12CodecContext_18__repr__(((struct __pyx_obj_2av_5codec_7context_CodecContext *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_18__repr__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "av/codec/context.pyx":298
 * 
 *     def __repr__(self):
 *         return '<av.%s %s/%s at 0x%x>' % (             # <<<<<<<<<<<<<<
 *             self.__class__.__name__,
 *             self.type or '<notype>',
 */
  __Pyx_XDECREF(__pyx_r);

  /* "av/codec/context.pyx":299
 *     def __repr__(self):
 *         return '<av.%s %s/%s at 0x%x>' % (
 *             self.__class__.__name__,             # <<<<<<<<<<<<<<
 *             self.type or '<notype>',
 *             self.name or '<nocodec>',
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "av/codec/context.pyx":300
 *         return '<av.%s %s/%s at 0x%x>' % (
 *             self.__class__.__name__,
 *             self.type or '<notype>',             # <<<<<<<<<<<<<<
 *             self.name or '<nocodec>',
 *             id(self),
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_type); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 300, __pyx_L1_error)
  if (!__pyx_t_4) {
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_1 = __pyx_t_3;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L3_bool_binop_done;
  }
  __Pyx_INCREF(__pyx_kp_s_notype);
  __pyx_t_1 = __pyx_kp_s_notype;
  __pyx_L3_bool_binop_done:;

  /* "av/codec/context.pyx":301
 *             self.__class__.__name__,
 *             self.type or '<notype>',
 *             self.name or '<nocodec>',             # <<<<<<<<<<<<<<
 *             id(self),
 *         )
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_name_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 301, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 301, __pyx_L1_error)
  if (!__pyx_t_4) {
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_5);
    __pyx_t_3 = __pyx_t_5;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L5_bool_binop_done;
  }
  __Pyx_INCREF(__pyx_kp_s_nocodec);
  __pyx_t_3 = __pyx_kp_s_nocodec;
  __pyx_L5_bool_binop_done:;

  /* "av/codec/context.pyx":302
 *             self.type or '<notype>',
 *             self.name or '<nocodec>',
 *             id(self),             # <<<<<<<<<<<<<<
 *         )
 * 
 */
  __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_id, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "av/codec/context.pyx":299
 *     def __repr__(self):
 *         return '<av.%s %s/%s at 0x%x>' % (
 *             self.__class__.__name__,             # <<<<<<<<<<<<<<
 *             self.type or '<notype>',
 *             self.name or '<nocodec>',
 */
  __pyx_t_6 = PyTuple_New(4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_6, 3, __pyx_t_5);
  __pyx_t_2 = 0;
  __pyx_t_1 = 0;
  __pyx_t_3 = 0;
  __pyx_t_5 = 0;

  /* "av/codec/context.pyx":298
 * 
 *     def __repr__(self):
 *         return '<av.%s %s/%s at 0x%x>' % (             # <<<<<<<<<<<<<<
 *             self.__class__.__name__,
 *             self.type or '<notype>',
 */
  __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_av_s_s_s_at_0x_x, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 298, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "av/codec/context.pyx":297
 *             lib.av_parser_close(self.parser)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return '<av.%s %s/%s at 0x%x>' % (
 *             self.__class__.__name__,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("av.codec.context.CodecContext.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pyx":305
 *         )
 * 
 *     def parse(self, raw_input=None):             # <<<<<<<<<<<<<<
 *         """Split up a byte stream into list of :class:`.Packet`.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_21parse(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_2av_5codec_7context_12CodecContext_20parse[] = "CodecContext.parse(raw_input=None)\nSplit up a byte stream into list of :class:`.Packet`.\n\n        This is only effectively splitting up a byte stream, and does no\n        actual interpretation of the data.\n\n        It will return all packets that are fully contained within the given\n        input, and will buffer partial packets until they are complete.\n\n        :param ByteSource raw_input: A chunk of a byte-stream to process.\n            Anything that can be turned into a :class:`.ByteSource` is fine.\n            ``None`` or empty inputs will flush the parser's buffers.\n\n        :return: ``list`` of :class:`.Packet` newly available.\n\n        ";
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_21parse(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_raw_input = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("parse (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_raw_input,0};
    PyObject* values[1] = {0};
    values[0] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_raw_input);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "parse") < 0)) __PYX_ERR(0, 305, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_raw_input = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("parse", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 305, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("av.codec.context.CodecContext.parse", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_2av_5codec_7context_12CodecContext_20parse(((struct __pyx_obj_2av_5codec_7context_CodecContext *)__pyx_v_self), __pyx_v_raw_input);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_20parse(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, PyObject *__pyx_v_raw_input) {
  struct __pyx_obj_2av_10bytesource_ByteSource *__pyx_v_source = 0;
  unsigned char *__pyx_v_in_data;
  int __pyx_v_in_size;
  unsigned char *__pyx_v_out_data;
  int __pyx_v_out_size;
  int __pyx_v_consumed;
  struct __pyx_obj_2av_6packet_Packet *__pyx_v_packet = 0;
  PyObject *__pyx_v_packets = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  struct __pyx_opt_args_2av_10bytesource_bytesource __pyx_t_4;
  unsigned char *__pyx_t_5;
  size_t __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parse", 0);

  /* "av/codec/context.pyx":322
 *         """
 * 
 *         if not self.parser:             # <<<<<<<<<<<<<<
 *             self.parser = lib.av_parser_init(self.codec.ptr.id)
 *             if not self.parser:
 */
  __pyx_t_1 = ((!(__pyx_v_self->parser != 0)) != 0);
  if (__pyx_t_1) {

    /* "av/codec/context.pyx":323
 * 
 *         if not self.parser:
 *             self.parser = lib.av_parser_init(self.codec.ptr.id)             # <<<<<<<<<<<<<<
 *             if not self.parser:
 *                 raise ValueError('No parser for %s' % self.codec.name)
 */
    __pyx_v_self->parser = av_parser_init(__pyx_v_self->codec->ptr->id);

    /* "av/codec/context.pyx":324
 *         if not self.parser:
 *             self.parser = lib.av_parser_init(self.codec.ptr.id)
 *             if not self.parser:             # <<<<<<<<<<<<<<
 *                 raise ValueError('No parser for %s' % self.codec.name)
 * 
 */
    __pyx_t_1 = ((!(__pyx_v_self->parser != 0)) != 0);
    if (unlikely(__pyx_t_1)) {

      /* "av/codec/context.pyx":325
 *             self.parser = lib.av_parser_init(self.codec.ptr.id)
 *             if not self.parser:
 *                 raise ValueError('No parser for %s' % self.codec.name)             # <<<<<<<<<<<<<<
 * 
 *         cdef ByteSource source = bytesource(raw_input, allow_none=True)
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->codec), __pyx_n_s_name_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 325, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = __Pyx_PyString_FormatSafe(__pyx_kp_s_No_parser_for_s, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 325, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 325, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 325, __pyx_L1_error)

      /* "av/codec/context.pyx":324
 *         if not self.parser:
 *             self.parser = lib.av_parser_init(self.codec.ptr.id)
 *             if not self.parser:             # <<<<<<<<<<<<<<
 *                 raise ValueError('No parser for %s' % self.codec.name)
 * 
 */
    }

    /* "av/codec/context.pyx":322
 *         """
 * 
 *         if not self.parser:             # <<<<<<<<<<<<<<
 *             self.parser = lib.av_parser_init(self.codec.ptr.id)
 *             if not self.parser:
 */
  }

  /* "av/codec/context.pyx":327
 *                 raise ValueError('No parser for %s' % self.codec.name)
 * 
 *         cdef ByteSource source = bytesource(raw_input, allow_none=True)             # <<<<<<<<<<<<<<
 * 
 *         cdef unsigned char *in_data = source.ptr if source is not None else NULL
 */
  __pyx_t_4.__pyx_n = 1;
  __pyx_t_4.allow_none = 1;
  __pyx_t_2 = ((PyObject *)__pyx_f_2av_10bytesource_bytesource(__pyx_v_raw_input, &__pyx_t_4)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 327, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_source = ((struct __pyx_obj_2av_10bytesource_ByteSource *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "av/codec/context.pyx":329
 *         cdef ByteSource source = bytesource(raw_input, allow_none=True)
 * 
 *         cdef unsigned char *in_data = source.ptr if source is not None else NULL             # <<<<<<<<<<<<<<
 *         cdef int in_size = source.length if source is not None else 0
 * 
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_source) != Py_None);
  if ((__pyx_t_1 != 0)) {
    __pyx_t_5 = __pyx_v_source->ptr;
  } else {
    __pyx_t_5 = NULL;
  }
  __pyx_v_in_data = __pyx_t_5;

  /* "av/codec/context.pyx":330
 * 
 *         cdef unsigned char *in_data = source.ptr if source is not None else NULL
 *         cdef int in_size = source.length if source is not None else 0             # <<<<<<<<<<<<<<
 * 
 *         cdef unsigned char *out_data
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_source) != Py_None);
  if ((__pyx_t_1 != 0)) {
    __pyx_t_6 = __pyx_v_source->length;
  } else {
    __pyx_t_6 = 0;
  }
  __pyx_v_in_size = __pyx_t_6;

  /* "av/codec/context.pyx":335
 *         cdef int out_size
 *         cdef int consumed
 *         cdef Packet packet = None             # <<<<<<<<<<<<<<
 * 
 *         packets = []
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_packet = ((struct __pyx_obj_2av_6packet_Packet *)Py_None);

  /* "av/codec/context.pyx":337
 *         cdef Packet packet = None
 * 
 *         packets = []             # <<<<<<<<<<<<<<
 * 
 *         while True:
 */
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 337, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_packets = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "av/codec/context.pyx":339
 *         packets = []
 * 
 *         while True:             # <<<<<<<<<<<<<<
 * 
 *             with nogil:
 */
  while (1) {

    /* "av/codec/context.pyx":341
 *         while True:
 * 
 *             with nogil:             # <<<<<<<<<<<<<<
 *                 consumed = lib.av_parser_parse2(
 *                     self.parser,
 */
    {
        #ifdef WITH_THREAD
        PyThreadState *_save;
        Py_UNBLOCK_THREADS
        __Pyx_FastGIL_Remember();
        #endif
        /*try:*/ {

          /* "av/codec/context.pyx":342
 * 
 *             with nogil:
 *                 consumed = lib.av_parser_parse2(             # <<<<<<<<<<<<<<
 *                     self.parser,
 *                     self.ptr,
 */
          __pyx_v_consumed = av_parser_parse2(__pyx_v_self->parser, __pyx_v_self->ptr, (&__pyx_v_out_data), (&__pyx_v_out_size), __pyx_v_in_data, __pyx_v_in_size, AV_NOPTS_VALUE, AV_NOPTS_VALUE, 0);
        }

        /* "av/codec/context.pyx":341
 *         while True:
 * 
 *             with nogil:             # <<<<<<<<<<<<<<
 *                 consumed = lib.av_parser_parse2(
 *                     self.parser,
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_FastGIL_Forget();
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L11;
          }
          __pyx_L11:;
        }
    }

    /* "av/codec/context.pyx":350
 *                     0
 *                 )
 *             err_check(consumed)             # <<<<<<<<<<<<<<
 * 
 *             if out_size:
 */
    __pyx_t_7 = __pyx_f_2av_5error_err_check(__pyx_v_consumed, 0, NULL); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 350, __pyx_L1_error)

    /* "av/codec/context.pyx":352
 *             err_check(consumed)
 * 
 *             if out_size:             # <<<<<<<<<<<<<<
 * 
 *                 # We copy the data immediately, as we have yet to figure out
 */
    __pyx_t_1 = (__pyx_v_out_size != 0);
    if (__pyx_t_1) {

      /* "av/codec/context.pyx":366
 *                 # ... but this results in corruption.
 * 
 *                 packet = Packet(out_size)             # <<<<<<<<<<<<<<
 *                 memcpy(packet.ptr.data, out_data, out_size)
 * 
 */
      __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_out_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 366, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_2av_6packet_Packet), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 366, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF_SET(__pyx_v_packet, ((struct __pyx_obj_2av_6packet_Packet *)__pyx_t_3));
      __pyx_t_3 = 0;

      /* "av/codec/context.pyx":367
 * 
 *                 packet = Packet(out_size)
 *                 memcpy(packet.ptr.data, out_data, out_size)             # <<<<<<<<<<<<<<
 * 
 *                 packets.append(packet)
 */
      (void)(memcpy(__pyx_v_packet->ptr->data, __pyx_v_out_data, __pyx_v_out_size));

      /* "av/codec/context.pyx":369
 *                 memcpy(packet.ptr.data, out_data, out_size)
 * 
 *                 packets.append(packet)             # <<<<<<<<<<<<<<
 * 
 *             if not in_size:
 */
      __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_packets, ((PyObject *)__pyx_v_packet)); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 369, __pyx_L1_error)

      /* "av/codec/context.pyx":352
 *             err_check(consumed)
 * 
 *             if out_size:             # <<<<<<<<<<<<<<
 * 
 *                 # We copy the data immediately, as we have yet to figure out
 */
    }

    /* "av/codec/context.pyx":371
 *                 packets.append(packet)
 * 
 *             if not in_size:             # <<<<<<<<<<<<<<
 *                 # This was a flush. Only one packet should ever be returned.
 *                 break
 */
    __pyx_t_1 = ((!(__pyx_v_in_size != 0)) != 0);
    if (__pyx_t_1) {

      /* "av/codec/context.pyx":373
 *             if not in_size:
 *                 # This was a flush. Only one packet should ever be returned.
 *                 break             # <<<<<<<<<<<<<<
 * 
 *             in_data += consumed
 */
      goto __pyx_L6_break;

      /* "av/codec/context.pyx":371
 *                 packets.append(packet)
 * 
 *             if not in_size:             # <<<<<<<<<<<<<<
 *                 # This was a flush. Only one packet should ever be returned.
 *                 break
 */
    }

    /* "av/codec/context.pyx":375
 *                 break
 * 
 *             in_data += consumed             # <<<<<<<<<<<<<<
 *             in_size -= consumed
 * 
 */
    __pyx_v_in_data = (__pyx_v_in_data + __pyx_v_consumed);

    /* "av/codec/context.pyx":376
 * 
 *             in_data += consumed
 *             in_size -= consumed             # <<<<<<<<<<<<<<
 * 
 *             if not in_size:
 */
    __pyx_v_in_size = (__pyx_v_in_size - __pyx_v_consumed);

    /* "av/codec/context.pyx":378
 *             in_size -= consumed
 * 
 *             if not in_size:             # <<<<<<<<<<<<<<
 *                 # Aaaand now we're done.
 *                 break
 */
    __pyx_t_1 = ((!(__pyx_v_in_size != 0)) != 0);
    if (__pyx_t_1) {

      /* "av/codec/context.pyx":380
 *             if not in_size:
 *                 # Aaaand now we're done.
 *                 break             # <<<<<<<<<<<<<<
 * 
 *         return packets
 */
      goto __pyx_L6_break;

      /* "av/codec/context.pyx":378
 *             in_size -= consumed
 * 
 *             if not in_size:             # <<<<<<<<<<<<<<
 *                 # Aaaand now we're done.
 *                 break
 */
    }
  }
  __pyx_L6_break:;

  /* "av/codec/context.pyx":382
 *                 break
 * 
 *         return packets             # <<<<<<<<<<<<<<
 * 
 *     cdef _send_frame_and_recv(self, Frame frame):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_packets);
  __pyx_r = __pyx_v_packets;
  goto __pyx_L0;

  /* "av/codec/context.pyx":305
 *         )
 * 
 *     def parse(self, raw_input=None):             # <<<<<<<<<<<<<<
 *         """Split up a byte stream into list of :class:`.Packet`.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("av.codec.context.CodecContext.parse", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_source);
  __Pyx_XDECREF((PyObject *)__pyx_v_packet);
  __Pyx_XDECREF(__pyx_v_packets);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pyx":384
 *         return packets
 * 
 *     cdef _send_frame_and_recv(self, Frame frame):             # <<<<<<<<<<<<<<
 * 
 *         cdef Packet packet
 */

static PyObject *__pyx_f_2av_5codec_7context_12CodecContext__send_frame_and_recv(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, struct __pyx_obj_2av_5frame_Frame *__pyx_v_frame) {
  struct __pyx_obj_2av_6packet_Packet *__pyx_v_packet = 0;
  int __pyx_v_res;
  PyObject *__pyx_v_out = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct AVFrame *__pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_send_frame_and_recv", 0);

  /* "av/codec/context.pyx":389
 * 
 *         cdef int res
 *         with nogil:             # <<<<<<<<<<<<<<
 *             res = lib.avcodec_send_frame(self.ptr, frame.ptr if frame is not None else NULL)
 *         err_check(res)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "av/codec/context.pyx":390
 *         cdef int res
 *         with nogil:
 *             res = lib.avcodec_send_frame(self.ptr, frame.ptr if frame is not None else NULL)             # <<<<<<<<<<<<<<
 *         err_check(res)
 * 
 */
        __pyx_t_2 = (((PyObject *)__pyx_v_frame) != Py_None);
        if ((__pyx_t_2 != 0)) {
          __pyx_t_1 = __pyx_v_frame->ptr;
        } else {
          __pyx_t_1 = NULL;
        }
        __pyx_v_res = avcodec_send_frame(__pyx_v_self->ptr, __pyx_t_1);
      }

      /* "av/codec/context.pyx":389
 * 
 *         cdef int res
 *         with nogil:             # <<<<<<<<<<<<<<
 *             res = lib.avcodec_send_frame(self.ptr, frame.ptr if frame is not None else NULL)
 *         err_check(res)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "av/codec/context.pyx":391
 *         with nogil:
 *             res = lib.avcodec_send_frame(self.ptr, frame.ptr if frame is not None else NULL)
 *         err_check(res)             # <<<<<<<<<<<<<<
 * 
 *         out = []
 */
  __pyx_t_3 = __pyx_f_2av_5error_err_check(__pyx_v_res, 0, NULL); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 391, __pyx_L1_error)

  /* "av/codec/context.pyx":393
 *         err_check(res)
 * 
 *         out = []             # <<<<<<<<<<<<<<
 *         while True:
 *             packet = self._recv_packet()
 */
  __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 393, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_out = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "av/codec/context.pyx":394
 * 
 *         out = []
 *         while True:             # <<<<<<<<<<<<<<
 *             packet = self._recv_packet()
 *             if packet:
 */
  while (1) {

    /* "av/codec/context.pyx":395
 *         out = []
 *         while True:
 *             packet = self._recv_packet()             # <<<<<<<<<<<<<<
 *             if packet:
 *                 out.append(packet)
 */
    __pyx_t_4 = ((struct __pyx_vtabstruct_2av_5codec_7context_CodecContext *)__pyx_v_self->__pyx_vtab)->_recv_packet(__pyx_v_self); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 395, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_2av_6packet_Packet))))) __PYX_ERR(0, 395, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_packet, ((struct __pyx_obj_2av_6packet_Packet *)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "av/codec/context.pyx":396
 *         while True:
 *             packet = self._recv_packet()
 *             if packet:             # <<<<<<<<<<<<<<
 *                 out.append(packet)
 *             else:
 */
    __pyx_t_2 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_packet)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 396, __pyx_L1_error)
    if (__pyx_t_2) {

      /* "av/codec/context.pyx":397
 *             packet = self._recv_packet()
 *             if packet:
 *                 out.append(packet)             # <<<<<<<<<<<<<<
 *             else:
 *                 break
 */
      __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_out, ((PyObject *)__pyx_v_packet)); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 397, __pyx_L1_error)

      /* "av/codec/context.pyx":396
 *         while True:
 *             packet = self._recv_packet()
 *             if packet:             # <<<<<<<<<<<<<<
 *                 out.append(packet)
 *             else:
 */
      goto __pyx_L8;
    }

    /* "av/codec/context.pyx":399
 *                 out.append(packet)
 *             else:
 *                 break             # <<<<<<<<<<<<<<
 *         return out
 * 
 */
    /*else*/ {
      goto __pyx_L7_break;
    }
    __pyx_L8:;
  }
  __pyx_L7_break:;

  /* "av/codec/context.pyx":400
 *             else:
 *                 break
 *         return out             # <<<<<<<<<<<<<<
 * 
 *     cdef _send_packet_and_recv(self, Packet packet):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_out);
  __pyx_r = __pyx_v_out;
  goto __pyx_L0;

  /* "av/codec/context.pyx":384
 *         return packets
 * 
 *     cdef _send_frame_and_recv(self, Frame frame):             # <<<<<<<<<<<<<<
 * 
 *         cdef Packet packet
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("av.codec.context.CodecContext._send_frame_and_recv", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_packet);
  __Pyx_XDECREF(__pyx_v_out);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pyx":402
 *         return out
 * 
 *     cdef _send_packet_and_recv(self, Packet packet):             # <<<<<<<<<<<<<<
 * 
 *         cdef Frame frame
 */

static PyObject *__pyx_f_2av_5codec_7context_12CodecContext__send_packet_and_recv(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, struct __pyx_obj_2av_6packet_Packet *__pyx_v_packet) {
  struct __pyx_obj_2av_5frame_Frame *__pyx_v_frame = 0;
  int __pyx_v_res;
  PyObject *__pyx_v_out = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct AVPacket *__pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_send_packet_and_recv", 0);

  /* "av/codec/context.pyx":407
 * 
 *         cdef int res
 *         with nogil:             # <<<<<<<<<<<<<<
 *             res = lib.avcodec_send_packet(self.ptr, packet.ptr if packet is not None else NULL)
 *         err_check(res)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "av/codec/context.pyx":408
 *         cdef int res
 *         with nogil:
 *             res = lib.avcodec_send_packet(self.ptr, packet.ptr if packet is not None else NULL)             # <<<<<<<<<<<<<<
 *         err_check(res)
 * 
 */
        __pyx_t_2 = (((PyObject *)__pyx_v_packet) != Py_None);
        if ((__pyx_t_2 != 0)) {
          __pyx_t_1 = __pyx_v_packet->ptr;
        } else {
          __pyx_t_1 = NULL;
        }
        __pyx_v_res = avcodec_send_packet(__pyx_v_self->ptr, __pyx_t_1);
      }

      /* "av/codec/context.pyx":407
 * 
 *         cdef int res
 *         with nogil:             # <<<<<<<<<<<<<<
 *             res = lib.avcodec_send_packet(self.ptr, packet.ptr if packet is not None else NULL)
 *         err_check(res)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "av/codec/context.pyx":409
 *         with nogil:
 *             res = lib.avcodec_send_packet(self.ptr, packet.ptr if packet is not None else NULL)
 *         err_check(res)             # <<<<<<<<<<<<<<
 * 
 *         out = []
 */
  __pyx_t_3 = __pyx_f_2av_5error_err_check(__pyx_v_res, 0, NULL); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 409, __pyx_L1_error)

  /* "av/codec/context.pyx":411
 *         err_check(res)
 * 
 *         out = []             # <<<<<<<<<<<<<<
 *         while True:
 *             frame = self._recv_frame()
 */
  __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 411, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_out = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "av/codec/context.pyx":412
 * 
 *         out = []
 *         while True:             # <<<<<<<<<<<<<<
 *             frame = self._recv_frame()
 *             if frame:
 */
  while (1) {

    /* "av/codec/context.pyx":413
 *         out = []
 *         while True:
 *             frame = self._recv_frame()             # <<<<<<<<<<<<<<
 *             if frame:
 *                 out.append(frame)
 */
    __pyx_t_4 = ((struct __pyx_vtabstruct_2av_5codec_7context_CodecContext *)__pyx_v_self->__pyx_vtab)->_recv_frame(__pyx_v_self); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_2av_5frame_Frame))))) __PYX_ERR(0, 413, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_frame, ((struct __pyx_obj_2av_5frame_Frame *)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "av/codec/context.pyx":414
 *         while True:
 *             frame = self._recv_frame()
 *             if frame:             # <<<<<<<<<<<<<<
 *                 out.append(frame)
 *             else:
 */
    __pyx_t_2 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_frame)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 414, __pyx_L1_error)
    if (__pyx_t_2) {

      /* "av/codec/context.pyx":415
 *             frame = self._recv_frame()
 *             if frame:
 *                 out.append(frame)             # <<<<<<<<<<<<<<
 *             else:
 *                 break
 */
      __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_out, ((PyObject *)__pyx_v_frame)); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 415, __pyx_L1_error)

      /* "av/codec/context.pyx":414
 *         while True:
 *             frame = self._recv_frame()
 *             if frame:             # <<<<<<<<<<<<<<
 *                 out.append(frame)
 *             else:
 */
      goto __pyx_L8;
    }

    /* "av/codec/context.pyx":417
 *                 out.append(frame)
 *             else:
 *                 break             # <<<<<<<<<<<<<<
 *         return out
 * 
 */
    /*else*/ {
      goto __pyx_L7_break;
    }
    __pyx_L8:;
  }
  __pyx_L7_break:;

  /* "av/codec/context.pyx":418
 *             else:
 *                 break
 *         return out             # <<<<<<<<<<<<<<
 * 
 *     cdef _prepare_frames_for_encode(self, Frame frame):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_out);
  __pyx_r = __pyx_v_out;
  goto __pyx_L0;

  /* "av/codec/context.pyx":402
 *         return out
 * 
 *     cdef _send_packet_and_recv(self, Packet packet):             # <<<<<<<<<<<<<<
 * 
 *         cdef Frame frame
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("av.codec.context.CodecContext._send_packet_and_recv", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_frame);
  __Pyx_XDECREF(__pyx_v_out);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pyx":420
 *         return out
 * 
 *     cdef _prepare_frames_for_encode(self, Frame frame):             # <<<<<<<<<<<<<<
 *         return [frame]
 * 
 */

static PyObject *__pyx_f_2av_5codec_7context_12CodecContext__prepare_frames_for_encode(CYTHON_UNUSED struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, struct __pyx_obj_2av_5frame_Frame *__pyx_v_frame) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_prepare_frames_for_encode", 0);

  /* "av/codec/context.pyx":421
 * 
 *     cdef _prepare_frames_for_encode(self, Frame frame):
 *         return [frame]             # <<<<<<<<<<<<<<
 * 
 *     cdef Frame _alloc_next_frame(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 421, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject *)__pyx_v_frame));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_frame));
  PyList_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_frame));
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "av/codec/context.pyx":420
 *         return out
 * 
 *     cdef _prepare_frames_for_encode(self, Frame frame):             # <<<<<<<<<<<<<<
 *         return [frame]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("av.codec.context.CodecContext._prepare_frames_for_encode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pyx":423
 *         return [frame]
 * 
 *     cdef Frame _alloc_next_frame(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError('Base CodecContext cannot decode.')
 * 
 */

static struct __pyx_obj_2av_5frame_Frame *__pyx_f_2av_5codec_7context_12CodecContext__alloc_next_frame(CYTHON_UNUSED struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self) {
  struct __pyx_obj_2av_5frame_Frame *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_alloc_next_frame", 0);

  /* "av/codec/context.pyx":424
 * 
 *     cdef Frame _alloc_next_frame(self):
 *         raise NotImplementedError('Base CodecContext cannot decode.')             # <<<<<<<<<<<<<<
 * 
 *     cdef _recv_frame(self):
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_NotImplementedError, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 424, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 424, __pyx_L1_error)

  /* "av/codec/context.pyx":423
 *         return [frame]
 * 
 *     cdef Frame _alloc_next_frame(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError('Base CodecContext cannot decode.')
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("av.codec.context.CodecContext._alloc_next_frame", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pyx":426
 *         raise NotImplementedError('Base CodecContext cannot decode.')
 * 
 *     cdef _recv_frame(self):             # <<<<<<<<<<<<<<
 * 
 *         if not self._next_frame:
 */

static PyObject *__pyx_f_2av_5codec_7context_12CodecContext__recv_frame(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self) {
  struct __pyx_obj_2av_5frame_Frame *__pyx_v_frame = 0;
  int __pyx_v_res;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_recv_frame", 0);

  /* "av/codec/context.pyx":428
 *     cdef _recv_frame(self):
 * 
 *         if not self._next_frame:             # <<<<<<<<<<<<<<
 *             self._next_frame = self._alloc_next_frame()
 *         cdef Frame frame = self._next_frame
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_self->_next_frame)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 428, __pyx_L1_error)
  __pyx_t_2 = ((!__pyx_t_1) != 0);
  if (__pyx_t_2) {

    /* "av/codec/context.pyx":429
 * 
 *         if not self._next_frame:
 *             self._next_frame = self._alloc_next_frame()             # <<<<<<<<<<<<<<
 *         cdef Frame frame = self._next_frame
 * 
 */
    __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_2av_5codec_7context_CodecContext *)__pyx_v_self->__pyx_vtab)->_alloc_next_frame(__pyx_v_self)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 429, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->_next_frame);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->_next_frame));
    __pyx_v_self->_next_frame = ((struct __pyx_obj_2av_5frame_Frame *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "av/codec/context.pyx":428
 *     cdef _recv_frame(self):
 * 
 *         if not self._next_frame:             # <<<<<<<<<<<<<<
 *             self._next_frame = self._alloc_next_frame()
 *         cdef Frame frame = self._next_frame
 */
  }

  /* "av/codec/context.pyx":430
 *         if not self._next_frame:
 *             self._next_frame = self._alloc_next_frame()
 *         cdef Frame frame = self._next_frame             # <<<<<<<<<<<<<<
 * 
 *         cdef int res
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_self->_next_frame);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_frame = ((struct __pyx_obj_2av_5frame_Frame *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "av/codec/context.pyx":433
 * 
 *         cdef int res
 *         with nogil:             # <<<<<<<<<<<<<<
 *             res = lib.avcodec_receive_frame(self.ptr, frame.ptr)
 * 
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "av/codec/context.pyx":434
 *         cdef int res
 *         with nogil:
 *             res = lib.avcodec_receive_frame(self.ptr, frame.ptr)             # <<<<<<<<<<<<<<
 * 
 *         if res == -EAGAIN or res == lib.AVERROR_EOF:
 */
        __pyx_v_res = avcodec_receive_frame(__pyx_v_self->ptr, __pyx_v_frame->ptr);
      }

      /* "av/codec/context.pyx":433
 * 
 *         cdef int res
 *         with nogil:             # <<<<<<<<<<<<<<
 *             res = lib.avcodec_receive_frame(self.ptr, frame.ptr)
 * 
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L6;
        }
        __pyx_L6:;
      }
  }

  /* "av/codec/context.pyx":436
 *             res = lib.avcodec_receive_frame(self.ptr, frame.ptr)
 * 
 *         if res == -EAGAIN or res == lib.AVERROR_EOF:             # <<<<<<<<<<<<<<
 *             return
 *         err_check(res)
 */
  __pyx_t_1 = ((__pyx_v_res == (-EAGAIN)) != 0);
  if (!__pyx_t_1) {
  } else {
    __pyx_t_2 = __pyx_t_1;
    goto __pyx_L8_bool_binop_done;
  }
  __pyx_t_1 = ((__pyx_v_res == AVERROR_EOF) != 0);
  __pyx_t_2 = __pyx_t_1;
  __pyx_L8_bool_binop_done:;
  if (__pyx_t_2) {

    /* "av/codec/context.pyx":437
 * 
 *         if res == -EAGAIN or res == lib.AVERROR_EOF:
 *             return             # <<<<<<<<<<<<<<
 *         err_check(res)
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "av/codec/context.pyx":436
 *             res = lib.avcodec_receive_frame(self.ptr, frame.ptr)
 * 
 *         if res == -EAGAIN or res == lib.AVERROR_EOF:             # <<<<<<<<<<<<<<
 *             return
 *         err_check(res)
 */
  }

  /* "av/codec/context.pyx":438
 *         if res == -EAGAIN or res == lib.AVERROR_EOF:
 *             return
 *         err_check(res)             # <<<<<<<<<<<<<<
 * 
 *         if not res:
 */
  __pyx_t_4 = __pyx_f_2av_5error_err_check(__pyx_v_res, 0, NULL); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 438, __pyx_L1_error)

  /* "av/codec/context.pyx":440
 *         err_check(res)
 * 
 *         if not res:             # <<<<<<<<<<<<<<
 *             self._next_frame = None
 *             return frame
 */
  __pyx_t_2 = ((!(__pyx_v_res != 0)) != 0);
  if (__pyx_t_2) {

    /* "av/codec/context.pyx":441
 * 
 *         if not res:
 *             self._next_frame = None             # <<<<<<<<<<<<<<
 *             return frame
 * 
 */
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_self->_next_frame);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->_next_frame));
    __pyx_v_self->_next_frame = ((struct __pyx_obj_2av_5frame_Frame *)Py_None);

    /* "av/codec/context.pyx":442
 *         if not res:
 *             self._next_frame = None
 *             return frame             # <<<<<<<<<<<<<<
 * 
 *     cdef _recv_packet(self):
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject *)__pyx_v_frame));
    __pyx_r = ((PyObject *)__pyx_v_frame);
    goto __pyx_L0;

    /* "av/codec/context.pyx":440
 *         err_check(res)
 * 
 *         if not res:             # <<<<<<<<<<<<<<
 *             self._next_frame = None
 *             return frame
 */
  }

  /* "av/codec/context.pyx":426
 *         raise NotImplementedError('Base CodecContext cannot decode.')
 * 
 *     cdef _recv_frame(self):             # <<<<<<<<<<<<<<
 * 
 *         if not self._next_frame:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("av.codec.context.CodecContext._recv_frame", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_frame);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pyx":444
 *             return frame
 * 
 *     cdef _recv_packet(self):             # <<<<<<<<<<<<<<
 * 
 *         cdef Packet packet = Packet()
 */

static PyObject *__pyx_f_2av_5codec_7context_12CodecContext__recv_packet(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self) {
  struct __pyx_obj_2av_6packet_Packet *__pyx_v_packet = 0;
  int __pyx_v_res;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_recv_packet", 0);

  /* "av/codec/context.pyx":446
 *     cdef _recv_packet(self):
 * 
 *         cdef Packet packet = Packet()             # <<<<<<<<<<<<<<
 * 
 *         cdef int res
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_2av_6packet_Packet)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 446, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_packet = ((struct __pyx_obj_2av_6packet_Packet *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "av/codec/context.pyx":449
 * 
 *         cdef int res
 *         with nogil:             # <<<<<<<<<<<<<<
 *             res = lib.avcodec_receive_packet(self.ptr, packet.ptr)
 *         if res == -EAGAIN or res == lib.AVERROR_EOF:
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "av/codec/context.pyx":450
 *         cdef int res
 *         with nogil:
 *             res = lib.avcodec_receive_packet(self.ptr, packet.ptr)             # <<<<<<<<<<<<<<
 *         if res == -EAGAIN or res == lib.AVERROR_EOF:
 *             return
 */
        __pyx_v_res = avcodec_receive_packet(__pyx_v_self->ptr, __pyx_v_packet->ptr);
      }

      /* "av/codec/context.pyx":449
 * 
 *         cdef int res
 *         with nogil:             # <<<<<<<<<<<<<<
 *             res = lib.avcodec_receive_packet(self.ptr, packet.ptr)
 *         if res == -EAGAIN or res == lib.AVERROR_EOF:
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "av/codec/context.pyx":451
 *         with nogil:
 *             res = lib.avcodec_receive_packet(self.ptr, packet.ptr)
 *         if res == -EAGAIN or res == lib.AVERROR_EOF:             # <<<<<<<<<<<<<<
 *             return
 *         err_check(res)
 */
  __pyx_t_3 = ((__pyx_v_res == (-EAGAIN)) != 0);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_3 = ((__pyx_v_res == AVERROR_EOF) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_2) {

    /* "av/codec/context.pyx":452
 *             res = lib.avcodec_receive_packet(self.ptr, packet.ptr)
 *         if res == -EAGAIN or res == lib.AVERROR_EOF:
 *             return             # <<<<<<<<<<<<<<
 *         err_check(res)
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "av/codec/context.pyx":451
 *         with nogil:
 *             res = lib.avcodec_receive_packet(self.ptr, packet.ptr)
 *         if res == -EAGAIN or res == lib.AVERROR_EOF:             # <<<<<<<<<<<<<<
 *             return
 *         err_check(res)
 */
  }

  /* "av/codec/context.pyx":453
 *         if res == -EAGAIN or res == lib.AVERROR_EOF:
 *             return
 *         err_check(res)             # <<<<<<<<<<<<<<
 * 
 *         if not res:
 */
  __pyx_t_4 = __pyx_f_2av_5error_err_check(__pyx_v_res, 0, NULL); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 453, __pyx_L1_error)

  /* "av/codec/context.pyx":455
 *         err_check(res)
 * 
 *         if not res:             # <<<<<<<<<<<<<<
 *             return packet
 * 
 */
  __pyx_t_2 = ((!(__pyx_v_res != 0)) != 0);
  if (__pyx_t_2) {

    /* "av/codec/context.pyx":456
 * 
 *         if not res:
 *             return packet             # <<<<<<<<<<<<<<
 * 
 *     cpdef encode(self, Frame frame=None):
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject *)__pyx_v_packet));
    __pyx_r = ((PyObject *)__pyx_v_packet);
    goto __pyx_L0;

    /* "av/codec/context.pyx":455
 *         err_check(res)
 * 
 *         if not res:             # <<<<<<<<<<<<<<
 *             return packet
 * 
 */
  }

  /* "av/codec/context.pyx":444
 *             return frame
 * 
 *     cdef _recv_packet(self):             # <<<<<<<<<<<<<<
 * 
 *         cdef Packet packet = Packet()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("av.codec.context.CodecContext._recv_packet", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_packet);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pyx":458
 *             return packet
 * 
 *     cpdef encode(self, Frame frame=None):             # <<<<<<<<<<<<<<
 *         """Encode a list of :class:`.Packet` from the given :class:`.Frame`."""
 * 
 */

static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_23encode(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_2av_5codec_7context_12CodecContext_encode(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, int __pyx_skip_dispatch, struct __pyx_opt_args_2av_5codec_7context_12CodecContext_encode *__pyx_optional_args) {
  struct __pyx_obj_2av_5frame_Frame *__pyx_v_frame = ((struct __pyx_obj_2av_5frame_Frame *)Py_None);
  PyObject *__pyx_v_frames = NULL;
  PyObject *__pyx_v_res = NULL;
  PyObject *__pyx_v_packet = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  struct __pyx_opt_args_2av_5codec_7context_12CodecContext_open __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  Py_ssize_t __pyx_t_10;
  PyObject *(*__pyx_t_11)(PyObject *);
  int __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("encode", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_frame = __pyx_optional_args->frame;
    }
  }
  __Pyx_INCREF((PyObject *)__pyx_v_frame);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_encode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 458, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_2av_5codec_7context_12CodecContext_23encode)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_frame)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_frame));
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 458, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "av/codec/context.pyx":461
 *         """Encode a list of :class:`.Packet` from the given :class:`.Frame`."""
 * 
 *         if self.ptr.codec_type not in [lib.AVMEDIA_TYPE_VIDEO, lib.AVMEDIA_TYPE_AUDIO]:             # <<<<<<<<<<<<<<
 *             raise NotImplementedError('Encoding is only supported for audio and video.')
 * 
 */
  switch (__pyx_v_self->ptr->codec_type) {
    case AVMEDIA_TYPE_VIDEO:
    case AVMEDIA_TYPE_AUDIO:
    __pyx_t_5 = 0;
    break;
    default:
    __pyx_t_5 = 1;
    break;
  }
  __pyx_t_6 = (__pyx_t_5 != 0);
  if (unlikely(__pyx_t_6)) {

    /* "av/codec/context.pyx":462
 * 
 *         if self.ptr.codec_type not in [lib.AVMEDIA_TYPE_VIDEO, lib.AVMEDIA_TYPE_AUDIO]:
 *             raise NotImplementedError('Encoding is only supported for audio and video.')             # <<<<<<<<<<<<<<
 * 
 *         self.open(strict=False)
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_NotImplementedError, __pyx_tuple__8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 462, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 462, __pyx_L1_error)

    /* "av/codec/context.pyx":461
 *         """Encode a list of :class:`.Packet` from the given :class:`.Frame`."""
 * 
 *         if self.ptr.codec_type not in [lib.AVMEDIA_TYPE_VIDEO, lib.AVMEDIA_TYPE_AUDIO]:             # <<<<<<<<<<<<<<
 *             raise NotImplementedError('Encoding is only supported for audio and video.')
 * 
 */
  }

  /* "av/codec/context.pyx":464
 *             raise NotImplementedError('Encoding is only supported for audio and video.')
 * 
 *         self.open(strict=False)             # <<<<<<<<<<<<<<
 * 
 *         frames = self._prepare_frames_for_encode(frame)
 */
  __pyx_t_7.__pyx_n = 1;
  __pyx_t_7.strict = 0;
  __pyx_t_1 = ((struct __pyx_vtabstruct_2av_5codec_7context_CodecContext *)__pyx_v_self->__pyx_vtab)->open(__pyx_v_self, 0, &__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "av/codec/context.pyx":466
 *         self.open(strict=False)
 * 
 *         frames = self._prepare_frames_for_encode(frame)             # <<<<<<<<<<<<<<
 * 
 *         # Assert the frames are in our time base.
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_2av_5codec_7context_CodecContext *)__pyx_v_self->__pyx_vtab)->_prepare_frames_for_encode(__pyx_v_self, __pyx_v_frame); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 466, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_frames = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "av/codec/context.pyx":470
 *         # Assert the frames are in our time base.
 *         # TODO: Don't mutate time.
 *         for frame in frames:             # <<<<<<<<<<<<<<
 *             if frame is not None:
 *                 frame._rebase_time(self.ptr.time_base)
 */
  if (likely(PyList_CheckExact(__pyx_v_frames)) || PyTuple_CheckExact(__pyx_v_frames)) {
    __pyx_t_1 = __pyx_v_frames; __Pyx_INCREF(__pyx_t_1); __pyx_t_8 = 0;
    __pyx_t_9 = NULL;
  } else {
    __pyx_t_8 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_frames); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 470, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 470, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_9)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_8); __Pyx_INCREF(__pyx_t_2); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 470, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 470, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_8); __Pyx_INCREF(__pyx_t_2); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 470, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 470, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_9(__pyx_t_1);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 470, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_2av_5frame_Frame))))) __PYX_ERR(0, 470, __pyx_L1_error)
    __Pyx_DECREF_SET(__pyx_v_frame, ((struct __pyx_obj_2av_5frame_Frame *)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "av/codec/context.pyx":471
 *         # TODO: Don't mutate time.
 *         for frame in frames:
 *             if frame is not None:             # <<<<<<<<<<<<<<
 *                 frame._rebase_time(self.ptr.time_base)
 * 
 */
    __pyx_t_6 = (((PyObject *)__pyx_v_frame) != Py_None);
    __pyx_t_5 = (__pyx_t_6 != 0);
    if (__pyx_t_5) {

      /* "av/codec/context.pyx":472
 *         for frame in frames:
 *             if frame is not None:
 *                 frame._rebase_time(self.ptr.time_base)             # <<<<<<<<<<<<<<
 * 
 *         res = []
 */
      __pyx_t_2 = ((struct __pyx_vtabstruct_2av_5frame_Frame *)__pyx_v_frame->__pyx_vtab)->_rebase_time(__pyx_v_frame, __pyx_v_self->ptr->time_base); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 472, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "av/codec/context.pyx":471
 *         # TODO: Don't mutate time.
 *         for frame in frames:
 *             if frame is not None:             # <<<<<<<<<<<<<<
 *                 frame._rebase_time(self.ptr.time_base)
 * 
 */
    }

    /* "av/codec/context.pyx":470
 *         # Assert the frames are in our time base.
 *         # TODO: Don't mutate time.
 *         for frame in frames:             # <<<<<<<<<<<<<<
 *             if frame is not None:
 *                 frame._rebase_time(self.ptr.time_base)
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "av/codec/context.pyx":474
 *                 frame._rebase_time(self.ptr.time_base)
 * 
 *         res = []             # <<<<<<<<<<<<<<
 *         for frame in frames:
 *             for packet in self._send_frame_and_recv(frame):
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 474, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_res = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "av/codec/context.pyx":475
 * 
 *         res = []
 *         for frame in frames:             # <<<<<<<<<<<<<<
 *             for packet in self._send_frame_and_recv(frame):
 *                 self._setup_encoded_packet(packet)
 */
  if (likely(PyList_CheckExact(__pyx_v_frames)) || PyTuple_CheckExact(__pyx_v_frames)) {
    __pyx_t_1 = __pyx_v_frames; __Pyx_INCREF(__pyx_t_1); __pyx_t_8 = 0;
    __pyx_t_9 = NULL;
  } else {
    __pyx_t_8 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_frames); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 475, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 475, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_9)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_8); __Pyx_INCREF(__pyx_t_2); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 475, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 475, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_8); __Pyx_INCREF(__pyx_t_2); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 475, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 475, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_9(__pyx_t_1);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 475, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_2av_5frame_Frame))))) __PYX_ERR(0, 475, __pyx_L1_error)
    __Pyx_DECREF_SET(__pyx_v_frame, ((struct __pyx_obj_2av_5frame_Frame *)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "av/codec/context.pyx":476
 *         res = []
 *         for frame in frames:
 *             for packet in self._send_frame_and_recv(frame):             # <<<<<<<<<<<<<<
 *                 self._setup_encoded_packet(packet)
 *                 res.append(packet)
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_2av_5codec_7context_CodecContext *)__pyx_v_self->__pyx_vtab)->_send_frame_and_recv(__pyx_v_self, __pyx_v_frame); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 476, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
      __pyx_t_3 = __pyx_t_2; __Pyx_INCREF(__pyx_t_3); __pyx_t_10 = 0;
      __pyx_t_11 = NULL;
    } else {
      __pyx_t_10 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 476, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_11 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 476, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    for (;;) {
      if (likely(!__pyx_t_11)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_10); __Pyx_INCREF(__pyx_t_2); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 476, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 476, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_10); __Pyx_INCREF(__pyx_t_2); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 476, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 476, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_11(__pyx_t_3);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 476, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_XDECREF_SET(__pyx_v_packet, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "av/codec/context.pyx":477
 *         for frame in frames:
 *             for packet in self._send_frame_and_recv(frame):
 *                 self._setup_encoded_packet(packet)             # <<<<<<<<<<<<<<
 *                 res.append(packet)
 *         return res
 */
      if (!(likely(((__pyx_v_packet) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_packet, __pyx_ptype_2av_6packet_Packet))))) __PYX_ERR(0, 477, __pyx_L1_error)
      __pyx_t_2 = ((struct __pyx_vtabstruct_2av_5codec_7context_CodecContext *)__pyx_v_self->__pyx_vtab)->_setup_encoded_packet(__pyx_v_self, ((struct __pyx_obj_2av_6packet_Packet *)__pyx_v_packet)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 477, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "av/codec/context.pyx":478
 *             for packet in self._send_frame_and_recv(frame):
 *                 self._setup_encoded_packet(packet)
 *                 res.append(packet)             # <<<<<<<<<<<<<<
 *         return res
 * 
 */
      __pyx_t_12 = __Pyx_PyList_Append(__pyx_v_res, __pyx_v_packet); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 478, __pyx_L1_error)

      /* "av/codec/context.pyx":476
 *         res = []
 *         for frame in frames:
 *             for packet in self._send_frame_and_recv(frame):             # <<<<<<<<<<<<<<
 *                 self._setup_encoded_packet(packet)
 *                 res.append(packet)
 */
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "av/codec/context.pyx":475
 * 
 *         res = []
 *         for frame in frames:             # <<<<<<<<<<<<<<
 *             for packet in self._send_frame_and_recv(frame):
 *                 self._setup_encoded_packet(packet)
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "av/codec/context.pyx":479
 *                 self._setup_encoded_packet(packet)
 *                 res.append(packet)
 *         return res             # <<<<<<<<<<<<<<
 * 
 *     cdef _setup_encoded_packet(self, Packet packet):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_res);
  __pyx_r = __pyx_v_res;
  goto __pyx_L0;

  /* "av/codec/context.pyx":458
 *             return packet
 * 
 *     cpdef encode(self, Frame frame=None):             # <<<<<<<<<<<<<<
 *         """Encode a list of :class:`.Packet` from the given :class:`.Frame`."""
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("av.codec.context.CodecContext.encode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_frames);
  __Pyx_XDECREF(__pyx_v_res);
  __Pyx_XDECREF(__pyx_v_packet);
  __Pyx_XDECREF((PyObject *)__pyx_v_frame);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_23encode(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_2av_5codec_7context_12CodecContext_22encode[] = "CodecContext.encode(Frame frame=None)\nEncode a list of :class:`.Packet` from the given :class:`.Frame`.";
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_23encode(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_2av_5frame_Frame *__pyx_v_frame = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("encode (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_frame,0};
    PyObject* values[1] = {0};
    values[0] = (PyObject *)((struct __pyx_obj_2av_5frame_Frame *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_frame);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "encode") < 0)) __PYX_ERR(0, 458, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_frame = ((struct __pyx_obj_2av_5frame_Frame *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("encode", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 458, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("av.codec.context.CodecContext.encode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_frame), __pyx_ptype_2av_5frame_Frame, 1, "frame", 0))) __PYX_ERR(0, 458, __pyx_L1_error)
  __pyx_r = __pyx_pf_2av_5codec_7context_12CodecContext_22encode(((struct __pyx_obj_2av_5codec_7context_CodecContext *)__pyx_v_self), __pyx_v_frame);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_22encode(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, struct __pyx_obj_2av_5frame_Frame *__pyx_v_frame) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_2av_5codec_7context_12CodecContext_encode __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("encode", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.frame = __pyx_v_frame;
  __pyx_t_1 = __pyx_vtabptr_2av_5codec_7context_CodecContext->encode(__pyx_v_self, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 458, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("av.codec.context.CodecContext.encode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pyx":481
 *         return res
 * 
 *     cdef _setup_encoded_packet(self, Packet packet):             # <<<<<<<<<<<<<<
 *         # We coerced the frame's time_base into the CodecContext's during encoding,
 *         # and FFmpeg copied the frame's pts/dts to the packet, so keep track of
 */

static PyObject *__pyx_f_2av_5codec_7context_12CodecContext__setup_encoded_packet(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, struct __pyx_obj_2av_6packet_Packet *__pyx_v_packet) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  AVRational __pyx_t_1;
  __Pyx_RefNannySetupContext("_setup_encoded_packet", 0);

  /* "av/codec/context.pyx":489
 *         # NOTE: if the CodecContext's time_base is altered during encoding, all bets
 *         # are off!
 *         packet._time_base = self.ptr.time_base             # <<<<<<<<<<<<<<
 * 
 *     cpdef decode(self, Packet packet=None):
 */
  __pyx_t_1 = __pyx_v_self->ptr->time_base;
  __pyx_v_packet->_time_base = __pyx_t_1;

  /* "av/codec/context.pyx":481
 *         return res
 * 
 *     cdef _setup_encoded_packet(self, Packet packet):             # <<<<<<<<<<<<<<
 *         # We coerced the frame's time_base into the CodecContext's during encoding,
 *         # and FFmpeg copied the frame's pts/dts to the packet, so keep track of
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pyx":491
 *         packet._time_base = self.ptr.time_base
 * 
 *     cpdef decode(self, Packet packet=None):             # <<<<<<<<<<<<<<
 *         """Decode a list of :class:`.Frame` from the given :class:`.Packet`.
 * 
 */

static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_25decode(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_2av_5codec_7context_12CodecContext_decode(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, int __pyx_skip_dispatch, struct __pyx_opt_args_2av_5codec_7context_12CodecContext_decode *__pyx_optional_args) {
  struct __pyx_obj_2av_6packet_Packet *__pyx_v_packet = ((struct __pyx_obj_2av_6packet_Packet *)Py_None);
  PyObject *__pyx_v_res = NULL;
  PyObject *__pyx_v_frame = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  struct __pyx_opt_args_2av_5codec_7context_12CodecContext_open __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("decode", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_packet = __pyx_optional_args->packet;
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_decode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 491, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_2av_5codec_7context_12CodecContext_25decode)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_packet)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_packet));
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 491, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "av/codec/context.pyx":500
 *         """
 * 
 *         if not self.codec.ptr:             # <<<<<<<<<<<<<<
 *             raise ValueError('cannot decode unknown codec')
 * 
 */
  __pyx_t_5 = ((!(__pyx_v_self->codec->ptr != 0)) != 0);
  if (unlikely(__pyx_t_5)) {

    /* "av/codec/context.pyx":501
 * 
 *         if not self.codec.ptr:
 *             raise ValueError('cannot decode unknown codec')             # <<<<<<<<<<<<<<
 * 
 *         self.open(strict=False)
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 501, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 501, __pyx_L1_error)

    /* "av/codec/context.pyx":500
 *         """
 * 
 *         if not self.codec.ptr:             # <<<<<<<<<<<<<<
 *             raise ValueError('cannot decode unknown codec')
 * 
 */
  }

  /* "av/codec/context.pyx":503
 *             raise ValueError('cannot decode unknown codec')
 * 
 *         self.open(strict=False)             # <<<<<<<<<<<<<<
 * 
 *         res = []
 */
  __pyx_t_6.__pyx_n = 1;
  __pyx_t_6.strict = 0;
  __pyx_t_1 = ((struct __pyx_vtabstruct_2av_5codec_7context_CodecContext *)__pyx_v_self->__pyx_vtab)->open(__pyx_v_self, 0, &__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "av/codec/context.pyx":505
 *         self.open(strict=False)
 * 
 *         res = []             # <<<<<<<<<<<<<<
 *         for frame in self._send_packet_and_recv(packet):
 *             if isinstance(frame, Frame):
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 505, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_res = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "av/codec/context.pyx":506
 * 
 *         res = []
 *         for frame in self._send_packet_and_recv(packet):             # <<<<<<<<<<<<<<
 *             if isinstance(frame, Frame):
 *                 self._setup_decoded_frame(frame, packet)
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_2av_5codec_7context_CodecContext *)__pyx_v_self->__pyx_vtab)->_send_packet_and_recv(__pyx_v_self, __pyx_v_packet); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 506, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
  } else {
    __pyx_t_7 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 506, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_8 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 506, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_8)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_7); __Pyx_INCREF(__pyx_t_1); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 506, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 506, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_7); __Pyx_INCREF(__pyx_t_1); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 506, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 506, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_8(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 506, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_frame, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "av/codec/context.pyx":507
 *         res = []
 *         for frame in self._send_packet_and_recv(packet):
 *             if isinstance(frame, Frame):             # <<<<<<<<<<<<<<
 *                 self._setup_decoded_frame(frame, packet)
 *             res.append(frame)
 */
    __pyx_t_5 = __Pyx_TypeCheck(__pyx_v_frame, __pyx_ptype_2av_5frame_Frame); 
    __pyx_t_9 = (__pyx_t_5 != 0);
    if (__pyx_t_9) {

      /* "av/codec/context.pyx":508
 *         for frame in self._send_packet_and_recv(packet):
 *             if isinstance(frame, Frame):
 *                 self._setup_decoded_frame(frame, packet)             # <<<<<<<<<<<<<<
 *             res.append(frame)
 *         return res
 */
      if (!(likely(((__pyx_v_frame) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_frame, __pyx_ptype_2av_5frame_Frame))))) __PYX_ERR(0, 508, __pyx_L1_error)
      __pyx_t_1 = ((struct __pyx_vtabstruct_2av_5codec_7context_CodecContext *)__pyx_v_self->__pyx_vtab)->_setup_decoded_frame(__pyx_v_self, ((struct __pyx_obj_2av_5frame_Frame *)__pyx_v_frame), __pyx_v_packet); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 508, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "av/codec/context.pyx":507
 *         res = []
 *         for frame in self._send_packet_and_recv(packet):
 *             if isinstance(frame, Frame):             # <<<<<<<<<<<<<<
 *                 self._setup_decoded_frame(frame, packet)
 *             res.append(frame)
 */
    }

    /* "av/codec/context.pyx":509
 *             if isinstance(frame, Frame):
 *                 self._setup_decoded_frame(frame, packet)
 *             res.append(frame)             # <<<<<<<<<<<<<<
 *         return res
 * 
 */
    __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_res, __pyx_v_frame); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 509, __pyx_L1_error)

    /* "av/codec/context.pyx":506
 * 
 *         res = []
 *         for frame in self._send_packet_and_recv(packet):             # <<<<<<<<<<<<<<
 *             if isinstance(frame, Frame):
 *                 self._setup_decoded_frame(frame, packet)
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "av/codec/context.pyx":510
 *                 self._setup_decoded_frame(frame, packet)
 *             res.append(frame)
 *         return res             # <<<<<<<<<<<<<<
 * 
 *     cdef _setup_decoded_frame(self, Frame frame, Packet packet):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_res);
  __pyx_r = __pyx_v_res;
  goto __pyx_L0;

  /* "av/codec/context.pyx":491
 *         packet._time_base = self.ptr.time_base
 * 
 *     cpdef decode(self, Packet packet=None):             # <<<<<<<<<<<<<<
 *         """Decode a list of :class:`.Frame` from the given :class:`.Packet`.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("av.codec.context.CodecContext.decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_res);
  __Pyx_XDECREF(__pyx_v_frame);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_25decode(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_2av_5codec_7context_12CodecContext_24decode[] = "CodecContext.decode(Packet packet=None)\nDecode a list of :class:`.Frame` from the given :class:`.Packet`.\n\n        If the packet is None, the buffers will be flushed. This is useful if\n        you do not want the library to automatically re-order frames for you\n        (if they are encoded with a codec that has B-frames).\n\n        ";
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_25decode(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_2av_6packet_Packet *__pyx_v_packet = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("decode (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_packet,0};
    PyObject* values[1] = {0};
    values[0] = (PyObject *)((struct __pyx_obj_2av_6packet_Packet *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_packet);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "decode") < 0)) __PYX_ERR(0, 491, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_packet = ((struct __pyx_obj_2av_6packet_Packet *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("decode", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 491, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("av.codec.context.CodecContext.decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_packet), __pyx_ptype_2av_6packet_Packet, 1, "packet", 0))) __PYX_ERR(0, 491, __pyx_L1_error)
  __pyx_r = __pyx_pf_2av_5codec_7context_12CodecContext_24decode(((struct __pyx_obj_2av_5codec_7context_CodecContext *)__pyx_v_self), __pyx_v_packet);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_24decode(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, struct __pyx_obj_2av_6packet_Packet *__pyx_v_packet) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_2av_5codec_7context_12CodecContext_decode __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("decode", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.packet = __pyx_v_packet;
  __pyx_t_1 = __pyx_vtabptr_2av_5codec_7context_CodecContext->decode(__pyx_v_self, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("av.codec.context.CodecContext.decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pyx":512
 *         return res
 * 
 *     cdef _setup_decoded_frame(self, Frame frame, Packet packet):             # <<<<<<<<<<<<<<
 * 
 *         # Propagate our manual times.
 */

static PyObject *__pyx_f_2av_5codec_7context_12CodecContext__setup_decoded_frame(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, struct __pyx_obj_2av_5frame_Frame *__pyx_v_frame, struct __pyx_obj_2av_6packet_Packet *__pyx_v_packet) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  AVRational __pyx_t_1;
  __Pyx_RefNannySetupContext("_setup_decoded_frame", 0);

  /* "av/codec/context.pyx":519
 *         # TODO: Somehow get this from the stream so we can not pass the
 *         # packet here (because flushing packets are bogus).
 *         frame._time_base = packet._time_base             # <<<<<<<<<<<<<<
 * 
 *         frame.index = self.ptr.frame_number - 1
 */
  __pyx_t_1 = __pyx_v_packet->_time_base;
  __pyx_v_frame->_time_base = __pyx_t_1;

  /* "av/codec/context.pyx":521
 *         frame._time_base = packet._time_base
 * 
 *         frame.index = self.ptr.frame_number - 1             # <<<<<<<<<<<<<<
 * 
 *     property name:
 */
  __pyx_v_frame->index = (__pyx_v_self->ptr->frame_number - 1);

  /* "av/codec/context.pyx":512
 *         return res
 * 
 *     cdef _setup_decoded_frame(self, Frame frame, Packet packet):             # <<<<<<<<<<<<<<
 * 
 *         # Propagate our manual times.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pyx":524
 * 
 *     property name:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.codec.name
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_4name_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_4name_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2av_5codec_7context_12CodecContext_4name___get__(((struct __pyx_obj_2av_5codec_7context_CodecContext *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_4name___get__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "av/codec/context.pyx":525
 *     property name:
 *         def __get__(self):
 *             return self.codec.name             # <<<<<<<<<<<<<<
 * 
 *     property type:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->codec), __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "av/codec/context.pyx":524
 * 
 *     property name:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.codec.name
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("av.codec.context.CodecContext.name.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pyx":528
 * 
 *     property type:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.codec.type
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_4type_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_4type_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2av_5codec_7context_12CodecContext_4type___get__(((struct __pyx_obj_2av_5codec_7context_CodecContext *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_4type___get__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "av/codec/context.pyx":529
 *     property type:
 *         def __get__(self):
 *             return self.codec.type             # <<<<<<<<<<<<<<
 * 
 *     property profile:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->codec), __pyx_n_s_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 529, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "av/codec/context.pyx":528
 * 
 *     property type:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.codec.type
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("av.codec.context.CodecContext.type.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pyx":532
 * 
 *     property profile:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             if self.ptr.codec and lib.av_get_profile_name(self.ptr.codec, self.ptr.profile):
 *                 return lib.av_get_profile_name(self.ptr.codec, self.ptr.profile)
 */

/* Python wrapper */
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_7profile_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_7profile_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2av_5codec_7context_12CodecContext_7profile___get__(((struct __pyx_obj_2av_5codec_7context_CodecContext *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_7profile___get__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "av/codec/context.pyx":533
 *     property profile:
 *         def __get__(self):
 *             if self.ptr.codec and lib.av_get_profile_name(self.ptr.codec, self.ptr.profile):             # <<<<<<<<<<<<<<
 *                 return lib.av_get_profile_name(self.ptr.codec, self.ptr.profile)
 * 
 */
  __pyx_t_2 = (__pyx_v_self->ptr->codec != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (av_get_profile_name(__pyx_v_self->ptr->codec, __pyx_v_self->ptr->profile) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "av/codec/context.pyx":534
 *         def __get__(self):
 *             if self.ptr.codec and lib.av_get_profile_name(self.ptr.codec, self.ptr.profile):
 *                 return lib.av_get_profile_name(self.ptr.codec, self.ptr.profile)             # <<<<<<<<<<<<<<
 * 
 *     property time_base:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_PyStr_FromString(av_get_profile_name(__pyx_v_self->ptr->codec, __pyx_v_self->ptr->profile)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 534, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "av/codec/context.pyx":533
 *     property profile:
 *         def __get__(self):
 *             if self.ptr.codec and lib.av_get_profile_name(self.ptr.codec, self.ptr.profile):             # <<<<<<<<<<<<<<
 *                 return lib.av_get_profile_name(self.ptr.codec, self.ptr.profile)
 * 
 */
  }

  /* "av/codec/context.pyx":532
 * 
 *     property profile:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             if self.ptr.codec and lib.av_get_profile_name(self.ptr.codec, self.ptr.profile):
 *                 return lib.av_get_profile_name(self.ptr.codec, self.ptr.profile)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("av.codec.context.CodecContext.profile.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pyx":537
 * 
 *     property time_base:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             if self.is_decoder:
 *                 warnings.warn(
 */

/* Python wrapper */
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_9time_base_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_9time_base_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2av_5codec_7context_12CodecContext_9time_base___get__(((struct __pyx_obj_2av_5codec_7context_CodecContext *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_9time_base___get__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "av/codec/context.pyx":538
 *     property time_base:
 *         def __get__(self):
 *             if self.is_decoder:             # <<<<<<<<<<<<<<
 *                 warnings.warn(
 *                     "Using CodecContext.time_base for decoders is deprecated.",
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_is_decoder); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 538, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "av/codec/context.pyx":539
 *         def __get__(self):
 *             if self.is_decoder:
 *                 warnings.warn(             # <<<<<<<<<<<<<<
 *                     "Using CodecContext.time_base for decoders is deprecated.",
 *                     AVDeprecationWarning
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_warnings); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 539, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_warn); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 539, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "av/codec/context.pyx":541
 *                 warnings.warn(
 *                     "Using CodecContext.time_base for decoders is deprecated.",
 *                     AVDeprecationWarning             # <<<<<<<<<<<<<<
 *                 )
 *             return avrational_to_fraction(&self.ptr.time_base)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_AVDeprecationWarning); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 541, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_kp_s_Using_CodecContext_time_base_for, __pyx_t_3};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 539, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_kp_s_Using_CodecContext_time_base_for, __pyx_t_3};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 539, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else
    #endif
    {
      __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 539, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (__pyx_t_5) {
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
      }
      __Pyx_INCREF(__pyx_kp_s_Using_CodecContext_time_base_for);
      __Pyx_GIVEREF(__pyx_kp_s_Using_CodecContext_time_base_for);
      PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_kp_s_Using_CodecContext_time_base_for);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
      __pyx_t_3 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 539, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "av/codec/context.pyx":538
 *     property time_base:
 *         def __get__(self):
 *             if self.is_decoder:             # <<<<<<<<<<<<<<
 *                 warnings.warn(
 *                     "Using CodecContext.time_base for decoders is deprecated.",
 */
  }

  /* "av/codec/context.pyx":543
 *                     AVDeprecationWarning
 *                 )
 *             return avrational_to_fraction(&self.ptr.time_base)             # <<<<<<<<<<<<<<
 * 
 *         def __set__(self, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_2av_5utils_avrational_to_fraction((&__pyx_v_self->ptr->time_base)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "av/codec/context.pyx":537
 * 
 *     property time_base:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             if self.is_decoder:
 *                 warnings.warn(
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("av.codec.context.CodecContext.time_base.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pyx":545
 *             return avrational_to_fraction(&self.ptr.time_base)
 * 
 *         def __set__(self, value):             # <<<<<<<<<<<<<<
 *             if self.is_decoder:
 *                 warnings.warn(
 */

/* Python wrapper */
static int __pyx_pw_2av_5codec_7context_12CodecContext_9time_base_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_2av_5codec_7context_12CodecContext_9time_base_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2av_5codec_7context_12CodecContext_9time_base_2__set__(((struct __pyx_obj_2av_5codec_7context_CodecContext *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2av_5codec_7context_12CodecContext_9time_base_2__set__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "av/codec/context.pyx":546
 * 
 *         def __set__(self, value):
 *             if self.is_decoder:             # <<<<<<<<<<<<<<
 *                 warnings.warn(
 *                     "Using CodecContext.time_base for decoders is deprecated.",
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_is_decoder); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 546, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "av/codec/context.pyx":547
 *         def __set__(self, value):
 *             if self.is_decoder:
 *                 warnings.warn(             # <<<<<<<<<<<<<<
 *                     "Using CodecContext.time_base for decoders is deprecated.",
 *                     AVDeprecationWarning
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_warnings); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 547, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_warn); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 547, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "av/codec/context.pyx":549
 *                 warnings.warn(
 *                     "Using CodecContext.time_base for decoders is deprecated.",
 *                     AVDeprecationWarning             # <<<<<<<<<<<<<<
 *                 )
 *             to_avrational(value, &self.ptr.time_base)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_AVDeprecationWarning); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 549, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_kp_s_Using_CodecContext_time_base_for, __pyx_t_3};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 547, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_kp_s_Using_CodecContext_time_base_for, __pyx_t_3};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 547, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else
    #endif
    {
      __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 547, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (__pyx_t_5) {
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
      }
      __Pyx_INCREF(__pyx_kp_s_Using_CodecContext_time_base_for);
      __Pyx_GIVEREF(__pyx_kp_s_Using_CodecContext_time_base_for);
      PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_kp_s_Using_CodecContext_time_base_for);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
      __pyx_t_3 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 547, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "av/codec/context.pyx":546
 * 
 *         def __set__(self, value):
 *             if self.is_decoder:             # <<<<<<<<<<<<<<
 *                 warnings.warn(
 *                     "Using CodecContext.time_base for decoders is deprecated.",
 */
  }

  /* "av/codec/context.pyx":551
 *                     AVDeprecationWarning
 *                 )
 *             to_avrational(value, &self.ptr.time_base)             # <<<<<<<<<<<<<<
 * 
 *     property codec_tag:
 */
  __pyx_t_1 = __pyx_f_2av_5utils_to_avrational(__pyx_v_value, (&__pyx_v_self->ptr->time_base)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 551, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "av/codec/context.pyx":545
 *             return avrational_to_fraction(&self.ptr.time_base)
 * 
 *         def __set__(self, value):             # <<<<<<<<<<<<<<
 *             if self.is_decoder:
 *                 warnings.warn(
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("av.codec.context.CodecContext.time_base.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pyx":554
 * 
 *     property codec_tag:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.ptr.codec_tag.to_bytes(4, byteorder="little", signed=False).decode(
 *                 encoding="ascii")
 */

/* Python wrapper */
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_9codec_tag_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_9codec_tag_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2av_5codec_7context_12CodecContext_9codec_tag___get__(((struct __pyx_obj_2av_5codec_7context_CodecContext *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_9codec_tag___get__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "av/codec/context.pyx":555
 *     property codec_tag:
 *         def __get__(self):
 *             return self.ptr.codec_tag.to_bytes(4, byteorder="little", signed=False).decode(             # <<<<<<<<<<<<<<
 *                 encoding="ascii")
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_int(__pyx_v_self->ptr->codec_tag); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 555, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_to_bytes); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 555, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 555, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_byteorder, __pyx_n_s_little) < 0) __PYX_ERR(0, 555, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_signed, Py_False) < 0) __PYX_ERR(0, 555, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__10, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 555, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_decode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 555, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "av/codec/context.pyx":556
 *         def __get__(self):
 *             return self.ptr.codec_tag.to_bytes(4, byteorder="little", signed=False).decode(
 *                 encoding="ascii")             # <<<<<<<<<<<<<<
 * 
 *         def __set__(self, value):
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 556, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_encoding, __pyx_n_s_ascii) < 0) __PYX_ERR(0, 556, __pyx_L1_error)

  /* "av/codec/context.pyx":555
 *     property codec_tag:
 *         def __get__(self):
 *             return self.ptr.codec_tag.to_bytes(4, byteorder="little", signed=False).decode(             # <<<<<<<<<<<<<<
 *                 encoding="ascii")
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 555, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "av/codec/context.pyx":554
 * 
 *     property codec_tag:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.ptr.codec_tag.to_bytes(4, byteorder="little", signed=False).decode(
 *                 encoding="ascii")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("av.codec.context.CodecContext.codec_tag.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pyx":558
 *                 encoding="ascii")
 * 
 *         def __set__(self, value):             # <<<<<<<<<<<<<<
 *             if isinstance(value, str) and len(value) == 4:
 *                 self.ptr.codec_tag = int.from_bytes(value.encode(encoding="ascii"),
 */

/* Python wrapper */
static int __pyx_pw_2av_5codec_7context_12CodecContext_9codec_tag_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_2av_5codec_7context_12CodecContext_9codec_tag_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2av_5codec_7context_12CodecContext_9codec_tag_2__set__(((struct __pyx_obj_2av_5codec_7context_CodecContext *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2av_5codec_7context_12CodecContext_9codec_tag_2__set__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  unsigned int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "av/codec/context.pyx":559
 * 
 *         def __set__(self, value):
 *             if isinstance(value, str) and len(value) == 4:             # <<<<<<<<<<<<<<
 *                 self.ptr.codec_tag = int.from_bytes(value.encode(encoding="ascii"),
 *                                                     byteorder="little", signed=False)
 */
  __pyx_t_2 = PyString_Check(__pyx_v_value); 
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = PyObject_Length(__pyx_v_value); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 559, __pyx_L1_error)
  __pyx_t_3 = ((__pyx_t_4 == 4) != 0);
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (likely(__pyx_t_1)) {

    /* "av/codec/context.pyx":560
 *         def __set__(self, value):
 *             if isinstance(value, str) and len(value) == 4:
 *                 self.ptr.codec_tag = int.from_bytes(value.encode(encoding="ascii"),             # <<<<<<<<<<<<<<
 *                                                     byteorder="little", signed=False)
 *             else:
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)(&PyInt_Type)), __pyx_n_s_from_bytes); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 560, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_encode); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 560, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 560, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_encoding, __pyx_n_s_ascii) < 0) __PYX_ERR(0, 560, __pyx_L1_error)
    __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_empty_tuple, __pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 560, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 560, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_8);
    __pyx_t_8 = 0;

    /* "av/codec/context.pyx":561
 *             if isinstance(value, str) and len(value) == 4:
 *                 self.ptr.codec_tag = int.from_bytes(value.encode(encoding="ascii"),
 *                                                     byteorder="little", signed=False)             # <<<<<<<<<<<<<<
 *             else:
 *                 raise ValueError("Codec tag should be a 4 character string.")
 */
    __pyx_t_8 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 561, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_byteorder, __pyx_n_s_little) < 0) __PYX_ERR(0, 561, __pyx_L1_error)
    if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_signed, Py_False) < 0) __PYX_ERR(0, 561, __pyx_L1_error)

    /* "av/codec/context.pyx":560
 *         def __set__(self, value):
 *             if isinstance(value, str) and len(value) == 4:
 *                 self.ptr.codec_tag = int.from_bytes(value.encode(encoding="ascii"),             # <<<<<<<<<<<<<<
 *                                                     byteorder="little", signed=False)
 *             else:
 */
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_7, __pyx_t_8); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 560, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_unsigned_int(__pyx_t_6); if (unlikely((__pyx_t_9 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 560, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_v_self->ptr->codec_tag = __pyx_t_9;

    /* "av/codec/context.pyx":559
 * 
 *         def __set__(self, value):
 *             if isinstance(value, str) and len(value) == 4:             # <<<<<<<<<<<<<<
 *                 self.ptr.codec_tag = int.from_bytes(value.encode(encoding="ascii"),
 *                                                     byteorder="little", signed=False)
 */
    goto __pyx_L3;
  }

  /* "av/codec/context.pyx":563
 *                                                     byteorder="little", signed=False)
 *             else:
 *                 raise ValueError("Codec tag should be a 4 character string.")             # <<<<<<<<<<<<<<
 * 
 *     property ticks_per_frame:
 */
  /*else*/ {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 563, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 563, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "av/codec/context.pyx":558
 *                 encoding="ascii")
 * 
 *         def __set__(self, value):             # <<<<<<<<<<<<<<
 *             if isinstance(value, str) and len(value) == 4:
 *                 self.ptr.codec_tag = int.from_bytes(value.encode(encoding="ascii"),
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("av.codec.context.CodecContext.codec_tag.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pyx":566
 * 
 *     property ticks_per_frame:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.ptr.ticks_per_frame
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_15ticks_per_frame_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_15ticks_per_frame_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2av_5codec_7context_12CodecContext_15ticks_per_frame___get__(((struct __pyx_obj_2av_5codec_7context_CodecContext *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_15ticks_per_frame___get__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "av/codec/context.pyx":567
 *     property ticks_per_frame:
 *         def __get__(self):
 *             return self.ptr.ticks_per_frame             # <<<<<<<<<<<<<<
 * 
 *     property bit_rate:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->ptr->ticks_per_frame); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "av/codec/context.pyx":566
 * 
 *     property ticks_per_frame:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.ptr.ticks_per_frame
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("av.codec.context.CodecContext.ticks_per_frame.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pyx":570
 * 
 *     property bit_rate:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.ptr.bit_rate if self.ptr.bit_rate > 0 else None
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_8bit_rate_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_8bit_rate_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2av_5codec_7context_12CodecContext_8bit_rate___get__(((struct __pyx_obj_2av_5codec_7context_CodecContext *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_8bit_rate___get__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "av/codec/context.pyx":571
 *     property bit_rate:
 *         def __get__(self):
 *             return self.ptr.bit_rate if self.ptr.bit_rate > 0 else None             # <<<<<<<<<<<<<<
 * 
 *         def __set__(self, int value):
 */
  __Pyx_XDECREF(__pyx_r);
  if (((__pyx_v_self->ptr->bit_rate > 0) != 0)) {
    __pyx_t_2 = __Pyx_PyInt_From_int64_t(__pyx_v_self->ptr->bit_rate); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 571, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __pyx_t_2;
    __pyx_t_2 = 0;
  } else {
    __Pyx_INCREF(Py_None);
    __pyx_t_1 = Py_None;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "av/codec/context.pyx":570
 * 
 *     property bit_rate:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.ptr.bit_rate if self.ptr.bit_rate > 0 else None
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("av.codec.context.CodecContext.bit_rate.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pyx":573
 *             return self.ptr.bit_rate if self.ptr.bit_rate > 0 else None
 * 
 *         def __set__(self, int value):             # <<<<<<<<<<<<<<
 *             self.ptr.bit_rate = value
 * 
 */

/* Python wrapper */
static int __pyx_pw_2av_5codec_7context_12CodecContext_8bit_rate_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value); /*proto*/
static int __pyx_pw_2av_5codec_7context_12CodecContext_8bit_rate_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value) {
  int __pyx_v_value;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  assert(__pyx_arg_value); {
    __pyx_v_value = __Pyx_PyInt_As_int(__pyx_arg_value); if (unlikely((__pyx_v_value == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 573, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("av.codec.context.CodecContext.bit_rate.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_2av_5codec_7context_12CodecContext_8bit_rate_2__set__(((struct __pyx_obj_2av_5codec_7context_CodecContext *)__pyx_v_self), ((int)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2av_5codec_7context_12CodecContext_8bit_rate_2__set__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, int __pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "av/codec/context.pyx":574
 * 
 *         def __set__(self, int value):
 *             self.ptr.bit_rate = value             # <<<<<<<<<<<<<<
 * 
 *     property max_bit_rate:
 */
  __pyx_v_self->ptr->bit_rate = __pyx_v_value;

  /* "av/codec/context.pyx":573
 *             return self.ptr.bit_rate if self.ptr.bit_rate > 0 else None
 * 
 *         def __set__(self, int value):             # <<<<<<<<<<<<<<
 *             self.ptr.bit_rate = value
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pyx":577
 * 
 *     property max_bit_rate:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             if self.ptr.rc_max_rate > 0:
 *                 return self.ptr.rc_max_rate
 */

/* Python wrapper */
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_12max_bit_rate_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_12max_bit_rate_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2av_5codec_7context_12CodecContext_12max_bit_rate___get__(((struct __pyx_obj_2av_5codec_7context_CodecContext *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_12max_bit_rate___get__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "av/codec/context.pyx":578
 *     property max_bit_rate:
 *         def __get__(self):
 *             if self.ptr.rc_max_rate > 0:             # <<<<<<<<<<<<<<
 *                 return self.ptr.rc_max_rate
 *             else:
 */
  __pyx_t_1 = ((__pyx_v_self->ptr->rc_max_rate > 0) != 0);
  if (__pyx_t_1) {

    /* "av/codec/context.pyx":579
 *         def __get__(self):
 *             if self.ptr.rc_max_rate > 0:
 *                 return self.ptr.rc_max_rate             # <<<<<<<<<<<<<<
 *             else:
 *                 return None
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->ptr->rc_max_rate); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 579, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "av/codec/context.pyx":578
 *     property max_bit_rate:
 *         def __get__(self):
 *             if self.ptr.rc_max_rate > 0:             # <<<<<<<<<<<<<<
 *                 return self.ptr.rc_max_rate
 *             else:
 */
  }

  /* "av/codec/context.pyx":581
 *                 return self.ptr.rc_max_rate
 *             else:
 *                 return None             # <<<<<<<<<<<<<<
 * 
 *     property bit_rate_tolerance:
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;
  }

  /* "av/codec/context.pyx":577
 * 
 *     property max_bit_rate:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             if self.ptr.rc_max_rate > 0:
 *                 return self.ptr.rc_max_rate
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("av.codec.context.CodecContext.max_bit_rate.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pyx":584
 * 
 *     property bit_rate_tolerance:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             self.ptr.bit_rate_tolerance
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_18bit_rate_tolerance_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_18bit_rate_tolerance_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2av_5codec_7context_12CodecContext_18bit_rate_tolerance___get__(((struct __pyx_obj_2av_5codec_7context_CodecContext *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_18bit_rate_tolerance___get__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "av/codec/context.pyx":585
 *     property bit_rate_tolerance:
 *         def __get__(self):
 *             self.ptr.bit_rate_tolerance             # <<<<<<<<<<<<<<
 * 
 *         def __set__(self, int value):
 */
  (void)(__pyx_v_self->ptr->bit_rate_tolerance);

  /* "av/codec/context.pyx":584
 * 
 *     property bit_rate_tolerance:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             self.ptr.bit_rate_tolerance
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pyx":587
 *             self.ptr.bit_rate_tolerance
 * 
 *         def __set__(self, int value):             # <<<<<<<<<<<<<<
 *             self.ptr.bit_rate_tolerance = value
 * 
 */

/* Python wrapper */
static int __pyx_pw_2av_5codec_7context_12CodecContext_18bit_rate_tolerance_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value); /*proto*/
static int __pyx_pw_2av_5codec_7context_12CodecContext_18bit_rate_tolerance_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value) {
  int __pyx_v_value;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  assert(__pyx_arg_value); {
    __pyx_v_value = __Pyx_PyInt_As_int(__pyx_arg_value); if (unlikely((__pyx_v_value == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 587, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("av.codec.context.CodecContext.bit_rate_tolerance.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_2av_5codec_7context_12CodecContext_18bit_rate_tolerance_2__set__(((struct __pyx_obj_2av_5codec_7context_CodecContext *)__pyx_v_self), ((int)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2av_5codec_7context_12CodecContext_18bit_rate_tolerance_2__set__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, int __pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "av/codec/context.pyx":588
 * 
 *         def __set__(self, int value):
 *             self.ptr.bit_rate_tolerance = value             # <<<<<<<<<<<<<<
 * 
 *     property thread_count:
 */
  __pyx_v_self->ptr->bit_rate_tolerance = __pyx_v_value;

  /* "av/codec/context.pyx":587
 *             self.ptr.bit_rate_tolerance
 * 
 *         def __set__(self, int value):             # <<<<<<<<<<<<<<
 *             self.ptr.bit_rate_tolerance = value
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pyx":597
 *         """
 * 
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.ptr.thread_count
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_12thread_count_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_12thread_count_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2av_5codec_7context_12CodecContext_12thread_count___get__(((struct __pyx_obj_2av_5codec_7context_CodecContext *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_12thread_count___get__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "av/codec/context.pyx":598
 * 
 *         def __get__(self):
 *             return self.ptr.thread_count             # <<<<<<<<<<<<<<
 * 
 *         def __set__(self, int value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->ptr->thread_count); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 598, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "av/codec/context.pyx":597
 *         """
 * 
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.ptr.thread_count
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("av.codec.context.CodecContext.thread_count.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pyx":600
 *             return self.ptr.thread_count
 * 
 *         def __set__(self, int value):             # <<<<<<<<<<<<<<
 *             if lib.avcodec_is_open(self.ptr):
 *                 raise RuntimeError("Cannot change thread_count after codec is open.")
 */

/* Python wrapper */
static int __pyx_pw_2av_5codec_7context_12CodecContext_12thread_count_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value); /*proto*/
static int __pyx_pw_2av_5codec_7context_12CodecContext_12thread_count_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value) {
  int __pyx_v_value;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  assert(__pyx_arg_value); {
    __pyx_v_value = __Pyx_PyInt_As_int(__pyx_arg_value); if (unlikely((__pyx_v_value == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 600, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("av.codec.context.CodecContext.thread_count.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_2av_5codec_7context_12CodecContext_12thread_count_2__set__(((struct __pyx_obj_2av_5codec_7context_CodecContext *)__pyx_v_self), ((int)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2av_5codec_7context_12CodecContext_12thread_count_2__set__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, int __pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "av/codec/context.pyx":601
 * 
 *         def __set__(self, int value):
 *             if lib.avcodec_is_open(self.ptr):             # <<<<<<<<<<<<<<
 *                 raise RuntimeError("Cannot change thread_count after codec is open.")
 *             self.ptr.thread_count = value
 */
  __pyx_t_1 = (avcodec_is_open(__pyx_v_self->ptr) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "av/codec/context.pyx":602
 *         def __set__(self, int value):
 *             if lib.avcodec_is_open(self.ptr):
 *                 raise RuntimeError("Cannot change thread_count after codec is open.")             # <<<<<<<<<<<<<<
 *             self.ptr.thread_count = value
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 602, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 602, __pyx_L1_error)

    /* "av/codec/context.pyx":601
 * 
 *         def __set__(self, int value):
 *             if lib.avcodec_is_open(self.ptr):             # <<<<<<<<<<<<<<
 *                 raise RuntimeError("Cannot change thread_count after codec is open.")
 *             self.ptr.thread_count = value
 */
  }

  /* "av/codec/context.pyx":603
 *             if lib.avcodec_is_open(self.ptr):
 *                 raise RuntimeError("Cannot change thread_count after codec is open.")
 *             self.ptr.thread_count = value             # <<<<<<<<<<<<<<
 * 
 *     property thread_type:
 */
  __pyx_v_self->ptr->thread_count = __pyx_v_value;

  /* "av/codec/context.pyx":600
 *             return self.ptr.thread_count
 * 
 *         def __set__(self, int value):             # <<<<<<<<<<<<<<
 *             if lib.avcodec_is_open(self.ptr):
 *                 raise RuntimeError("Cannot change thread_count after codec is open.")
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("av.codec.context.CodecContext.thread_count.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pyx":612
 *         """
 * 
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return ThreadType.get(self.ptr.thread_type, create=True)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_11thread_type_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_11thread_type_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2av_5codec_7context_12CodecContext_11thread_type___get__(((struct __pyx_obj_2av_5codec_7context_CodecContext *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_11thread_type___get__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "av/codec/context.pyx":613
 * 
 *         def __get__(self):
 *             return ThreadType.get(self.ptr.thread_type, create=True)             # <<<<<<<<<<<<<<
 * 
 *         def __set__(self, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ThreadType); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_get); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->ptr->thread_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_create, Py_True) < 0) __PYX_ERR(0, 613, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "av/codec/context.pyx":612
 *         """
 * 
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return ThreadType.get(self.ptr.thread_type, create=True)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("av.codec.context.CodecContext.thread_type.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pyx":615
 *             return ThreadType.get(self.ptr.thread_type, create=True)
 * 
 *         def __set__(self, value):             # <<<<<<<<<<<<<<
 *             if lib.avcodec_is_open(self.ptr):
 *                 raise RuntimeError("Cannot change thread_type after codec is open.")
 */

/* Python wrapper */
static int __pyx_pw_2av_5codec_7context_12CodecContext_11thread_type_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_2av_5codec_7context_12CodecContext_11thread_type_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2av_5codec_7context_12CodecContext_11thread_type_2__set__(((struct __pyx_obj_2av_5codec_7context_CodecContext *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2av_5codec_7context_12CodecContext_11thread_type_2__set__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "av/codec/context.pyx":616
 * 
 *         def __set__(self, value):
 *             if lib.avcodec_is_open(self.ptr):             # <<<<<<<<<<<<<<
 *                 raise RuntimeError("Cannot change thread_type after codec is open.")
 *             self.ptr.thread_type = ThreadType[value].value
 */
  __pyx_t_1 = (avcodec_is_open(__pyx_v_self->ptr) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "av/codec/context.pyx":617
 *         def __set__(self, value):
 *             if lib.avcodec_is_open(self.ptr):
 *                 raise RuntimeError("Cannot change thread_type after codec is open.")             # <<<<<<<<<<<<<<
 *             self.ptr.thread_type = ThreadType[value].value
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 617, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 617, __pyx_L1_error)

    /* "av/codec/context.pyx":616
 * 
 *         def __set__(self, value):
 *             if lib.avcodec_is_open(self.ptr):             # <<<<<<<<<<<<<<
 *                 raise RuntimeError("Cannot change thread_type after codec is open.")
 *             self.ptr.thread_type = ThreadType[value].value
 */
  }

  /* "av/codec/context.pyx":618
 *             if lib.avcodec_is_open(self.ptr):
 *                 raise RuntimeError("Cannot change thread_type after codec is open.")
 *             self.ptr.thread_type = ThreadType[value].value             # <<<<<<<<<<<<<<
 * 
 *     property skip_frame:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ThreadType); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 618, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_t_2, __pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 618, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 618, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 618, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_self->ptr->thread_type = __pyx_t_4;

  /* "av/codec/context.pyx":615
 *             return ThreadType.get(self.ptr.thread_type, create=True)
 * 
 *         def __set__(self, value):             # <<<<<<<<<<<<<<
 *             if lib.avcodec_is_open(self.ptr):
 *                 raise RuntimeError("Cannot change thread_type after codec is open.")
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("av.codec.context.CodecContext.thread_type.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pyx":627
 *         """
 * 
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return SkipType._get(self.ptr.skip_frame, create=True)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_10skip_frame_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_10skip_frame_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2av_5codec_7context_12CodecContext_10skip_frame___get__(((struct __pyx_obj_2av_5codec_7context_CodecContext *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_10skip_frame___get__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "av/codec/context.pyx":628
 * 
 *         def __get__(self):
 *             return SkipType._get(self.ptr.skip_frame, create=True)             # <<<<<<<<<<<<<<
 * 
 *         def __set__(self, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_SkipType); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_get_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_enum__AVDiscard(__pyx_v_self->ptr->skip_frame); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_create, Py_True) < 0) __PYX_ERR(0, 628, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "av/codec/context.pyx":627
 *         """
 * 
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return SkipType._get(self.ptr.skip_frame, create=True)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("av.codec.context.CodecContext.skip_frame.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pyx":630
 *             return SkipType._get(self.ptr.skip_frame, create=True)
 * 
 *         def __set__(self, value):             # <<<<<<<<<<<<<<
 *             self.ptr.skip_frame = SkipType[value].value
 */

/* Python wrapper */
static int __pyx_pw_2av_5codec_7context_12CodecContext_10skip_frame_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_2av_5codec_7context_12CodecContext_10skip_frame_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2av_5codec_7context_12CodecContext_10skip_frame_2__set__(((struct __pyx_obj_2av_5codec_7context_CodecContext *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2av_5codec_7context_12CodecContext_10skip_frame_2__set__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  enum AVDiscard __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "av/codec/context.pyx":631
 * 
 *         def __set__(self, value):
 *             self.ptr.skip_frame = SkipType[value].value             # <<<<<<<<<<<<<<
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_SkipType); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = ((enum AVDiscard)__Pyx_PyInt_As_enum__AVDiscard(__pyx_t_1)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 631, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->ptr->skip_frame = __pyx_t_3;

  /* "av/codec/context.pyx":630
 *             return SkipType._get(self.ptr.skip_frame, create=True)
 * 
 *         def __set__(self, value):             # <<<<<<<<<<<<<<
 *             self.ptr.skip_frame = SkipType[value].value
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("av.codec.context.CodecContext.skip_frame.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pxd":26
 *     cdef _init(self, lib.AVCodecContext *ptr, const lib.AVCodec *codec)
 * 
 *     cdef readonly Codec codec             # <<<<<<<<<<<<<<
 * 
 *     cdef public dict options
 */

/* Python wrapper */
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_5codec_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_5codec_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2av_5codec_7context_12CodecContext_5codec___get__(((struct __pyx_obj_2av_5codec_7context_CodecContext *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_5codec___get__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->codec));
  __pyx_r = ((PyObject *)__pyx_v_self->codec);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "av/codec/context.pxd":28
 *     cdef readonly Codec codec
 * 
 *     cdef public dict options             # <<<<<<<<<<<<<<
 * 
 *     # Public API.
 */

/* Python wrapper */
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_7options_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_7options_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2av_5codec_7context_12CodecContext_7options___get__(((struct __pyx_obj_2av_5codec_7context_CodecContext *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_7options___get__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->options);
  __pyx_r = __pyx_v_self->options;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2av_5codec_7context_12CodecContext_7options_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_2av_5codec_7context_12CodecContext_7options_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2av_5codec_7context_12CodecContext_7options_2__set__(((struct __pyx_obj_2av_5codec_7context_CodecContext *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2av_5codec_7context_12CodecContext_7options_2__set__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(PyDict_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(2, 28, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->options);
  __Pyx_DECREF(__pyx_v_self->options);
  __pyx_v_self->options = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("av.codec.context.CodecContext.options.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2av_5codec_7context_12CodecContext_7options_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_2av_5codec_7context_12CodecContext_7options_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2av_5codec_7context_12CodecContext_7options_4__del__(((struct __pyx_obj_2av_5codec_7context_CodecContext *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2av_5codec_7context_12CodecContext_7options_4__del__(struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->options);
  __Pyx_DECREF(__pyx_v_self->options);
  __pyx_v_self->options = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_27__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_2av_5codec_7context_12CodecContext_26__reduce_cython__[] = "CodecContext.__reduce_cython__()";
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_27__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2av_5codec_7context_12CodecContext_26__reduce_cython__(((struct __pyx_obj_2av_5codec_7context_CodecContext *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_26__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__14, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("av.codec.context.CodecContext.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_29__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static char __pyx_doc_2av_5codec_7context_12CodecContext_28__setstate_cython__[] = "CodecContext.__setstate_cython__(__pyx_state)";
static PyObject *__pyx_pw_2av_5codec_7context_12CodecContext_29__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2av_5codec_7context_12CodecContext_28__setstate_cython__(((struct __pyx_obj_2av_5codec_7context_CodecContext *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2av_5codec_7context_12CodecContext_28__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2av_5codec_7context_CodecContext *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("av.codec.context.CodecContext.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_2av_5codec_7context_CodecContext __pyx_vtable_2av_5codec_7context_CodecContext;

static PyObject *__pyx_tp_new_2av_5codec_7context_CodecContext(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_2av_5codec_7context_CodecContext *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_2av_5codec_7context_CodecContext *)o);
  p->__pyx_vtab = __pyx_vtabptr_2av_5codec_7context_CodecContext;
  p->codec = ((struct __pyx_obj_2av_5codec_5codec_Codec *)Py_None); Py_INCREF(Py_None);
  p->options = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_next_frame = ((struct __pyx_obj_2av_5frame_Frame *)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_2av_5codec_7context_12CodecContext_3__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_2av_5codec_7context_CodecContext(PyObject *o) {
  struct __pyx_obj_2av_5codec_7context_CodecContext *p = (struct __pyx_obj_2av_5codec_7context_CodecContext *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_2av_5codec_7context_12CodecContext_17__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->codec);
  Py_CLEAR(p->options);
  Py_CLEAR(p->_next_frame);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_2av_5codec_7context_CodecContext(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_2av_5codec_7context_CodecContext *p = (struct __pyx_obj_2av_5codec_7context_CodecContext *)o;
  if (p->codec) {
    e = (*v)(((PyObject *)p->codec), a); if (e) return e;
  }
  if (p->options) {
    e = (*v)(p->options, a); if (e) return e;
  }
  if (p->_next_frame) {
    e = (*v)(((PyObject *)p->_next_frame), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_2av_5codec_7context_CodecContext(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_2av_5codec_7context_CodecContext *p = (struct __pyx_obj_2av_5codec_7context_CodecContext *)o;
  tmp = ((PyObject*)p->codec);
  p->codec = ((struct __pyx_obj_2av_5codec_5codec_Codec *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->options);
  p->options = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_next_frame);
  p->_next_frame = ((struct __pyx_obj_2av_5frame_Frame *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_2av_5codec_7context_12CodecContext_extradata(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2av_5codec_7context_12CodecContext_9extradata_1__get__(o);
}

static int __pyx_setprop_2av_5codec_7context_12CodecContext_extradata(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_2av_5codec_7context_12CodecContext_9extradata_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_2av_5codec_7context_12CodecContext_extradata_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2av_5codec_7context_12CodecContext_14extradata_size_1__get__(o);
}

static PyObject *__pyx_getprop_2av_5codec_7context_12CodecContext_is_open(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2av_5codec_7context_12CodecContext_7is_open_1__get__(o);
}

static PyObject *__pyx_getprop_2av_5codec_7context_12CodecContext_is_encoder(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2av_5codec_7context_12CodecContext_10is_encoder_1__get__(o);
}

static PyObject *__pyx_getprop_2av_5codec_7context_12CodecContext_is_decoder(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2av_5codec_7context_12CodecContext_10is_decoder_1__get__(o);
}

static PyObject *__pyx_getprop_2av_5codec_7context_12CodecContext_name(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2av_5codec_7context_12CodecContext_4name_1__get__(o);
}

static PyObject *__pyx_getprop_2av_5codec_7context_12CodecContext_type(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2av_5codec_7context_12CodecContext_4type_1__get__(o);
}

static PyObject *__pyx_getprop_2av_5codec_7context_12CodecContext_profile(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2av_5codec_7context_12CodecContext_7profile_1__get__(o);
}

static PyObject *__pyx_getprop_2av_5codec_7context_12CodecContext_time_base(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2av_5codec_7context_12CodecContext_9time_base_1__get__(o);
}

static int __pyx_setprop_2av_5codec_7context_12CodecContext_time_base(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_2av_5codec_7context_12CodecContext_9time_base_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_2av_5codec_7context_12CodecContext_codec_tag(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2av_5codec_7context_12CodecContext_9codec_tag_1__get__(o);
}

static int __pyx_setprop_2av_5codec_7context_12CodecContext_codec_tag(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_2av_5codec_7context_12CodecContext_9codec_tag_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_2av_5codec_7context_12CodecContext_ticks_per_frame(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2av_5codec_7context_12CodecContext_15ticks_per_frame_1__get__(o);
}

static PyObject *__pyx_getprop_2av_5codec_7context_12CodecContext_bit_rate(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2av_5codec_7context_12CodecContext_8bit_rate_1__get__(o);
}

static int __pyx_setprop_2av_5codec_7context_12CodecContext_bit_rate(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_2av_5codec_7context_12CodecContext_8bit_rate_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_2av_5codec_7context_12CodecContext_max_bit_rate(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2av_5codec_7context_12CodecContext_12max_bit_rate_1__get__(o);
}

static PyObject *__pyx_getprop_2av_5codec_7context_12CodecContext_bit_rate_tolerance(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2av_5codec_7context_12CodecContext_18bit_rate_tolerance_1__get__(o);
}

static int __pyx_setprop_2av_5codec_7context_12CodecContext_bit_rate_tolerance(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_2av_5codec_7context_12CodecContext_18bit_rate_tolerance_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_2av_5codec_7context_12CodecContext_thread_count(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2av_5codec_7context_12CodecContext_12thread_count_1__get__(o);
}

static int __pyx_setprop_2av_5codec_7context_12CodecContext_thread_count(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_2av_5codec_7context_12CodecContext_12thread_count_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_2av_5codec_7context_12CodecContext_thread_type(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2av_5codec_7context_12CodecContext_11thread_type_1__get__(o);
}

static int __pyx_setprop_2av_5codec_7context_12CodecContext_thread_type(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_2av_5codec_7context_12CodecContext_11thread_type_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_2av_5codec_7context_12CodecContext_skip_frame(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2av_5codec_7context_12CodecContext_10skip_frame_1__get__(o);
}

static int __pyx_setprop_2av_5codec_7context_12CodecContext_skip_frame(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_2av_5codec_7context_12CodecContext_10skip_frame_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_2av_5codec_7context_12CodecContext_codec(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2av_5codec_7context_12CodecContext_5codec_1__get__(o);
}

static PyObject *__pyx_getprop_2av_5codec_7context_12CodecContext_options(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2av_5codec_7context_12CodecContext_7options_1__get__(o);
}

static int __pyx_setprop_2av_5codec_7context_12CodecContext_options(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_2av_5codec_7context_12CodecContext_7options_3__set__(o, v);
  }
  else {
    return __pyx_pw_2av_5codec_7context_12CodecContext_7options_5__del__(o);
  }
}

static PyMethodDef __pyx_methods_2av_5codec_7context_CodecContext[] = {
  {"create", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_2av_5codec_7context_12CodecContext_1create, METH_VARARGS|METH_KEYWORDS, __pyx_doc_2av_5codec_7context_12CodecContext_create},
  {"_get_flags", (PyCFunction)__pyx_pw_2av_5codec_7context_12CodecContext_5_get_flags, METH_NOARGS, __pyx_doc_2av_5codec_7context_12CodecContext_4_get_flags},
  {"_set_flags", (PyCFunction)__pyx_pw_2av_5codec_7context_12CodecContext_7_set_flags, METH_O, __pyx_doc_2av_5codec_7context_12CodecContext_6_set_flags},
  {"_get_flags2", (PyCFunction)__pyx_pw_2av_5codec_7context_12CodecContext_9_get_flags2, METH_NOARGS, __pyx_doc_2av_5codec_7context_12CodecContext_8_get_flags2},
  {"_set_flags2", (PyCFunction)__pyx_pw_2av_5codec_7context_12CodecContext_11_set_flags2, METH_O, __pyx_doc_2av_5codec_7context_12CodecContext_10_set_flags2},
  {"open", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_2av_5codec_7context_12CodecContext_13open, METH_VARARGS|METH_KEYWORDS, __pyx_doc_2av_5codec_7context_12CodecContext_12open},
  {"close", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_2av_5codec_7context_12CodecContext_15close, METH_VARARGS|METH_KEYWORDS, __pyx_doc_2av_5codec_7context_12CodecContext_14close},
  {"parse", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_2av_5codec_7context_12CodecContext_21parse, METH_VARARGS|METH_KEYWORDS, __pyx_doc_2av_5codec_7context_12CodecContext_20parse},
  {"encode", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_2av_5codec_7context_12CodecContext_23encode, METH_VARARGS|METH_KEYWORDS, __pyx_doc_2av_5codec_7context_12CodecContext_22encode},
  {"decode", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_2av_5codec_7context_12CodecContext_25decode, METH_VARARGS|METH_KEYWORDS, __pyx_doc_2av_5codec_7context_12CodecContext_24decode},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_2av_5codec_7context_12CodecContext_27__reduce_cython__, METH_NOARGS, __pyx_doc_2av_5codec_7context_12CodecContext_26__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_2av_5codec_7context_12CodecContext_29__setstate_cython__, METH_O, __pyx_doc_2av_5codec_7context_12CodecContext_28__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_2av_5codec_7context_CodecContext[] = {
  {(char *)"extradata", __pyx_getprop_2av_5codec_7context_12CodecContext_extradata, __pyx_setprop_2av_5codec_7context_12CodecContext_extradata, (char *)0, 0},
  {(char *)"extradata_size", __pyx_getprop_2av_5codec_7context_12CodecContext_extradata_size, 0, (char *)0, 0},
  {(char *)"is_open", __pyx_getprop_2av_5codec_7context_12CodecContext_is_open, 0, (char *)0, 0},
  {(char *)"is_encoder", __pyx_getprop_2av_5codec_7context_12CodecContext_is_encoder, 0, (char *)0, 0},
  {(char *)"is_decoder", __pyx_getprop_2av_5codec_7context_12CodecContext_is_decoder, 0, (char *)0, 0},
  {(char *)"name", __pyx_getprop_2av_5codec_7context_12CodecContext_name, 0, (char *)0, 0},
  {(char *)"type", __pyx_getprop_2av_5codec_7context_12CodecContext_type, 0, (char *)0, 0},
  {(char *)"profile", __pyx_getprop_2av_5codec_7context_12CodecContext_profile, 0, (char *)0, 0},
  {(char *)"time_base", __pyx_getprop_2av_5codec_7context_12CodecContext_time_base, __pyx_setprop_2av_5codec_7context_12CodecContext_time_base, (char *)0, 0},
  {(char *)"codec_tag", __pyx_getprop_2av_5codec_7context_12CodecContext_codec_tag, __pyx_setprop_2av_5codec_7context_12CodecContext_codec_tag, (char *)0, 0},
  {(char *)"ticks_per_frame", __pyx_getprop_2av_5codec_7context_12CodecContext_ticks_per_frame, 0, (char *)0, 0},
  {(char *)"bit_rate", __pyx_getprop_2av_5codec_7context_12CodecContext_bit_rate, __pyx_setprop_2av_5codec_7context_12CodecContext_bit_rate, (char *)0, 0},
  {(char *)"max_bit_rate", __pyx_getprop_2av_5codec_7context_12CodecContext_max_bit_rate, 0, (char *)0, 0},
  {(char *)"bit_rate_tolerance", __pyx_getprop_2av_5codec_7context_12CodecContext_bit_rate_tolerance, __pyx_setprop_2av_5codec_7context_12CodecContext_bit_rate_tolerance, (char *)0, 0},
  {(char *)"thread_count", __pyx_getprop_2av_5codec_7context_12CodecContext_thread_count, __pyx_setprop_2av_5codec_7context_12CodecContext_thread_count, (char *)"How many threads to use; 0 means auto.\n\n        Wraps :ffmpeg:`AVCodecContext.thread_count`.\n\n        ", 0},
  {(char *)"thread_type", __pyx_getprop_2av_5codec_7context_12CodecContext_thread_type, __pyx_setprop_2av_5codec_7context_12CodecContext_thread_type, (char *)"One of :class:`.ThreadType`.\n\n        Wraps :ffmpeg:`AVCodecContext.thread_type`.\n\n        ", 0},
  {(char *)"skip_frame", __pyx_getprop_2av_5codec_7context_12CodecContext_skip_frame, __pyx_setprop_2av_5codec_7context_12CodecContext_skip_frame, (char *)"One of :class:`.SkipType`.\n\n        Wraps ffmpeg:`AVCodecContext.skip_frame`.\n\n        ", 0},
  {(char *)"codec", __pyx_getprop_2av_5codec_7context_12CodecContext_codec, 0, (char *)0, 0},
  {(char *)"options", __pyx_getprop_2av_5codec_7context_12CodecContext_options, __pyx_setprop_2av_5codec_7context_12CodecContext_options, (char *)"options: dict", 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_2av_5codec_7context_CodecContext = {
  PyVarObject_HEAD_INIT(0, 0)
  "av.codec.context.CodecContext", /*tp_name*/
  sizeof(struct __pyx_obj_2av_5codec_7context_CodecContext), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_2av_5codec_7context_CodecContext, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_2av_5codec_7context_12CodecContext_19__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_2av_5codec_7context_CodecContext, /*tp_traverse*/
  __pyx_tp_clear_2av_5codec_7context_CodecContext, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_2av_5codec_7context_CodecContext, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_2av_5codec_7context_CodecContext, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_2av_5codec_7context_CodecContext, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_context(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_context},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "context",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_s_4MV, __pyx_k_4MV, sizeof(__pyx_k_4MV), 0, 0, 1, 0},
  {&__pyx_kp_s_4_MV_per_MB_allowed_advanced_pre, __pyx_k_4_MV_per_MB_allowed_advanced_pre, sizeof(__pyx_k_4_MV_per_MB_allowed_advanced_pre), 0, 0, 1, 0},
  {&__pyx_n_s_AC_PRED, __pyx_k_AC_PRED, sizeof(__pyx_k_AC_PRED), 0, 0, 1, 1},
  {&__pyx_n_s_ALL, __pyx_k_ALL, sizeof(__pyx_k_ALL), 0, 0, 1, 1},
  {&__pyx_n_s_AUTO, __pyx_k_AUTO, sizeof(__pyx_k_AUTO), 0, 0, 1, 1},
  {&__pyx_n_s_AVDeprecationWarning, __pyx_k_AVDeprecationWarning, sizeof(__pyx_k_AVDeprecationWarning), 0, 0, 1, 1},
  {&__pyx_kp_s_Allow_decoders_to_produce_frames, __pyx_k_Allow_decoders_to_produce_frames, sizeof(__pyx_k_Allow_decoders_to_produce_frames), 0, 0, 1, 0},
  {&__pyx_kp_s_Allow_non_spec_compliant_speedup, __pyx_k_Allow_non_spec_compliant_speedup, sizeof(__pyx_k_Allow_non_spec_compliant_speedup), 0, 0, 1, 0},
  {&__pyx_n_s_AudioCodecContext, __pyx_k_AudioCodecContext, sizeof(__pyx_k_AudioCodecContext), 0, 0, 1, 1},
  {&__pyx_n_s_BIDIR, __pyx_k_BIDIR, sizeof(__pyx_k_BIDIR), 0, 0, 1, 1},
  {&__pyx_n_s_BITEXACT, __pyx_k_BITEXACT, sizeof(__pyx_k_BITEXACT), 0, 0, 1, 1},
  {&__pyx_kp_s_Base_CodecContext_cannot_decode, __pyx_k_Base_CodecContext_cannot_decode, sizeof(__pyx_k_Base_CodecContext_cannot_decode), 0, 0, 1, 0},
  {&__pyx_n_s_CHUNKS, __pyx_k_CHUNKS, sizeof(__pyx_k_CHUNKS), 0, 0, 1, 1},
  {&__pyx_n_s_CLOSED_GOP, __pyx_k_CLOSED_GOP, sizeof(__pyx_k_CLOSED_GOP), 0, 0, 1, 1},
  {&__pyx_kp_s_Can_only_set_extradata_for_decod, __pyx_k_Can_only_set_extradata_for_decod, sizeof(__pyx_k_Can_only_set_extradata_for_decod), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_allocate_extradata, __pyx_k_Cannot_allocate_extradata, sizeof(__pyx_k_Cannot_allocate_extradata), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_change_thread_count_after, __pyx_k_Cannot_change_thread_count_after, sizeof(__pyx_k_Cannot_change_thread_count_after), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_change_thread_type_after, __pyx_k_Cannot_change_thread_type_after, sizeof(__pyx_k_Cannot_change_thread_type_after), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_instantiate_CodecContext, __pyx_k_Cannot_instantiate_CodecContext, sizeof(__pyx_k_Cannot_instantiate_CodecContext), 0, 0, 1, 0},
  {&__pyx_n_s_CodecContext, __pyx_k_CodecContext, sizeof(__pyx_k_CodecContext), 0, 0, 1, 1},
  {&__pyx_kp_s_CodecContext_is_already_closed, __pyx_k_CodecContext_is_already_closed, sizeof(__pyx_k_CodecContext_is_already_closed), 0, 0, 1, 0},
  {&__pyx_kp_s_CodecContext_is_already_open, __pyx_k_CodecContext_is_already_open, sizeof(__pyx_k_CodecContext_is_already_open), 0, 0, 1, 0},
  {&__pyx_kp_s_Codec_tag_should_be_a_4_characte, __pyx_k_Codec_tag_should_be_a_4_characte, sizeof(__pyx_k_Codec_tag_should_be_a_4_characte), 0, 0, 1, 0},
  {&__pyx_n_s_DEFAULT, __pyx_k_DEFAULT, sizeof(__pyx_k_DEFAULT), 0, 0, 1, 1},
  {&__pyx_n_s_DROPCHANGED, __pyx_k_DROPCHANGED, sizeof(__pyx_k_DROPCHANGED), 0, 0, 1, 1},
  {&__pyx_kp_s_Decode_more_than_one_frame_at_on, __pyx_k_Decode_more_than_one_frame_at_on, sizeof(__pyx_k_Decode_more_than_one_frame_at_on), 0, 0, 1, 0},
  {&__pyx_kp_s_Decode_more_than_one_part_of_a_s, __pyx_k_Decode_more_than_one_part_of_a_s, sizeof(__pyx_k_Decode_more_than_one_part_of_a_s), 0, 0, 1, 0},
  {&__pyx_kp_s_Decode_using_both_FRAME_and_SLIC, __pyx_k_Decode_using_both_FRAME_and_SLIC, sizeof(__pyx_k_Decode_using_both_FRAME_and_SLIC), 0, 0, 1, 0},
  {&__pyx_n_s_Dictionary, __pyx_k_Dictionary, sizeof(__pyx_k_Dictionary), 0, 0, 1, 1},
  {&__pyx_kp_s_Discard_all, __pyx_k_Discard_all, sizeof(__pyx_k_Discard_all), 0, 0, 1, 0},
  {&__pyx_kp_s_Discard_all_bidirectional_frames, __pyx_k_Discard_all_bidirectional_frames, sizeof(__pyx_k_Discard_all_bidirectional_frames), 0, 0, 1, 0},
  {&__pyx_kp_s_Discard_all_frames_except_keyfra, __pyx_k_Discard_all_frames_except_keyfra, sizeof(__pyx_k_Discard_all_frames_except_keyfra), 0, 0, 1, 0},
  {&__pyx_kp_s_Discard_all_non_intra_frames, __pyx_k_Discard_all_non_intra_frames, sizeof(__pyx_k_Discard_all_non_intra_frames), 0, 0, 1, 0},
  {&__pyx_kp_s_Discard_all_non_reference, __pyx_k_Discard_all_non_reference, sizeof(__pyx_k_Discard_all_non_reference), 0, 0, 1, 0},
  {&__pyx_kp_s_Discard_cropping_information_fro, __pyx_k_Discard_cropping_information_fro, sizeof(__pyx_k_Discard_cropping_information_fro), 0, 0, 1, 0},
  {&__pyx_kp_s_Discard_nothing, __pyx_k_Discard_nothing, sizeof(__pyx_k_Discard_nothing), 0, 0, 1, 0},
  {&__pyx_kp_s_Discard_useless_packets_like_0_s, __pyx_k_Discard_useless_packets_like_0_s, sizeof(__pyx_k_Discard_useless_packets_like_0_s), 0, 0, 1, 0},
  {&__pyx_kp_s_Do_not_reset_ASS_ReadOrder_field, __pyx_k_Do_not_reset_ASS_ReadOrder_field, sizeof(__pyx_k_Do_not_reset_ASS_ReadOrder_field), 0, 0, 1, 0},
  {&__pyx_kp_s_Do_not_skip_samples_and_export_s, __pyx_k_Do_not_skip_samples_and_export_s, sizeof(__pyx_k_Do_not_skip_samples_and_export_s), 0, 0, 1, 0},
  {&__pyx_kp_s_Don_t_output_frames_whose_parame, __pyx_k_Don_t_output_frames_whose_parame, sizeof(__pyx_k_Don_t_output_frames_whose_parame), 0, 0, 1, 0},
  {&__pyx_n_s_EXPORT_MVS, __pyx_k_EXPORT_MVS, sizeof(__pyx_k_EXPORT_MVS), 0, 0, 1, 1},
  {&__pyx_kp_s_Encoding_is_only_supported_for_a, __pyx_k_Encoding_is_only_supported_for_a, sizeof(__pyx_k_Encoding_is_only_supported_for_a), 0, 0, 1, 0},
  {&__pyx_kp_s_Export_motion_vectors_through_fr, __pyx_k_Export_motion_vectors_through_fr, sizeof(__pyx_k_Export_motion_vectors_through_fr), 0, 0, 1, 0},
  {&__pyx_n_s_FAST, __pyx_k_FAST, sizeof(__pyx_k_FAST), 0, 0, 1, 1},
  {&__pyx_n_s_FRAME, __pyx_k_FRAME, sizeof(__pyx_k_FRAME), 0, 0, 1, 1},
  {&__pyx_kp_s_Flag_property_of_class_Flags, __pyx_k_Flag_property_of_class_Flags, sizeof(__pyx_k_Flag_property_of_class_Flags), 0, 0, 1, 0},
  {&__pyx_kp_s_Flag_property_of_class_Flags2, __pyx_k_Flag_property_of_class_Flags2, sizeof(__pyx_k_Flag_property_of_class_Flags2), 0, 0, 1, 0},
  {&__pyx_n_s_Flags, __pyx_k_Flags, sizeof(__pyx_k_Flags), 0, 0, 1, 1},
  {&__pyx_n_s_Flags2, __pyx_k_Flags2, sizeof(__pyx_k_Flags2), 0, 0, 1, 1},
  {&__pyx_kp_s_Force_low_delay, __pyx_k_Force_low_delay, sizeof(__pyx_k_Force_low_delay), 0, 0, 1, 0},
  {&__pyx_n_s_GLOBAL_HEADER, __pyx_k_GLOBAL_HEADER, sizeof(__pyx_k_GLOBAL_HEADER), 0, 0, 1, 1},
  {&__pyx_n_s_GRAY, __pyx_k_GRAY, sizeof(__pyx_k_GRAY), 0, 0, 1, 1},
  {&__pyx_kp_s_H_263_advanced_intra_coding_MPEG, __pyx_k_H_263_advanced_intra_coding_MPEG, sizeof(__pyx_k_H_263_advanced_intra_coding_MPEG), 0, 0, 1, 0},
  {&__pyx_n_s_IGNORE_CROP, __pyx_k_IGNORE_CROP, sizeof(__pyx_k_IGNORE_CROP), 0, 0, 1, 1},
  {&__pyx_n_s_INTERLACED_DCT, __pyx_k_INTERLACED_DCT, sizeof(__pyx_k_INTERLACED_DCT), 0, 0, 1, 1},
  {&__pyx_n_s_INTERLACED_ME, __pyx_k_INTERLACED_ME, sizeof(__pyx_k_INTERLACED_ME), 0, 0, 1, 1},
  {&__pyx_kp_s_Input_bitstream_might_be_truncat, __pyx_k_Input_bitstream_might_be_truncat, sizeof(__pyx_k_Input_bitstream_might_be_truncat), 0, 0, 1, 0},
  {&__pyx_kp_s_Interlaced_motion_estimation, __pyx_k_Interlaced_motion_estimation, sizeof(__pyx_k_Interlaced_motion_estimation), 0, 0, 1, 0},
  {&__pyx_n_s_LOCAL_HEADER, __pyx_k_LOCAL_HEADER, sizeof(__pyx_k_LOCAL_HEADER), 0, 0, 1, 1},
  {&__pyx_n_s_LOOP_FILTER, __pyx_k_LOOP_FILTER, sizeof(__pyx_k_LOOP_FILTER), 0, 0, 1, 1},
  {&__pyx_n_s_LOW_DELAY, __pyx_k_LOW_DELAY, sizeof(__pyx_k_LOW_DELAY), 0, 0, 1, 1},
  {&__pyx_n_s_MemoryError, __pyx_k_MemoryError, sizeof(__pyx_k_MemoryError), 0, 0, 1, 1},
  {&__pyx_n_s_NONE, __pyx_k_NONE, sizeof(__pyx_k_NONE), 0, 0, 1, 1},
  {&__pyx_n_s_NONINTRA, __pyx_k_NONINTRA, sizeof(__pyx_k_NONINTRA), 0, 0, 1, 1},
  {&__pyx_n_s_NONKEY, __pyx_k_NONKEY, sizeof(__pyx_k_NONKEY), 0, 0, 1, 1},
  {&__pyx_n_s_NONREF, __pyx_k_NONREF, sizeof(__pyx_k_NONREF), 0, 0, 1, 1},
  {&__pyx_n_s_NO_OUTPUT, __pyx_k_NO_OUTPUT, sizeof(__pyx_k_NO_OUTPUT), 0, 0, 1, 1},
  {&__pyx_kp_s_No_parser_for_s, __pyx_k_No_parser_for_s, sizeof(__pyx_k_No_parser_for_s), 0, 0, 1, 0},
  {&__pyx_n_s_NotImplementedError, __pyx_k_NotImplementedError, sizeof(__pyx_k_NotImplementedError), 0, 0, 1, 1},
  {&__pyx_n_s_OUTPUT_CORRUPT, __pyx_k_OUTPUT_CORRUPT, sizeof(__pyx_k_OUTPUT_CORRUPT), 0, 0, 1, 1},
  {&__pyx_kp_s_Only_decode_encode_grayscale, __pyx_k_Only_decode_encode_grayscale, sizeof(__pyx_k_Only_decode_encode_grayscale), 0, 0, 1, 0},
  {&__pyx_kp_s_Output_even_those_frames_that_mi, __pyx_k_Output_even_those_frames_that_mi, sizeof(__pyx_k_Output_even_those_frames_that_mi), 0, 0, 1, 0},
  {&__pyx_n_s_PASS1, __pyx_k_PASS1, sizeof(__pyx_k_PASS1), 0, 0, 1, 1},
  {&__pyx_n_s_PASS2, __pyx_k_PASS2, sizeof(__pyx_k_PASS2), 0, 0, 1, 1},
  {&__pyx_n_s_PSNR, __pyx_k_PSNR, sizeof(__pyx_k_PSNR), 0, 0, 1, 1},
  {&__pyx_kp_s_Place_global_headers_at_every_ke, __pyx_k_Place_global_headers_at_every_ke, sizeof(__pyx_k_Place_global_headers_at_every_ke), 0, 0, 1, 0},
  {&__pyx_kp_s_Place_global_headers_in_extradat, __pyx_k_Place_global_headers_in_extradat, sizeof(__pyx_k_Place_global_headers_in_extradat), 0, 0, 1, 0},
  {&__pyx_n_s_QPEL, __pyx_k_QPEL, sizeof(__pyx_k_QPEL), 0, 0, 1, 1},
  {&__pyx_n_s_QSCALE, __pyx_k_QSCALE, sizeof(__pyx_k_QSCALE), 0, 0, 1, 1},
  {&__pyx_n_s_RO_FLUSH_NOOP, __pyx_k_RO_FLUSH_NOOP, sizeof(__pyx_k_RO_FLUSH_NOOP), 0, 0, 1, 1},
  {&__pyx_n_s_RuntimeError, __pyx_k_RuntimeError, sizeof(__pyx_k_RuntimeError), 0, 0, 1, 1},
  {&__pyx_n_s_SHOW_ALL, __pyx_k_SHOW_ALL, sizeof(__pyx_k_SHOW_ALL), 0, 0, 1, 1},
  {&__pyx_n_s_SKIP_MANUAL, __pyx_k_SKIP_MANUAL, sizeof(__pyx_k_SKIP_MANUAL), 0, 0, 1, 1},
  {&__pyx_n_s_SLICE, __pyx_k_SLICE, sizeof(__pyx_k_SLICE), 0, 0, 1, 1},
  {&__pyx_kp_s_Show_all_frames_before_the_first, __pyx_k_Show_all_frames_before_the_first, sizeof(__pyx_k_Show_all_frames_before_the_first), 0, 0, 1, 0},
  {&__pyx_n_s_SkipType, __pyx_k_SkipType, sizeof(__pyx_k_SkipType), 0, 0, 1, 1},
  {&__pyx_kp_s_Skip_bitstream_encoding, __pyx_k_Skip_bitstream_encoding, sizeof(__pyx_k_Skip_bitstream_encoding), 0, 0, 1, 0},
  {&__pyx_n_s_SubtitleCodecContext, __pyx_k_SubtitleCodecContext, sizeof(__pyx_k_SubtitleCodecContext), 0, 0, 1, 1},
  {&__pyx_n_s_ThreadType, __pyx_k_ThreadType, sizeof(__pyx_k_ThreadType), 0, 0, 1, 1},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_n_s_UNALIGNED, __pyx_k_UNALIGNED, sizeof(__pyx_k_UNALIGNED), 0, 0, 1, 1},
  {&__pyx_kp_s_Use_fixed_qscale, __pyx_k_Use_fixed_qscale, sizeof(__pyx_k_Use_fixed_qscale), 0, 0, 1, 0},
  {&__pyx_kp_s_Use_interlaced_DCT, __pyx_k_Use_interlaced_DCT, sizeof(__pyx_k_Use_interlaced_DCT), 0, 0, 1, 0},
  {&__pyx_kp_s_Use_internal_2pass_ratecontrol_i, __pyx_k_Use_internal_2pass_ratecontrol_i, sizeof(__pyx_k_Use_internal_2pass_ratecontrol_i), 0, 0, 1, 0},
  {&__pyx_kp_s_Use_internal_2pass_ratecontrol_i_2, __pyx_k_Use_internal_2pass_ratecontrol_i_2, sizeof(__pyx_k_Use_internal_2pass_ratecontrol_i_2), 0, 0, 1, 0},
  {&__pyx_kp_s_Use_only_bitexact_stuff_except_I, __pyx_k_Use_only_bitexact_stuff_except_I, sizeof(__pyx_k_Use_only_bitexact_stuff_except_I), 0, 0, 1, 0},
  {&__pyx_kp_s_Use_qpel_MC, __pyx_k_Use_qpel_MC, sizeof(__pyx_k_Use_qpel_MC), 0, 0, 1, 0},
  {&__pyx_kp_s_Using_CodecContext_time_base_for, __pyx_k_Using_CodecContext_time_base_for, sizeof(__pyx_k_Using_CodecContext_time_base_for), 0, 0, 1, 0},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_n_s_VideoCodecContext, __pyx_k_VideoCodecContext, sizeof(__pyx_k_VideoCodecContext), 0, 0, 1, 1},
  {&__pyx_kp_s_Wrapping_CodecContext_with_misma, __pyx_k_Wrapping_CodecContext_with_misma, sizeof(__pyx_k_Wrapping_CodecContext_with_misma), 0, 0, 1, 0},
  {&__pyx_n_s_ac_pred, __pyx_k_ac_pred, sizeof(__pyx_k_ac_pred), 0, 0, 1, 1},
  {&__pyx_n_s_ascii, __pyx_k_ascii, sizeof(__pyx_k_ascii), 0, 0, 1, 1},
  {&__pyx_n_s_av_audio_codeccontext, __pyx_k_av_audio_codeccontext, sizeof(__pyx_k_av_audio_codeccontext), 0, 0, 1, 1},
  {&__pyx_n_s_av_codec_context, __pyx_k_av_codec_context, sizeof(__pyx_k_av_codec_context), 0, 0, 1, 1},
  {&__pyx_kp_s_av_codec_context_pyx, __pyx_k_av_codec_context_pyx, sizeof(__pyx_k_av_codec_context_pyx), 0, 0, 1, 0},
  {&__pyx_n_s_av_deprecation, __pyx_k_av_deprecation, sizeof(__pyx_k_av_deprecation), 0, 0, 1, 1},
  {&__pyx_n_s_av_dictionary, __pyx_k_av_dictionary, sizeof(__pyx_k_av_dictionary), 0, 0, 1, 1},
  {&__pyx_kp_s_av_s_s_s_at_0x_x, __pyx_k_av_s_s_s_at_0x_x, sizeof(__pyx_k_av_s_s_s_at_0x_x), 0, 0, 1, 0},
  {&__pyx_n_s_av_subtitles_codeccontext, __pyx_k_av_subtitles_codeccontext, sizeof(__pyx_k_av_subtitles_codeccontext), 0, 0, 1, 1},
  {&__pyx_n_s_av_video_codeccontext, __pyx_k_av_video_codeccontext, sizeof(__pyx_k_av_video_codeccontext), 0, 0, 1, 1},
  {&__pyx_n_s_bitexact, __pyx_k_bitexact, sizeof(__pyx_k_bitexact), 0, 0, 1, 1},
  {&__pyx_n_s_byteorder, __pyx_k_byteorder, sizeof(__pyx_k_byteorder), 0, 0, 1, 1},
  {&__pyx_n_s_c_ctx, __pyx_k_c_ctx, sizeof(__pyx_k_c_ctx), 0, 0, 1, 1},
  {&__pyx_kp_s_cannot_decode_unknown_codec, __pyx_k_cannot_decode_unknown_codec, sizeof(__pyx_k_cannot_decode_unknown_codec), 0, 0, 1, 0},
  {&__pyx_n_s_chunks, __pyx_k_chunks, sizeof(__pyx_k_chunks), 0, 0, 1, 1},
  {&__pyx_n_s_class, __pyx_k_class, sizeof(__pyx_k_class), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
  {&__pyx_n_s_closed_gop, __pyx_k_closed_gop, sizeof(__pyx_k_closed_gop), 0, 0, 1, 1},
  {&__pyx_n_s_codec, __pyx_k_codec, sizeof(__pyx_k_codec), 0, 0, 1, 1},
  {&__pyx_n_s_create, __pyx_k_create, sizeof(__pyx_k_create), 0, 0, 1, 1},
  {&__pyx_n_s_cy_codec, __pyx_k_cy_codec, sizeof(__pyx_k_cy_codec), 0, 0, 1, 1},
  {&__pyx_n_s_decode, __pyx_k_decode, sizeof(__pyx_k_decode), 0, 0, 1, 1},
  {&__pyx_n_s_drop_changed, __pyx_k_drop_changed, sizeof(__pyx_k_drop_changed), 0, 0, 1, 1},
  {&__pyx_n_s_encode, __pyx_k_encode, sizeof(__pyx_k_encode), 0, 0, 1, 1},
  {&__pyx_n_s_encoding, __pyx_k_encoding, sizeof(__pyx_k_encoding), 0, 0, 1, 1},
  {&__pyx_kp_s_error_variables_will_be_set_duri, __pyx_k_error_variables_will_be_set_duri, sizeof(__pyx_k_error_variables_will_be_set_duri), 0, 0, 1, 0},
  {&__pyx_n_s_export_mvs, __pyx_k_export_mvs, sizeof(__pyx_k_export_mvs), 0, 0, 1, 1},
  {&__pyx_n_s_fast, __pyx_k_fast, sizeof(__pyx_k_fast), 0, 0, 1, 1},
  {&__pyx_n_s_flag_property, __pyx_k_flag_property, sizeof(__pyx_k_flag_property), 0, 0, 1, 1},
  {&__pyx_n_s_flags, __pyx_k_flags, sizeof(__pyx_k_flags), 0, 0, 1, 1},
  {&__pyx_n_s_flags2, __pyx_k_flags2, sizeof(__pyx_k_flags2), 0, 0, 1, 1},
  {&__pyx_n_s_four_mv, __pyx_k_four_mv, sizeof(__pyx_k_four_mv), 0, 0, 1, 1},
  {&__pyx_n_s_frame, __pyx_k_frame, sizeof(__pyx_k_frame), 0, 0, 1, 1},
  {&__pyx_n_s_from_bytes, __pyx_k_from_bytes, sizeof(__pyx_k_from_bytes), 0, 0, 1, 1},
  {&__pyx_n_s_get, __pyx_k_get, sizeof(__pyx_k_get), 0, 0, 1, 1},
  {&__pyx_n_s_get_2, __pyx_k_get_2, sizeof(__pyx_k_get_2), 0, 0, 1, 1},
  {&__pyx_n_s_get_flags, __pyx_k_get_flags, sizeof(__pyx_k_get_flags), 0, 0, 1, 1},
  {&__pyx_n_s_get_flags2, __pyx_k_get_flags2, sizeof(__pyx_k_get_flags2), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_n_s_global_header, __pyx_k_global_header, sizeof(__pyx_k_global_header), 0, 0, 1, 1},
  {&__pyx_n_s_gray, __pyx_k_gray, sizeof(__pyx_k_gray), 0, 0, 1, 1},
  {&__pyx_n_s_id, __pyx_k_id, sizeof(__pyx_k_id), 0, 0, 1, 1},
  {&__pyx_n_s_ignore_crop, __pyx_k_ignore_crop, sizeof(__pyx_k_ignore_crop), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_interlaced_dct, __pyx_k_interlaced_dct, sizeof(__pyx_k_interlaced_dct), 0, 0, 1, 1},
  {&__pyx_n_s_interlaced_me, __pyx_k_interlaced_me, sizeof(__pyx_k_interlaced_me), 0, 0, 1, 1},
  {&__pyx_n_s_is_decoder, __pyx_k_is_decoder, sizeof(__pyx_k_is_decoder), 0, 0, 1, 1},
  {&__pyx_n_s_is_encoder, __pyx_k_is_encoder, sizeof(__pyx_k_is_encoder), 0, 0, 1, 1},
  {&__pyx_n_s_little, __pyx_k_little, sizeof(__pyx_k_little), 0, 0, 1, 1},
  {&__pyx_n_s_local_header, __pyx_k_local_header, sizeof(__pyx_k_local_header), 0, 0, 1, 1},
  {&__pyx_kp_s_loop_filter, __pyx_k_loop_filter, sizeof(__pyx_k_loop_filter), 0, 0, 1, 0},
  {&__pyx_n_s_loop_filter_2, __pyx_k_loop_filter_2, sizeof(__pyx_k_loop_filter_2), 0, 0, 1, 1},
  {&__pyx_n_s_low_delay, __pyx_k_low_delay, sizeof(__pyx_k_low_delay), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_mode, __pyx_k_mode, sizeof(__pyx_k_mode), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
  {&__pyx_kp_s_no_default___reduce___due_to_non, __pyx_k_no_default___reduce___due_to_non, sizeof(__pyx_k_no_default___reduce___due_to_non), 0, 0, 1, 0},
  {&__pyx_n_s_no_output, __pyx_k_no_output, sizeof(__pyx_k_no_output), 0, 0, 1, 1},
  {&__pyx_kp_s_nocodec, __pyx_k_nocodec, sizeof(__pyx_k_nocodec), 0, 0, 1, 0},
  {&__pyx_kp_s_notype, __pyx_k_notype, sizeof(__pyx_k_notype), 0, 0, 1, 0},
  {&__pyx_n_s_object, __pyx_k_object, sizeof(__pyx_k_object), 0, 0, 1, 1},
  {&__pyx_n_s_open, __pyx_k_open, sizeof(__pyx_k_open), 0, 0, 1, 1},
  {&__pyx_n_s_output_corrupt, __pyx_k_output_corrupt, sizeof(__pyx_k_output_corrupt), 0, 0, 1, 1},
  {&__pyx_n_s_packet, __pyx_k_packet, sizeof(__pyx_k_packet), 0, 0, 1, 1},
  {&__pyx_n_s_pass1, __pyx_k_pass1, sizeof(__pyx_k_pass1), 0, 0, 1, 1},
  {&__pyx_n_s_pass2, __pyx_k_pass2, sizeof(__pyx_k_pass2), 0, 0, 1, 1},
  {&__pyx_n_s_property, __pyx_k_property, sizeof(__pyx_k_property), 0, 0, 1, 1},
  {&__pyx_n_s_psnr, __pyx_k_psnr, sizeof(__pyx_k_psnr), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_qpel, __pyx_k_qpel, sizeof(__pyx_k_qpel), 0, 0, 1, 1},
  {&__pyx_n_s_qscale, __pyx_k_qscale, sizeof(__pyx_k_qscale), 0, 0, 1, 1},
  {&__pyx_n_s_raw_input, __pyx_k_raw_input, sizeof(__pyx_k_raw_input), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_ro_flush_noop, __pyx_k_ro_flush_noop, sizeof(__pyx_k_ro_flush_noop), 0, 0, 1, 1},
  {&__pyx_n_s_sentinel, __pyx_k_sentinel, sizeof(__pyx_k_sentinel), 0, 0, 1, 1},
  {&__pyx_n_s_set_flags, __pyx_k_set_flags, sizeof(__pyx_k_set_flags), 0, 0, 1, 1},
  {&__pyx_n_s_set_flags2, __pyx_k_set_flags2, sizeof(__pyx_k_set_flags2), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_show_all, __pyx_k_show_all, sizeof(__pyx_k_show_all), 0, 0, 1, 1},
  {&__pyx_n_s_signed, __pyx_k_signed, sizeof(__pyx_k_signed), 0, 0, 1, 1},
  {&__pyx_n_s_skip_manual, __pyx_k_skip_manual, sizeof(__pyx_k_skip_manual), 0, 0, 1, 1},
  {&__pyx_n_s_staticmethod, __pyx_k_staticmethod, sizeof(__pyx_k_staticmethod), 0, 0, 1, 1},
  {&__pyx_n_s_strict, __pyx_k_strict, sizeof(__pyx_k_strict), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_to_bytes, __pyx_k_to_bytes, sizeof(__pyx_k_to_bytes), 0, 0, 1, 1},
  {&__pyx_n_s_type, __pyx_k_type, sizeof(__pyx_k_type), 0, 0, 1, 1},
  {&__pyx_n_s_unaligned, __pyx_k_unaligned, sizeof(__pyx_k_unaligned), 0, 0, 1, 1},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {&__pyx_n_s_value, __pyx_k_value, sizeof(__pyx_k_value), 0, 0, 1, 1},
  {&__pyx_n_s_warn, __pyx_k_warn, sizeof(__pyx_k_warn), 0, 0, 1, 1},
  {&__pyx_n_s_warnings, __pyx_k_warnings, sizeof(__pyx_k_warnings), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_object = __Pyx_GetBuiltinName(__pyx_n_s_object); if (!__pyx_builtin_object) __PYX_ERR(0, 20, __pyx_L1_error)
  __pyx_builtin_staticmethod = __Pyx_GetBuiltinName(__pyx_n_s_staticmethod); if (!__pyx_builtin_staticmethod) __PYX_ERR(0, 140, __pyx_L1_error)
  __pyx_builtin_RuntimeError = __Pyx_GetBuiltinName(__pyx_n_s_RuntimeError); if (!__pyx_builtin_RuntimeError) __PYX_ERR(0, 148, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 228, __pyx_L1_error)
  __pyx_builtin_MemoryError = __Pyx_GetBuiltinName(__pyx_n_s_MemoryError); if (!__pyx_builtin_MemoryError) __PYX_ERR(0, 237, __pyx_L1_error)
  __pyx_builtin_id = __Pyx_GetBuiltinName(__pyx_n_s_id); if (!__pyx_builtin_id) __PYX_ERR(0, 302, __pyx_L1_error)
  __pyx_builtin_NotImplementedError = __Pyx_GetBuiltinName(__pyx_n_s_NotImplementedError); if (!__pyx_builtin_NotImplementedError) __PYX_ERR(0, 424, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(1, 2, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "av/codec/context.pyx":148
 *     def __cinit__(self, sentinel=None, *args, **kwargs):
 *         if sentinel is not _cinit_sentinel:
 *             raise RuntimeError('Cannot instantiate CodecContext')             # <<<<<<<<<<<<<<
 * 
 *         self.options = {}
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_s_Cannot_instantiate_CodecContext); if (unlikely(!__pyx_tuple_)) __PYX_ERR(0, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "av/codec/context.pyx":157
 *         self.ptr = ptr
 *         if self.ptr.codec and codec and self.ptr.codec != codec:
 *             raise RuntimeError('Wrapping CodecContext with mismatched codec.')             # <<<<<<<<<<<<<<
 *         self.codec = wrap_codec(codec if codec != NULL else self.ptr.codec)
 * 
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_s_Wrapping_CodecContext_with_misma); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(0, 157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "av/codec/context.pyx":228
 *         def __set__(self, data):
 *             if not self.is_decoder:
 *                 raise ValueError("Can only set extradata for decoders.")             # <<<<<<<<<<<<<<
 * 
 *             if data is None:
 */
  __pyx_tuple__3 = PyTuple_Pack(1, __pyx_kp_s_Can_only_set_extradata_for_decod); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(0, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "av/codec/context.pyx":237
 *                 self.ptr.extradata = <uint8_t*>lib.av_realloc(self.ptr.extradata, source.length + lib.AV_INPUT_BUFFER_PADDING_SIZE)
 *                 if not self.ptr.extradata:
 *                     raise MemoryError("Cannot allocate extradata")             # <<<<<<<<<<<<<<
 *                 memcpy(self.ptr.extradata, source.ptr, source.length)
 *                 self.ptr.extradata_size = source.length
 */
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_kp_s_Cannot_allocate_extradata); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(0, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "av/codec/context.pyx":262
 *         if lib.avcodec_is_open(self.ptr):
 *             if strict:
 *                 raise ValueError('CodecContext is already open.')             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_kp_s_CodecContext_is_already_open); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(0, 262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "av/codec/context.pyx":284
 *         if not lib.avcodec_is_open(self.ptr):
 *             if strict:
 *                 raise ValueError('CodecContext is already closed.')             # <<<<<<<<<<<<<<
 *             return
 *         err_check(lib.avcodec_close(self.ptr))
 */
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_kp_s_CodecContext_is_already_closed); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "av/codec/context.pyx":424
 * 
 *     cdef Frame _alloc_next_frame(self):
 *         raise NotImplementedError('Base CodecContext cannot decode.')             # <<<<<<<<<<<<<<
 * 
 *     cdef _recv_frame(self):
 */
  __pyx_tuple__7 = PyTuple_Pack(1, __pyx_kp_s_Base_CodecContext_cannot_decode); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(0, 424, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "av/codec/context.pyx":462
 * 
 *         if self.ptr.codec_type not in [lib.AVMEDIA_TYPE_VIDEO, lib.AVMEDIA_TYPE_AUDIO]:
 *             raise NotImplementedError('Encoding is only supported for audio and video.')             # <<<<<<<<<<<<<<
 * 
 *         self.open(strict=False)
 */
  __pyx_tuple__8 = PyTuple_Pack(1, __pyx_kp_s_Encoding_is_only_supported_for_a); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(0, 462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "av/codec/context.pyx":501
 * 
 *         if not self.codec.ptr:
 *             raise ValueError('cannot decode unknown codec')             # <<<<<<<<<<<<<<
 * 
 *         self.open(strict=False)
 */
  __pyx_tuple__9 = PyTuple_Pack(1, __pyx_kp_s_cannot_decode_unknown_codec); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(0, 501, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "av/codec/context.pyx":555
 *     property codec_tag:
 *         def __get__(self):
 *             return self.ptr.codec_tag.to_bytes(4, byteorder="little", signed=False).decode(             # <<<<<<<<<<<<<<
 *                 encoding="ascii")
 * 
 */
  __pyx_tuple__10 = PyTuple_Pack(1, __pyx_int_4); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(0, 555, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "av/codec/context.pyx":563
 *                                                     byteorder="little", signed=False)
 *             else:
 *                 raise ValueError("Codec tag should be a 4 character string.")             # <<<<<<<<<<<<<<
 * 
 *     property ticks_per_frame:
 */
  __pyx_tuple__11 = PyTuple_Pack(1, __pyx_kp_s_Codec_tag_should_be_a_4_characte); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(0, 563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "av/codec/context.pyx":602
 *         def __set__(self, int value):
 *             if lib.avcodec_is_open(self.ptr):
 *                 raise RuntimeError("Cannot change thread_count after codec is open.")             # <<<<<<<<<<<<<<
 *             self.ptr.thread_count = value
 * 
 */
  __pyx_tuple__12 = PyTuple_Pack(1, __pyx_kp_s_Cannot_change_thread_count_after); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(0, 602, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "av/codec/context.pyx":617
 *         def __set__(self, value):
 *             if lib.avcodec_is_open(self.ptr):
 *                 raise RuntimeError("Cannot change thread_type after codec is open.")             # <<<<<<<<<<<<<<
 *             self.ptr.thread_type = ThreadType[value].value
 * 
 */
  __pyx_tuple__13 = PyTuple_Pack(1, __pyx_kp_s_Cannot_change_thread_type_after); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(0, 617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__14 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__15 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);

  /* "av/codec/context.pyx":47
 * 
 * ThreadType = define_enum('ThreadType', __name__, (
 *     ('NONE', 0),             # <<<<<<<<<<<<<<
 *     ('FRAME', lib.FF_THREAD_FRAME,
 *         """Decode more than one frame at once"""),
 */
  __pyx_tuple__16 = PyTuple_Pack(2, __pyx_n_s_NONE, __pyx_int_0); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);

  /* "av/codec/context.pyx":86
 *     ('QPEL', lib.AV_CODEC_FLAG_QPEL,
 *         """Use qpel MC."""),
 *     ('DROPCHANGED', 1 << 5,             # <<<<<<<<<<<<<<
 *         """Don't output frames whose parameters differ from first
 *         decoded frame in stream."""),
 */
  __pyx_tuple__17 = PyTuple_Pack(3, __pyx_n_s_DROPCHANGED, __pyx_int_32, __pyx_kp_s_Don_t_output_frames_whose_parame); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(0, 86, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);

  /* "av/codec/context.pyx":141
 * 
 *     @staticmethod
 *     def create(codec, mode=None):             # <<<<<<<<<<<<<<
 *         cdef Codec cy_codec = codec if isinstance(codec, Codec) else Codec(codec, mode)
 *         cdef lib.AVCodecContext *c_ctx = lib.avcodec_alloc_context3(cy_codec.ptr)
 */
  __pyx_tuple__18 = PyTuple_Pack(4, __pyx_n_s_codec, __pyx_n_s_mode, __pyx_n_s_cy_codec, __pyx_n_s_c_ctx); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(0, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);
  __pyx_codeobj__19 = (PyObject*)__Pyx_PyCode_New(2, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__18, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_av_codec_context_pyx, __pyx_n_s_create, 141, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__19)) __PYX_ERR(0, 141, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_4 = PyInt_FromLong(4); if (unlikely(!__pyx_int_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_32 = PyInt_FromLong(32); if (unlikely(!__pyx_int_32)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __pyx_v_2av_5codec_7context__cinit_sentinel = Py_None; Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  if (__Pyx_ExportFunction("wrap_codec_context", (void (*)(void))__pyx_f_2av_5codec_7context_wrap_codec_context, "struct __pyx_obj_2av_5codec_7context_CodecContext *(struct AVCodecContext *, struct AVCodec const *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_2av_5codec_7context_CodecContext = &__pyx_vtable_2av_5codec_7context_CodecContext;
  __pyx_vtable_2av_5codec_7context_CodecContext._init = (PyObject *(*)(struct __pyx_obj_2av_5codec_7context_CodecContext *, struct AVCodecContext *, struct AVCodec const *))__pyx_f_2av_5codec_7context_12CodecContext__init;
  __pyx_vtable_2av_5codec_7context_CodecContext.open = (PyObject *(*)(struct __pyx_obj_2av_5codec_7context_CodecContext *, int __pyx_skip_dispatch, struct __pyx_opt_args_2av_5codec_7context_12CodecContext_open *__pyx_optional_args))__pyx_f_2av_5codec_7context_12CodecContext_open;
  __pyx_vtable_2av_5codec_7context_CodecContext.close = (PyObject *(*)(struct __pyx_obj_2av_5codec_7context_CodecContext *, int __pyx_skip_dispatch, struct __pyx_opt_args_2av_5codec_7context_12CodecContext_close *__pyx_optional_args))__pyx_f_2av_5codec_7context_12CodecContext_close;
  __pyx_vtable_2av_5codec_7context_CodecContext._set_default_time_base = (PyObject *(*)(struct __pyx_obj_2av_5codec_7context_CodecContext *))__pyx_f_2av_5codec_7context_12CodecContext__set_default_time_base;
  __pyx_vtable_2av_5codec_7context_CodecContext.encode = (PyObject *(*)(struct __pyx_obj_2av_5codec_7context_CodecContext *, int __pyx_skip_dispatch, struct __pyx_opt_args_2av_5codec_7context_12CodecContext_encode *__pyx_optional_args))__pyx_f_2av_5codec_7context_12CodecContext_encode;
  __pyx_vtable_2av_5codec_7context_CodecContext.decode = (PyObject *(*)(struct __pyx_obj_2av_5codec_7context_CodecContext *, int __pyx_skip_dispatch, struct __pyx_opt_args_2av_5codec_7context_12CodecContext_decode *__pyx_optional_args))__pyx_f_2av_5codec_7context_12CodecContext_decode;
  __pyx_vtable_2av_5codec_7context_CodecContext._prepare_frames_for_encode = (PyObject *(*)(struct __pyx_obj_2av_5codec_7context_CodecContext *, struct __pyx_obj_2av_5frame_Frame *))__pyx_f_2av_5codec_7context_12CodecContext__prepare_frames_for_encode;
  __pyx_vtable_2av_5codec_7context_CodecContext._setup_encoded_packet = (PyObject *(*)(struct __pyx_obj_2av_5codec_7context_CodecContext *, struct __pyx_obj_2av_6packet_Packet *))__pyx_f_2av_5codec_7context_12CodecContext__setup_encoded_packet;
  __pyx_vtable_2av_5codec_7context_CodecContext._setup_decoded_frame = (PyObject *(*)(struct __pyx_obj_2av_5codec_7context_CodecContext *, struct __pyx_obj_2av_5frame_Frame *, struct __pyx_obj_2av_6packet_Packet *))__pyx_f_2av_5codec_7context_12CodecContext__setup_decoded_frame;
  __pyx_vtable_2av_5codec_7context_CodecContext._send_frame_and_recv = (PyObject *(*)(struct __pyx_obj_2av_5codec_7context_CodecContext *, struct __pyx_obj_2av_5frame_Frame *))__pyx_f_2av_5codec_7context_12CodecContext__send_frame_and_recv;
  __pyx_vtable_2av_5codec_7context_CodecContext._recv_packet = (PyObject *(*)(struct __pyx_obj_2av_5codec_7context_CodecContext *))__pyx_f_2av_5codec_7context_12CodecContext__recv_packet;
  __pyx_vtable_2av_5codec_7context_CodecContext._send_packet_and_recv = (PyObject *(*)(struct __pyx_obj_2av_5codec_7context_CodecContext *, struct __pyx_obj_2av_6packet_Packet *))__pyx_f_2av_5codec_7context_12CodecContext__send_packet_and_recv;
  __pyx_vtable_2av_5codec_7context_CodecContext._recv_frame = (PyObject *(*)(struct __pyx_obj_2av_5codec_7context_CodecContext *))__pyx_f_2av_5codec_7context_12CodecContext__recv_frame;
  __pyx_vtable_2av_5codec_7context_CodecContext._alloc_next_frame = (struct __pyx_obj_2av_5frame_Frame *(*)(struct __pyx_obj_2av_5codec_7context_CodecContext *))__pyx_f_2av_5codec_7context_12CodecContext__alloc_next_frame;
  if (PyType_Ready(&__pyx_type_2av_5codec_7context_CodecContext) < 0) __PYX_ERR(0, 138, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_2av_5codec_7context_CodecContext.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_2av_5codec_7context_CodecContext.tp_dictoffset && __pyx_type_2av_5codec_7context_CodecContext.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_2av_5codec_7context_CodecContext.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_2av_5codec_7context_CodecContext.tp_dict, __pyx_vtabptr_2av_5codec_7context_CodecContext) < 0) __PYX_ERR(0, 138, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_CodecContext, (PyObject *)&__pyx_type_2av_5codec_7context_CodecContext) < 0) __PYX_ERR(0, 138, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_2av_5codec_7context_CodecContext) < 0) __PYX_ERR(0, 138, __pyx_L1_error)
  __pyx_ptype_2av_5codec_7context_CodecContext = &__pyx_type_2av_5codec_7context_CodecContext;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule("av.bytesource"); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_2av_10bytesource_ByteSource = __Pyx_ImportType(__pyx_t_1, "av.bytesource", "ByteSource", sizeof(struct __pyx_obj_2av_10bytesource_ByteSource), __PYX_GET_STRUCT_ALIGNMENT(struct __pyx_obj_2av_10bytesource_ByteSource),
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2av_10bytesource_ByteSource) __PYX_ERR(3, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("av.codec.codec"); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_2av_5codec_5codec_Codec = __Pyx_ImportType(__pyx_t_1, "av.codec.codec", "Codec", sizeof(struct __pyx_obj_2av_5codec_5codec_Codec), __PYX_GET_STRUCT_ALIGNMENT(struct __pyx_obj_2av_5codec_5codec_Codec),
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2av_5codec_5codec_Codec) __PYX_ERR(4, 4, __pyx_L1_error)
  __pyx_vtabptr_2av_5codec_5codec_Codec = (struct __pyx_vtabstruct_2av_5codec_5codec_Codec*)__Pyx_GetVtable(__pyx_ptype_2av_5codec_5codec_Codec->tp_dict); if (unlikely(!__pyx_vtabptr_2av_5codec_5codec_Codec)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("av.buffer"); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_2av_6buffer_Buffer = __Pyx_ImportType(__pyx_t_1, "av.buffer", "Buffer", sizeof(struct __pyx_obj_2av_6buffer_Buffer), __PYX_GET_STRUCT_ALIGNMENT(struct __pyx_obj_2av_6buffer_Buffer),
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2av_6buffer_Buffer) __PYX_ERR(5, 2, __pyx_L1_error)
  __pyx_vtabptr_2av_6buffer_Buffer = (struct __pyx_vtabstruct_2av_6buffer_Buffer*)__Pyx_GetVtable(__pyx_ptype_2av_6buffer_Buffer->tp_dict); if (unlikely(!__pyx_vtabptr_2av_6buffer_Buffer)) __PYX_ERR(5, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("av.container.pyio"); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_2av_9container_4pyio_PyIOFile = __Pyx_ImportType(__pyx_t_1, "av.container.pyio", "PyIOFile", sizeof(struct __pyx_obj_2av_9container_4pyio_PyIOFile), __PYX_GET_STRUCT_ALIGNMENT(struct __pyx_obj_2av_9container_4pyio_PyIOFile),
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2av_9container_4pyio_PyIOFile) __PYX_ERR(6, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("av.container.streams"); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_2av_9container_7streams_StreamContainer = __Pyx_ImportType(__pyx_t_1, "av.container.streams", "StreamContainer", sizeof(struct __pyx_obj_2av_9container_7streams_StreamContainer), __PYX_GET_STRUCT_ALIGNMENT(struct __pyx_obj_2av_9container_7streams_StreamContainer),
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2av_9container_7streams_StreamContainer) __PYX_ERR(7, 4, __pyx_L1_error)
  __pyx_vtabptr_2av_9container_7streams_StreamContainer = (struct __pyx_vtabstruct_2av_9container_7streams_StreamContainer*)__Pyx_GetVtable(__pyx_ptype_2av_9container_7streams_StreamContainer->tp_dict); if (unlikely(!__pyx_vtabptr_2av_9container_7streams_StreamContainer)) __PYX_ERR(7, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("av.dictionary"); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_2av_10dictionary__Dictionary = __Pyx_ImportType(__pyx_t_1, "av.dictionary", "_Dictionary", sizeof(struct __pyx_obj_2av_10dictionary__Dictionary), __PYX_GET_STRUCT_ALIGNMENT(struct __pyx_obj_2av_10dictionary__Dictionary),
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2av_10dictionary__Dictionary) __PYX_ERR(8, 4, __pyx_L1_error)
  __pyx_vtabptr_2av_10dictionary__Dictionary = (struct __pyx_vtabstruct_2av_10dictionary__Dictionary*)__Pyx_GetVtable(__pyx_ptype_2av_10dictionary__Dictionary->tp_dict); if (unlikely(!__pyx_vtabptr_2av_10dictionary__Dictionary)) __PYX_ERR(8, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("av.format"); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_2av_6format_ContainerFormat = __Pyx_ImportType(__pyx_t_1, "av.format", "ContainerFormat", sizeof(struct __pyx_obj_2av_6format_ContainerFormat), __PYX_GET_STRUCT_ALIGNMENT(struct __pyx_obj_2av_6format_ContainerFormat),
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2av_6format_ContainerFormat) __PYX_ERR(9, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("av.container.core"); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_2av_9container_4core_Container = __Pyx_ImportType(__pyx_t_1, "av.container.core", "Container", sizeof(struct __pyx_obj_2av_9container_4core_Container), __PYX_GET_STRUCT_ALIGNMENT(struct __pyx_obj_2av_9container_4core_Container),
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2av_9container_4core_Container) __PYX_ERR(10, 16, __pyx_L1_error)
  __pyx_vtabptr_2av_9container_4core_Container = (struct __pyx_vtabstruct_2av_9container_4core_Container*)__Pyx_GetVtable(__pyx_ptype_2av_9container_4core_Container->tp_dict); if (unlikely(!__pyx_vtabptr_2av_9container_4core_Container)) __PYX_ERR(10, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("av.stream"); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_2av_6stream_Stream = __Pyx_ImportType(__pyx_t_1, "av.stream", "Stream", sizeof(struct __pyx_obj_2av_6stream_Stream), __PYX_GET_STRUCT_ALIGNMENT(struct __pyx_obj_2av_6stream_Stream),
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2av_6stream_Stream) __PYX_ERR(11, 10, __pyx_L1_error)
  __pyx_vtabptr_2av_6stream_Stream = (struct __pyx_vtabstruct_2av_6stream_Stream*)__Pyx_GetVtable(__pyx_ptype_2av_6stream_Stream->tp_dict); if (unlikely(!__pyx_vtabptr_2av_6stream_Stream)) __PYX_ERR(11, 10, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("av.packet"); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_2av_6packet_Packet = __Pyx_ImportType(__pyx_t_1, "av.packet", "Packet", sizeof(struct __pyx_obj_2av_6packet_Packet), __PYX_GET_STRUCT_ALIGNMENT(struct __pyx_obj_2av_6packet_Packet),
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2av_6packet_Packet) __PYX_ERR(12, 8, __pyx_L1_error)
  __pyx_vtabptr_2av_6packet_Packet = (struct __pyx_vtabstruct_2av_6packet_Packet*)__Pyx_GetVtable(__pyx_ptype_2av_6packet_Packet->tp_dict); if (unlikely(!__pyx_vtabptr_2av_6packet_Packet)) __PYX_ERR(12, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("av.sidedata.sidedata"); if (unlikely(!__pyx_t_1)) __PYX_ERR(13, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_2av_8sidedata_8sidedata_SideData = __Pyx_ImportType(__pyx_t_1, "av.sidedata.sidedata", "SideData", sizeof(struct __pyx_obj_2av_8sidedata_8sidedata_SideData), __PYX_GET_STRUCT_ALIGNMENT(struct __pyx_obj_2av_8sidedata_8sidedata_SideData),
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2av_8sidedata_8sidedata_SideData) __PYX_ERR(13, 9, __pyx_L1_error)
  __pyx_vtabptr_2av_8sidedata_8sidedata_SideData = (struct __pyx_vtabstruct_2av_8sidedata_8sidedata_SideData*)__Pyx_GetVtable(__pyx_ptype_2av_8sidedata_8sidedata_SideData->tp_dict); if (unlikely(!__pyx_vtabptr_2av_8sidedata_8sidedata_SideData)) __PYX_ERR(13, 9, __pyx_L1_error)
  __pyx_ptype_2av_8sidedata_8sidedata__SideDataContainer = __Pyx_ImportType(__pyx_t_1, "av.sidedata.sidedata", "_SideDataContainer", sizeof(struct __pyx_obj_2av_8sidedata_8sidedata__SideDataContainer), __PYX_GET_STRUCT_ALIGNMENT(struct __pyx_obj_2av_8sidedata_8sidedata__SideDataContainer),
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2av_8sidedata_8sidedata__SideDataContainer) __PYX_ERR(13, 18, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("av.frame"); if (unlikely(!__pyx_t_1)) __PYX_ERR(14, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_2av_5frame_Frame = __Pyx_ImportType(__pyx_t_1, "av.frame", "Frame", sizeof(struct __pyx_obj_2av_5frame_Frame), __PYX_GET_STRUCT_ALIGNMENT(struct __pyx_obj_2av_5frame_Frame),
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2av_5frame_Frame) __PYX_ERR(14, 7, __pyx_L1_error)
  __pyx_vtabptr_2av_5frame_Frame = (struct __pyx_vtabstruct_2av_5frame_Frame*)__Pyx_GetVtable(__pyx_ptype_2av_5frame_Frame->tp_dict); if (unlikely(!__pyx_vtabptr_2av_5frame_Frame)) __PYX_ERR(14, 7, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __pyx_t_1 = PyImport_ImportModule("av.bytesource"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportFunction(__pyx_t_1, "bytesource", (void (**)(void))&__pyx_f_2av_10bytesource_bytesource, "struct __pyx_obj_2av_10bytesource_ByteSource *(PyObject *, struct __pyx_opt_args_2av_10bytesource_bytesource *__pyx_optional_args)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("av.codec.codec"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportFunction(__pyx_t_1, "wrap_codec", (void (**)(void))&__pyx_f_2av_5codec_5codec_wrap_codec, "struct __pyx_obj_2av_5codec_5codec_Codec *(struct AVCodec const *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("av.dictionary"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportFunction(__pyx_t_1, "wrap_dictionary", (void (**)(void))&__pyx_f_2av_10dictionary_wrap_dictionary, "struct __pyx_obj_2av_10dictionary__Dictionary *(AVDictionary *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("av.enum"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportFunction(__pyx_t_1, "define_enum", (void (**)(void))&__pyx_f_2av_4enum_define_enum, "PyObject *(PyObject *, PyObject *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_2av_4enum_define_enum *__pyx_optional_args)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("av.error"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportFunction(__pyx_t_1, "err_check", (void (**)(void))&__pyx_f_2av_5error_err_check, "int (int, int __pyx_skip_dispatch, struct __pyx_opt_args_2av_5error_err_check *__pyx_optional_args)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("av.utils"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportFunction(__pyx_t_1, "avrational_to_fraction", (void (**)(void))&__pyx_f_2av_5utils_avrational_to_fraction, "PyObject *(AVRational const *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "to_avrational", (void (**)(void))&__pyx_f_2av_5utils_to_avrational, "PyObject *(PyObject *, AVRational *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initcontext(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initcontext(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_context(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_context(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_context(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  struct __pyx_opt_args_2av_4enum_define_enum __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  int __pyx_t_21;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'context' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_context(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("context", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_av__codec__context) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "av.codec.context")) {
      if (unlikely(PyDict_SetItemString(modules, "av.codec.context", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  if (unlikely(__Pyx_modinit_function_export_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code();
  if (unlikely(__Pyx_modinit_function_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "av/codec/context.pyx":1
 * import warnings             # <<<<<<<<<<<<<<
 * 
 * from libc.errno cimport EAGAIN
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_warnings, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_warnings, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "av/codec/context.pyx":16
 * from av.utils cimport avrational_to_fraction, to_avrational
 * 
 * from av.deprecation import AVDeprecationWarning             # <<<<<<<<<<<<<<
 * from av.dictionary import Dictionary
 * 
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_AVDeprecationWarning);
  __Pyx_GIVEREF(__pyx_n_s_AVDeprecationWarning);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_AVDeprecationWarning);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_av_deprecation, __pyx_t_1, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_AVDeprecationWarning); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_AVDeprecationWarning, __pyx_t_1) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "av/codec/context.pyx":17
 * 
 * from av.deprecation import AVDeprecationWarning
 * from av.dictionary import Dictionary             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_Dictionary);
  __Pyx_GIVEREF(__pyx_n_s_Dictionary);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_Dictionary);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_av_dictionary, __pyx_t_2, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_Dictionary); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Dictionary, __pyx_t_2) < 0) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "av/codec/context.pyx":20
 * 
 * 
 * cdef object _cinit_sentinel = object()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_builtin_object); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_XGOTREF(__pyx_v_2av_5codec_7context__cinit_sentinel);
  __Pyx_DECREF_SET(__pyx_v_2av_5codec_7context__cinit_sentinel, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "av/codec/context.pyx":46
 * 
 * 
 * ThreadType = define_enum('ThreadType', __name__, (             # <<<<<<<<<<<<<<
 *     ('NONE', 0),
 *     ('FRAME', lib.FF_THREAD_FRAME,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "av/codec/context.pyx":48
 * ThreadType = define_enum('ThreadType', __name__, (
 *     ('NONE', 0),
 *     ('FRAME', lib.FF_THREAD_FRAME,             # <<<<<<<<<<<<<<
 *         """Decode more than one frame at once"""),
 *     ('SLICE', lib.FF_THREAD_SLICE,
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(FF_THREAD_FRAME); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_FRAME);
  __Pyx_GIVEREF(__pyx_n_s_FRAME);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_FRAME);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __Pyx_INCREF(__pyx_kp_s_Decode_more_than_one_frame_at_on);
  __Pyx_GIVEREF(__pyx_kp_s_Decode_more_than_one_frame_at_on);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_s_Decode_more_than_one_frame_at_on);
  __pyx_t_2 = 0;

  /* "av/codec/context.pyx":50
 *     ('FRAME', lib.FF_THREAD_FRAME,
 *         """Decode more than one frame at once"""),
 *     ('SLICE', lib.FF_THREAD_SLICE,             # <<<<<<<<<<<<<<
 *         """Decode more than one part of a single frame at once"""),
 *     ('AUTO', lib.FF_THREAD_SLICE | lib.FF_THREAD_FRAME,
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(FF_THREAD_SLICE); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_n_s_SLICE);
  __Pyx_GIVEREF(__pyx_n_s_SLICE);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_n_s_SLICE);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
  __Pyx_INCREF(__pyx_kp_s_Decode_more_than_one_part_of_a_s);
  __Pyx_GIVEREF(__pyx_kp_s_Decode_more_than_one_part_of_a_s);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_kp_s_Decode_more_than_one_part_of_a_s);
  __pyx_t_2 = 0;

  /* "av/codec/context.pyx":52
 *     ('SLICE', lib.FF_THREAD_SLICE,
 *         """Decode more than one part of a single frame at once"""),
 *     ('AUTO', lib.FF_THREAD_SLICE | lib.FF_THREAD_FRAME,             # <<<<<<<<<<<<<<
 *         """Decode using both FRAME and SLICE methods."""),
 * ), is_flags=True)
 */
  __pyx_t_2 = __Pyx_PyInt_From_int((FF_THREAD_SLICE | FF_THREAD_FRAME)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_n_s_AUTO);
  __Pyx_GIVEREF(__pyx_n_s_AUTO);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_n_s_AUTO);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
  __Pyx_INCREF(__pyx_kp_s_Decode_using_both_FRAME_and_SLIC);
  __Pyx_GIVEREF(__pyx_kp_s_Decode_using_both_FRAME_and_SLIC);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_kp_s_Decode_using_both_FRAME_and_SLIC);
  __pyx_t_2 = 0;

  /* "av/codec/context.pyx":47
 * 
 * ThreadType = define_enum('ThreadType', __name__, (
 *     ('NONE', 0),             # <<<<<<<<<<<<<<
 *     ('FRAME', lib.FF_THREAD_FRAME,
 *         """Decode more than one frame at once"""),
 */
  __pyx_t_2 = PyTuple_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_5);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;

  /* "av/codec/context.pyx":46
 * 
 * 
 * ThreadType = define_enum('ThreadType', __name__, (             # <<<<<<<<<<<<<<
 *     ('NONE', 0),
 *     ('FRAME', lib.FF_THREAD_FRAME,
 */
  __pyx_t_6.__pyx_n = 1;
  __pyx_t_6.is_flags = 1;
  __pyx_t_5 = __pyx_f_2av_4enum_define_enum(__pyx_n_s_ThreadType, __pyx_t_1, __pyx_t_2, 0, &__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ThreadType, __pyx_t_5) < 0) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "av/codec/context.pyx":56
 * ), is_flags=True)
 * 
 * SkipType = define_enum('SkipType', __name__, (             # <<<<<<<<<<<<<<
 *     ('NONE', lib.AVDISCARD_NONE,
 *         """Discard nothing"""),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "av/codec/context.pyx":57
 * 
 * SkipType = define_enum('SkipType', __name__, (
 *     ('NONE', lib.AVDISCARD_NONE,             # <<<<<<<<<<<<<<
 *         """Discard nothing"""),
 *     ('DEFAULT', lib.AVDISCARD_DEFAULT,
 */
  __pyx_t_2 = __Pyx_PyInt_From_enum__AVDiscard(AVDISCARD_NONE); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_NONE);
  __Pyx_GIVEREF(__pyx_n_s_NONE);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_NONE);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
  __Pyx_INCREF(__pyx_kp_s_Discard_nothing);
  __Pyx_GIVEREF(__pyx_kp_s_Discard_nothing);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_s_Discard_nothing);
  __pyx_t_2 = 0;

  /* "av/codec/context.pyx":59
 *     ('NONE', lib.AVDISCARD_NONE,
 *         """Discard nothing"""),
 *     ('DEFAULT', lib.AVDISCARD_DEFAULT,             # <<<<<<<<<<<<<<
 *         """Discard useless packets like 0 size packets in AVI"""),
 *     ('NONREF', lib.AVDISCARD_NONREF,
 */
  __pyx_t_2 = __Pyx_PyInt_From_enum__AVDiscard(AVDISCARD_DEFAULT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_n_s_DEFAULT);
  __Pyx_GIVEREF(__pyx_n_s_DEFAULT);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_n_s_DEFAULT);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
  __Pyx_INCREF(__pyx_kp_s_Discard_useless_packets_like_0_s);
  __Pyx_GIVEREF(__pyx_kp_s_Discard_useless_packets_like_0_s);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_kp_s_Discard_useless_packets_like_0_s);
  __pyx_t_2 = 0;

  /* "av/codec/context.pyx":61
 *     ('DEFAULT', lib.AVDISCARD_DEFAULT,
 *         """Discard useless packets like 0 size packets in AVI"""),
 *     ('NONREF', lib.AVDISCARD_NONREF,             # <<<<<<<<<<<<<<
 *         """Discard all non reference"""),
 *     ('BIDIR', lib.AVDISCARD_BIDIR,
 */
  __pyx_t_2 = __Pyx_PyInt_From_enum__AVDiscard(AVDISCARD_NONREF); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_NONREF);
  __Pyx_GIVEREF(__pyx_n_s_NONREF);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_NONREF);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __Pyx_INCREF(__pyx_kp_s_Discard_all_non_reference);
  __Pyx_GIVEREF(__pyx_kp_s_Discard_all_non_reference);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_s_Discard_all_non_reference);
  __pyx_t_2 = 0;

  /* "av/codec/context.pyx":63
 *     ('NONREF', lib.AVDISCARD_NONREF,
 *         """Discard all non reference"""),
 *     ('BIDIR', lib.AVDISCARD_BIDIR,             # <<<<<<<<<<<<<<
 *         """Discard all bidirectional frames"""),
 *     ('NONINTRA', lib.AVDISCARD_NONINTRA,
 */
  __pyx_t_2 = __Pyx_PyInt_From_enum__AVDiscard(AVDISCARD_BIDIR); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = PyTuple_New(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_n_s_BIDIR);
  __Pyx_GIVEREF(__pyx_n_s_BIDIR);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_n_s_BIDIR);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_2);
  __Pyx_INCREF(__pyx_kp_s_Discard_all_bidirectional_frames);
  __Pyx_GIVEREF(__pyx_kp_s_Discard_all_bidirectional_frames);
  PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_kp_s_Discard_all_bidirectional_frames);
  __pyx_t_2 = 0;

  /* "av/codec/context.pyx":65
 *     ('BIDIR', lib.AVDISCARD_BIDIR,
 *         """Discard all bidirectional frames"""),
 *     ('NONINTRA', lib.AVDISCARD_NONINTRA,             # <<<<<<<<<<<<<<
 *         """Discard all non intra frames"""),
 *     ('NONKEY', lib.AVDISCARD_NONKEY,
 */
  __pyx_t_2 = __Pyx_PyInt_From_enum__AVDiscard(AVDISCARD_NONINTRA); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_8 = PyTuple_New(3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_INCREF(__pyx_n_s_NONINTRA);
  __Pyx_GIVEREF(__pyx_n_s_NONINTRA);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_n_s_NONINTRA);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_2);
  __Pyx_INCREF(__pyx_kp_s_Discard_all_non_intra_frames);
  __Pyx_GIVEREF(__pyx_kp_s_Discard_all_non_intra_frames);
  PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_kp_s_Discard_all_non_intra_frames);
  __pyx_t_2 = 0;

  /* "av/codec/context.pyx":67
 *     ('NONINTRA', lib.AVDISCARD_NONINTRA,
 *         """Discard all non intra frames"""),
 *     ('NONKEY', lib.AVDISCARD_NONKEY,             # <<<<<<<<<<<<<<
 *         """Discard all frames except keyframes"""),
 *     ('ALL', lib.AVDISCARD_ALL,
 */
  __pyx_t_2 = __Pyx_PyInt_From_enum__AVDiscard(AVDISCARD_NONKEY); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_9 = PyTuple_New(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_INCREF(__pyx_n_s_NONKEY);
  __Pyx_GIVEREF(__pyx_n_s_NONKEY);
  PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_n_s_NONKEY);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_2);
  __Pyx_INCREF(__pyx_kp_s_Discard_all_frames_except_keyfra);
  __Pyx_GIVEREF(__pyx_kp_s_Discard_all_frames_except_keyfra);
  PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_kp_s_Discard_all_frames_except_keyfra);
  __pyx_t_2 = 0;

  /* "av/codec/context.pyx":69
 *     ('NONKEY', lib.AVDISCARD_NONKEY,
 *         """Discard all frames except keyframes"""),
 *     ('ALL', lib.AVDISCARD_ALL,             # <<<<<<<<<<<<<<
 *         """Discard all"""),
 * ))
 */
  __pyx_t_2 = __Pyx_PyInt_From_enum__AVDiscard(AVDISCARD_ALL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_10 = PyTuple_New(3); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_INCREF(__pyx_n_s_ALL);
  __Pyx_GIVEREF(__pyx_n_s_ALL);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_n_s_ALL);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_2);
  __Pyx_INCREF(__pyx_kp_s_Discard_all);
  __Pyx_GIVEREF(__pyx_kp_s_Discard_all);
  PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_kp_s_Discard_all);
  __pyx_t_2 = 0;

  /* "av/codec/context.pyx":57
 * 
 * SkipType = define_enum('SkipType', __name__, (
 *     ('NONE', lib.AVDISCARD_NONE,             # <<<<<<<<<<<<<<
 *         """Discard nothing"""),
 *     ('DEFAULT', lib.AVDISCARD_DEFAULT,
 */
  __pyx_t_2 = PyTuple_New(7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_2, 4, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_2, 5, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_2, 6, __pyx_t_10);
  __pyx_t_1 = 0;
  __pyx_t_4 = 0;
  __pyx_t_3 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;

  /* "av/codec/context.pyx":56
 * ), is_flags=True)
 * 
 * SkipType = define_enum('SkipType', __name__, (             # <<<<<<<<<<<<<<
 *     ('NONE', lib.AVDISCARD_NONE,
 *         """Discard nothing"""),
 */
  __pyx_t_10 = __pyx_f_2av_4enum_define_enum(__pyx_n_s_SkipType, __pyx_t_5, __pyx_t_2, 0, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_SkipType, __pyx_t_10) < 0) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "av/codec/context.pyx":73
 * ))
 * 
 * Flags = define_enum('Flags', __name__, (             # <<<<<<<<<<<<<<
 *     ('NONE', 0),
 *     ('UNALIGNED', lib.AV_CODEC_FLAG_UNALIGNED,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_name); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);

  /* "av/codec/context.pyx":75
 * Flags = define_enum('Flags', __name__, (
 *     ('NONE', 0),
 *     ('UNALIGNED', lib.AV_CODEC_FLAG_UNALIGNED,             # <<<<<<<<<<<<<<
 *         """Allow decoders to produce frames with data planes that are not aligned
 *         to CPU requirements (e.g. due to cropping)."""),
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(AV_CODEC_FLAG_UNALIGNED); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_n_s_UNALIGNED);
  __Pyx_GIVEREF(__pyx_n_s_UNALIGNED);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_n_s_UNALIGNED);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
  __Pyx_INCREF(__pyx_kp_s_Allow_decoders_to_produce_frames);
  __Pyx_GIVEREF(__pyx_kp_s_Allow_decoders_to_produce_frames);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_kp_s_Allow_decoders_to_produce_frames);
  __pyx_t_2 = 0;

  /* "av/codec/context.pyx":78
 *         """Allow decoders to produce frames with data planes that are not aligned
 *         to CPU requirements (e.g. due to cropping)."""),
 *     ('QSCALE', lib.AV_CODEC_FLAG_QSCALE,             # <<<<<<<<<<<<<<
 *         """Use fixed qscale."""),
 *     ('4MV', lib.AV_CODEC_FLAG_4MV,
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(AV_CODEC_FLAG_QSCALE); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_9 = PyTuple_New(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_INCREF(__pyx_n_s_QSCALE);
  __Pyx_GIVEREF(__pyx_n_s_QSCALE);
  PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_n_s_QSCALE);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_2);
  __Pyx_INCREF(__pyx_kp_s_Use_fixed_qscale);
  __Pyx_GIVEREF(__pyx_kp_s_Use_fixed_qscale);
  PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_kp_s_Use_fixed_qscale);
  __pyx_t_2 = 0;

  /* "av/codec/context.pyx":80
 *     ('QSCALE', lib.AV_CODEC_FLAG_QSCALE,
 *         """Use fixed qscale."""),
 *     ('4MV', lib.AV_CODEC_FLAG_4MV,             # <<<<<<<<<<<<<<
 *         """4 MV per MB allowed / advanced prediction for H.263."""),
 *     ('OUTPUT_CORRUPT', lib.AV_CODEC_FLAG_OUTPUT_CORRUPT,
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(AV_CODEC_FLAG_4MV); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_8 = PyTuple_New(3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_INCREF(__pyx_kp_s_4MV);
  __Pyx_GIVEREF(__pyx_kp_s_4MV);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_kp_s_4MV);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_2);
  __Pyx_INCREF(__pyx_kp_s_4_MV_per_MB_allowed_advanced_pre);
  __Pyx_GIVEREF(__pyx_kp_s_4_MV_per_MB_allowed_advanced_pre);
  PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_kp_s_4_MV_per_MB_allowed_advanced_pre);
  __pyx_t_2 = 0;

  /* "av/codec/context.pyx":82
 *     ('4MV', lib.AV_CODEC_FLAG_4MV,
 *         """4 MV per MB allowed / advanced prediction for H.263."""),
 *     ('OUTPUT_CORRUPT', lib.AV_CODEC_FLAG_OUTPUT_CORRUPT,             # <<<<<<<<<<<<<<
 *         """Output even those frames that might be corrupted."""),
 *     ('QPEL', lib.AV_CODEC_FLAG_QPEL,
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(AV_CODEC_FLAG_OUTPUT_CORRUPT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 82, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = PyTuple_New(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 82, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_n_s_OUTPUT_CORRUPT);
  __Pyx_GIVEREF(__pyx_n_s_OUTPUT_CORRUPT);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_n_s_OUTPUT_CORRUPT);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_2);
  __Pyx_INCREF(__pyx_kp_s_Output_even_those_frames_that_mi);
  __Pyx_GIVEREF(__pyx_kp_s_Output_even_those_frames_that_mi);
  PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_kp_s_Output_even_those_frames_that_mi);
  __pyx_t_2 = 0;

  /* "av/codec/context.pyx":84
 *     ('OUTPUT_CORRUPT', lib.AV_CODEC_FLAG_OUTPUT_CORRUPT,
 *         """Output even those frames that might be corrupted."""),
 *     ('QPEL', lib.AV_CODEC_FLAG_QPEL,             # <<<<<<<<<<<<<<
 *         """Use qpel MC."""),
 *     ('DROPCHANGED', 1 << 5,
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(AV_CODEC_FLAG_QPEL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_QPEL);
  __Pyx_GIVEREF(__pyx_n_s_QPEL);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_QPEL);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __Pyx_INCREF(__pyx_kp_s_Use_qpel_MC);
  __Pyx_GIVEREF(__pyx_kp_s_Use_qpel_MC);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_s_Use_qpel_MC);
  __pyx_t_2 = 0;

  /* "av/codec/context.pyx":89
 *         """Don't output frames whose parameters differ from first
 *         decoded frame in stream."""),
 *     ('PASS1', lib.AV_CODEC_FLAG_PASS1,             # <<<<<<<<<<<<<<
 *         """Use internal 2pass ratecontrol in first pass mode."""),
 *     ('PASS2', lib.AV_CODEC_FLAG_PASS2,
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(AV_CODEC_FLAG_PASS1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_n_s_PASS1);
  __Pyx_GIVEREF(__pyx_n_s_PASS1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_n_s_PASS1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
  __Pyx_INCREF(__pyx_kp_s_Use_internal_2pass_ratecontrol_i);
  __Pyx_GIVEREF(__pyx_kp_s_Use_internal_2pass_ratecontrol_i);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_kp_s_Use_internal_2pass_ratecontrol_i);
  __pyx_t_2 = 0;

  /* "av/codec/context.pyx":91
 *     ('PASS1', lib.AV_CODEC_FLAG_PASS1,
 *         """Use internal 2pass ratecontrol in first pass mode."""),
 *     ('PASS2', lib.AV_CODEC_FLAG_PASS2,             # <<<<<<<<<<<<<<
 *         """Use internal 2pass ratecontrol in second pass mode."""),
 *     ('LOOP_FILTER', lib.AV_CODEC_FLAG_LOOP_FILTER,
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(AV_CODEC_FLAG_PASS2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 91, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 91, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_PASS2);
  __Pyx_GIVEREF(__pyx_n_s_PASS2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PASS2);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
  __Pyx_INCREF(__pyx_kp_s_Use_internal_2pass_ratecontrol_i_2);
  __Pyx_GIVEREF(__pyx_kp_s_Use_internal_2pass_ratecontrol_i_2);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_s_Use_internal_2pass_ratecontrol_i_2);
  __pyx_t_2 = 0;

  /* "av/codec/context.pyx":93
 *     ('PASS2', lib.AV_CODEC_FLAG_PASS2,
 *         """Use internal 2pass ratecontrol in second pass mode."""),
 *     ('LOOP_FILTER', lib.AV_CODEC_FLAG_LOOP_FILTER,             # <<<<<<<<<<<<<<
 *         """loop filter."""),
 *     ('GRAY', lib.AV_CODEC_FLAG_GRAY,
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(AV_CODEC_FLAG_LOOP_FILTER); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_11 = PyTuple_New(3); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_INCREF(__pyx_n_s_LOOP_FILTER);
  __Pyx_GIVEREF(__pyx_n_s_LOOP_FILTER);
  PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_n_s_LOOP_FILTER);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_t_2);
  __Pyx_INCREF(__pyx_kp_s_loop_filter);
  __Pyx_GIVEREF(__pyx_kp_s_loop_filter);
  PyTuple_SET_ITEM(__pyx_t_11, 2, __pyx_kp_s_loop_filter);
  __pyx_t_2 = 0;

  /* "av/codec/context.pyx":95
 *     ('LOOP_FILTER', lib.AV_CODEC_FLAG_LOOP_FILTER,
 *         """loop filter."""),
 *     ('GRAY', lib.AV_CODEC_FLAG_GRAY,             # <<<<<<<<<<<<<<
 *         """Only decode/encode grayscale."""),
 *     ('PSNR', lib.AV_CODEC_FLAG_PSNR,
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(AV_CODEC_FLAG_GRAY); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_12 = PyTuple_New(3); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_INCREF(__pyx_n_s_GRAY);
  __Pyx_GIVEREF(__pyx_n_s_GRAY);
  PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_n_s_GRAY);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_t_2);
  __Pyx_INCREF(__pyx_kp_s_Only_decode_encode_grayscale);
  __Pyx_GIVEREF(__pyx_kp_s_Only_decode_encode_grayscale);
  PyTuple_SET_ITEM(__pyx_t_12, 2, __pyx_kp_s_Only_decode_encode_grayscale);
  __pyx_t_2 = 0;

  /* "av/codec/context.pyx":97
 *     ('GRAY', lib.AV_CODEC_FLAG_GRAY,
 *         """Only decode/encode grayscale."""),
 *     ('PSNR', lib.AV_CODEC_FLAG_PSNR,             # <<<<<<<<<<<<<<
 *         """error[?] variables will be set during encoding."""),
 *     ('INTERLACED_DCT', lib.AV_CODEC_FLAG_INTERLACED_DCT,
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(AV_CODEC_FLAG_PSNR); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_13 = PyTuple_New(3); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_INCREF(__pyx_n_s_PSNR);
  __Pyx_GIVEREF(__pyx_n_s_PSNR);
  PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_n_s_PSNR);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_t_2);
  __Pyx_INCREF(__pyx_kp_s_error_variables_will_be_set_duri);
  __Pyx_GIVEREF(__pyx_kp_s_error_variables_will_be_set_duri);
  PyTuple_SET_ITEM(__pyx_t_13, 2, __pyx_kp_s_error_variables_will_be_set_duri);
  __pyx_t_2 = 0;

  /* "av/codec/context.pyx":99
 *     ('PSNR', lib.AV_CODEC_FLAG_PSNR,
 *         """error[?] variables will be set during encoding."""),
 *     ('INTERLACED_DCT', lib.AV_CODEC_FLAG_INTERLACED_DCT,             # <<<<<<<<<<<<<<
 *         """Use interlaced DCT."""),
 *     ('LOW_DELAY', lib.AV_CODEC_FLAG_LOW_DELAY,
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(AV_CODEC_FLAG_INTERLACED_DCT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_14 = PyTuple_New(3); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_INCREF(__pyx_n_s_INTERLACED_DCT);
  __Pyx_GIVEREF(__pyx_n_s_INTERLACED_DCT);
  PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_n_s_INTERLACED_DCT);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_t_2);
  __Pyx_INCREF(__pyx_kp_s_Use_interlaced_DCT);
  __Pyx_GIVEREF(__pyx_kp_s_Use_interlaced_DCT);
  PyTuple_SET_ITEM(__pyx_t_14, 2, __pyx_kp_s_Use_interlaced_DCT);
  __pyx_t_2 = 0;

  /* "av/codec/context.pyx":101
 *     ('INTERLACED_DCT', lib.AV_CODEC_FLAG_INTERLACED_DCT,
 *         """Use interlaced DCT."""),
 *     ('LOW_DELAY', lib.AV_CODEC_FLAG_LOW_DELAY,             # <<<<<<<<<<<<<<
 *         """Force low delay."""),
 *     ('GLOBAL_HEADER', lib.AV_CODEC_FLAG_GLOBAL_HEADER,
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(AV_CODEC_FLAG_LOW_DELAY); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_15 = PyTuple_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_INCREF(__pyx_n_s_LOW_DELAY);
  __Pyx_GIVEREF(__pyx_n_s_LOW_DELAY);
  PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_n_s_LOW_DELAY);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_2);
  __Pyx_INCREF(__pyx_kp_s_Force_low_delay);
  __Pyx_GIVEREF(__pyx_kp_s_Force_low_delay);
  PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_kp_s_Force_low_delay);
  __pyx_t_2 = 0;

  /* "av/codec/context.pyx":103
 *     ('LOW_DELAY', lib.AV_CODEC_FLAG_LOW_DELAY,
 *         """Force low delay."""),
 *     ('GLOBAL_HEADER', lib.AV_CODEC_FLAG_GLOBAL_HEADER,             # <<<<<<<<<<<<<<
 *         """Place global headers in extradata instead of every keyframe."""),
 *     ('BITEXACT', lib.AV_CODEC_FLAG_BITEXACT,
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(AV_CODEC_FLAG_GLOBAL_HEADER); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_16 = PyTuple_New(3); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_INCREF(__pyx_n_s_GLOBAL_HEADER);
  __Pyx_GIVEREF(__pyx_n_s_GLOBAL_HEADER);
  PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_n_s_GLOBAL_HEADER);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_16, 1, __pyx_t_2);
  __Pyx_INCREF(__pyx_kp_s_Place_global_headers_in_extradat);
  __Pyx_GIVEREF(__pyx_kp_s_Place_global_headers_in_extradat);
  PyTuple_SET_ITEM(__pyx_t_16, 2, __pyx_kp_s_Place_global_headers_in_extradat);
  __pyx_t_2 = 0;

  /* "av/codec/context.pyx":105
 *     ('GLOBAL_HEADER', lib.AV_CODEC_FLAG_GLOBAL_HEADER,
 *         """Place global headers in extradata instead of every keyframe."""),
 *     ('BITEXACT', lib.AV_CODEC_FLAG_BITEXACT,             # <<<<<<<<<<<<<<
 *         """Use only bitexact stuff (except (I)DCT)."""),
 *     ('AC_PRED', lib.AV_CODEC_FLAG_AC_PRED,
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(AV_CODEC_FLAG_BITEXACT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_17 = PyTuple_New(3); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_INCREF(__pyx_n_s_BITEXACT);
  __Pyx_GIVEREF(__pyx_n_s_BITEXACT);
  PyTuple_SET_ITEM(__pyx_t_17, 0, __pyx_n_s_BITEXACT);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_17, 1, __pyx_t_2);
  __Pyx_INCREF(__pyx_kp_s_Use_only_bitexact_stuff_except_I);
  __Pyx_GIVEREF(__pyx_kp_s_Use_only_bitexact_stuff_except_I);
  PyTuple_SET_ITEM(__pyx_t_17, 2, __pyx_kp_s_Use_only_bitexact_stuff_except_I);
  __pyx_t_2 = 0;

  /* "av/codec/context.pyx":107
 *     ('BITEXACT', lib.AV_CODEC_FLAG_BITEXACT,
 *         """Use only bitexact stuff (except (I)DCT)."""),
 *     ('AC_PRED', lib.AV_CODEC_FLAG_AC_PRED,             # <<<<<<<<<<<<<<
 *         """H.263 advanced intra coding / MPEG-4 AC prediction"""),
 *     ('INTERLACED_ME', lib.AV_CODEC_FLAG_INTERLACED_ME,
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(AV_CODEC_FLAG_AC_PRED); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_18 = PyTuple_New(3); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_INCREF(__pyx_n_s_AC_PRED);
  __Pyx_GIVEREF(__pyx_n_s_AC_PRED);
  PyTuple_SET_ITEM(__pyx_t_18, 0, __pyx_n_s_AC_PRED);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_18, 1, __pyx_t_2);
  __Pyx_INCREF(__pyx_kp_s_H_263_advanced_intra_coding_MPEG);
  __Pyx_GIVEREF(__pyx_kp_s_H_263_advanced_intra_coding_MPEG);
  PyTuple_SET_ITEM(__pyx_t_18, 2, __pyx_kp_s_H_263_advanced_intra_coding_MPEG);
  __pyx_t_2 = 0;

  /* "av/codec/context.pyx":109
 *     ('AC_PRED', lib.AV_CODEC_FLAG_AC_PRED,
 *         """H.263 advanced intra coding / MPEG-4 AC prediction"""),
 *     ('INTERLACED_ME', lib.AV_CODEC_FLAG_INTERLACED_ME,             # <<<<<<<<<<<<<<
 *         """Interlaced motion estimation"""),
 *     ('CLOSED_GOP', lib.AV_CODEC_FLAG_CLOSED_GOP),
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(AV_CODEC_FLAG_INTERLACED_ME); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_19 = PyTuple_New(3); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_INCREF(__pyx_n_s_INTERLACED_ME);
  __Pyx_GIVEREF(__pyx_n_s_INTERLACED_ME);
  PyTuple_SET_ITEM(__pyx_t_19, 0, __pyx_n_s_INTERLACED_ME);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_19, 1, __pyx_t_2);
  __Pyx_INCREF(__pyx_kp_s_Interlaced_motion_estimation);
  __Pyx_GIVEREF(__pyx_kp_s_Interlaced_motion_estimation);
  PyTuple_SET_ITEM(__pyx_t_19, 2, __pyx_kp_s_Interlaced_motion_estimation);
  __pyx_t_2 = 0;

  /* "av/codec/context.pyx":111
 *     ('INTERLACED_ME', lib.AV_CODEC_FLAG_INTERLACED_ME,
 *         """Interlaced motion estimation"""),
 *     ('CLOSED_GOP', lib.AV_CODEC_FLAG_CLOSED_GOP),             # <<<<<<<<<<<<<<
 * ), is_flags=True)
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(AV_CODEC_FLAG_CLOSED_GOP); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_20 = PyTuple_New(2); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_INCREF(__pyx_n_s_CLOSED_GOP);
  __Pyx_GIVEREF(__pyx_n_s_CLOSED_GOP);
  PyTuple_SET_ITEM(__pyx_t_20, 0, __pyx_n_s_CLOSED_GOP);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_20, 1, __pyx_t_2);
  __pyx_t_2 = 0;

  /* "av/codec/context.pyx":74
 * 
 * Flags = define_enum('Flags', __name__, (
 *     ('NONE', 0),             # <<<<<<<<<<<<<<
 *     ('UNALIGNED', lib.AV_CODEC_FLAG_UNALIGNED,
 *         """Allow decoders to produce frames with data planes that are not aligned
 */
  __pyx_t_2 = PyTuple_New(19); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_2, 4, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_2, 5, __pyx_t_3);
  __Pyx_INCREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);
  PyTuple_SET_ITEM(__pyx_t_2, 6, __pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_2, 7, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 8, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_11);
  PyTuple_SET_ITEM(__pyx_t_2, 9, __pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_12);
  PyTuple_SET_ITEM(__pyx_t_2, 10, __pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_13);
  PyTuple_SET_ITEM(__pyx_t_2, 11, __pyx_t_13);
  __Pyx_GIVEREF(__pyx_t_14);
  PyTuple_SET_ITEM(__pyx_t_2, 12, __pyx_t_14);
  __Pyx_GIVEREF(__pyx_t_15);
  PyTuple_SET_ITEM(__pyx_t_2, 13, __pyx_t_15);
  __Pyx_GIVEREF(__pyx_t_16);
  PyTuple_SET_ITEM(__pyx_t_2, 14, __pyx_t_16);
  __Pyx_GIVEREF(__pyx_t_17);
  PyTuple_SET_ITEM(__pyx_t_2, 15, __pyx_t_17);
  __Pyx_GIVEREF(__pyx_t_18);
  PyTuple_SET_ITEM(__pyx_t_2, 16, __pyx_t_18);
  __Pyx_GIVEREF(__pyx_t_19);
  PyTuple_SET_ITEM(__pyx_t_2, 17, __pyx_t_19);
  __Pyx_GIVEREF(__pyx_t_20);
  PyTuple_SET_ITEM(__pyx_t_2, 18, __pyx_t_20);
  __pyx_t_5 = 0;
  __pyx_t_9 = 0;
  __pyx_t_8 = 0;
  __pyx_t_7 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_1 = 0;
  __pyx_t_11 = 0;
  __pyx_t_12 = 0;
  __pyx_t_13 = 0;
  __pyx_t_14 = 0;
  __pyx_t_15 = 0;
  __pyx_t_16 = 0;
  __pyx_t_17 = 0;
  __pyx_t_18 = 0;
  __pyx_t_19 = 0;
  __pyx_t_20 = 0;

  /* "av/codec/context.pyx":73
 * ))
 * 
 * Flags = define_enum('Flags', __name__, (             # <<<<<<<<<<<<<<
 *     ('NONE', 0),
 *     ('UNALIGNED', lib.AV_CODEC_FLAG_UNALIGNED,
 */
  __pyx_t_6.__pyx_n = 1;
  __pyx_t_6.is_flags = 1;
  __pyx_t_20 = __pyx_f_2av_4enum_define_enum(__pyx_n_s_Flags, __pyx_t_10, __pyx_t_2, 0, &__pyx_t_6); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Flags, __pyx_t_20) < 0) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "av/codec/context.pyx":114
 * ), is_flags=True)
 * 
 * Flags2 = define_enum('Flags2', __name__, (             # <<<<<<<<<<<<<<
 *     ('NONE', 0),
 *     ('FAST', lib.AV_CODEC_FLAG2_FAST,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_20, __pyx_n_s_name); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);

  /* "av/codec/context.pyx":116
 * Flags2 = define_enum('Flags2', __name__, (
 *     ('NONE', 0),
 *     ('FAST', lib.AV_CODEC_FLAG2_FAST,             # <<<<<<<<<<<<<<
 *         """Allow non spec compliant speedup tricks."""),
 *     ('NO_OUTPUT', lib.AV_CODEC_FLAG2_NO_OUTPUT,
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(AV_CODEC_FLAG2_FAST); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_10 = PyTuple_New(3); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_INCREF(__pyx_n_s_FAST);
  __Pyx_GIVEREF(__pyx_n_s_FAST);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_n_s_FAST);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_2);
  __Pyx_INCREF(__pyx_kp_s_Allow_non_spec_compliant_speedup);
  __Pyx_GIVEREF(__pyx_kp_s_Allow_non_spec_compliant_speedup);
  PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_kp_s_Allow_non_spec_compliant_speedup);
  __pyx_t_2 = 0;

  /* "av/codec/context.pyx":118
 *     ('FAST', lib.AV_CODEC_FLAG2_FAST,
 *         """Allow non spec compliant speedup tricks."""),
 *     ('NO_OUTPUT', lib.AV_CODEC_FLAG2_NO_OUTPUT,             # <<<<<<<<<<<<<<
 *         """Skip bitstream encoding."""),
 *     ('LOCAL_HEADER', lib.AV_CODEC_FLAG2_LOCAL_HEADER,
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(AV_CODEC_FLAG2_NO_OUTPUT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_19 = PyTuple_New(3); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_INCREF(__pyx_n_s_NO_OUTPUT);
  __Pyx_GIVEREF(__pyx_n_s_NO_OUTPUT);
  PyTuple_SET_ITEM(__pyx_t_19, 0, __pyx_n_s_NO_OUTPUT);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_19, 1, __pyx_t_2);
  __Pyx_INCREF(__pyx_kp_s_Skip_bitstream_encoding);
  __Pyx_GIVEREF(__pyx_kp_s_Skip_bitstream_encoding);
  PyTuple_SET_ITEM(__pyx_t_19, 2, __pyx_kp_s_Skip_bitstream_encoding);
  __pyx_t_2 = 0;

  /* "av/codec/context.pyx":120
 *     ('NO_OUTPUT', lib.AV_CODEC_FLAG2_NO_OUTPUT,
 *         """Skip bitstream encoding."""),
 *     ('LOCAL_HEADER', lib.AV_CODEC_FLAG2_LOCAL_HEADER,             # <<<<<<<<<<<<<<
 *         """Place global headers at every keyframe instead of in extradata."""),
 *     ('CHUNKS', lib.AV_CODEC_FLAG2_CHUNKS,
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(AV_CODEC_FLAG2_LOCAL_HEADER); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_18 = PyTuple_New(3); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_INCREF(__pyx_n_s_LOCAL_HEADER);
  __Pyx_GIVEREF(__pyx_n_s_LOCAL_HEADER);
  PyTuple_SET_ITEM(__pyx_t_18, 0, __pyx_n_s_LOCAL_HEADER);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_18, 1, __pyx_t_2);
  __Pyx_INCREF(__pyx_kp_s_Place_global_headers_at_every_ke);
  __Pyx_GIVEREF(__pyx_kp_s_Place_global_headers_at_every_ke);
  PyTuple_SET_ITEM(__pyx_t_18, 2, __pyx_kp_s_Place_global_headers_at_every_ke);
  __pyx_t_2 = 0;

  /* "av/codec/context.pyx":122
 *     ('LOCAL_HEADER', lib.AV_CODEC_FLAG2_LOCAL_HEADER,
 *         """Place global headers at every keyframe instead of in extradata."""),
 *     ('CHUNKS', lib.AV_CODEC_FLAG2_CHUNKS,             # <<<<<<<<<<<<<<
 *         """Input bitstream might be truncated at a packet boundaries
 *         instead of only at frame boundaries."""),
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(AV_CODEC_FLAG2_CHUNKS); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_17 = PyTuple_New(3); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_INCREF(__pyx_n_s_CHUNKS);
  __Pyx_GIVEREF(__pyx_n_s_CHUNKS);
  PyTuple_SET_ITEM(__pyx_t_17, 0, __pyx_n_s_CHUNKS);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_17, 1, __pyx_t_2);
  __Pyx_INCREF(__pyx_kp_s_Input_bitstream_might_be_truncat);
  __Pyx_GIVEREF(__pyx_kp_s_Input_bitstream_might_be_truncat);
  PyTuple_SET_ITEM(__pyx_t_17, 2, __pyx_kp_s_Input_bitstream_might_be_truncat);
  __pyx_t_2 = 0;

  /* "av/codec/context.pyx":125
 *         """Input bitstream might be truncated at a packet boundaries
 *         instead of only at frame boundaries."""),
 *     ('IGNORE_CROP', lib.AV_CODEC_FLAG2_IGNORE_CROP,             # <<<<<<<<<<<<<<
 *         """Discard cropping information from SPS."""),
 *     ('SHOW_ALL', lib.AV_CODEC_FLAG2_SHOW_ALL,
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(AV_CODEC_FLAG2_IGNORE_CROP); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_16 = PyTuple_New(3); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_INCREF(__pyx_n_s_IGNORE_CROP);
  __Pyx_GIVEREF(__pyx_n_s_IGNORE_CROP);
  PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_n_s_IGNORE_CROP);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_16, 1, __pyx_t_2);
  __Pyx_INCREF(__pyx_kp_s_Discard_cropping_information_fro);
  __Pyx_GIVEREF(__pyx_kp_s_Discard_cropping_information_fro);
  PyTuple_SET_ITEM(__pyx_t_16, 2, __pyx_kp_s_Discard_cropping_information_fro);
  __pyx_t_2 = 0;

  /* "av/codec/context.pyx":127
 *     ('IGNORE_CROP', lib.AV_CODEC_FLAG2_IGNORE_CROP,
 *         """Discard cropping information from SPS."""),
 *     ('SHOW_ALL', lib.AV_CODEC_FLAG2_SHOW_ALL,             # <<<<<<<<<<<<<<
 *         """Show all frames before the first keyframe"""),
 *     ('EXPORT_MVS', lib.AV_CODEC_FLAG2_EXPORT_MVS,
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(AV_CODEC_FLAG2_SHOW_ALL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_15 = PyTuple_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_INCREF(__pyx_n_s_SHOW_ALL);
  __Pyx_GIVEREF(__pyx_n_s_SHOW_ALL);
  PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_n_s_SHOW_ALL);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_2);
  __Pyx_INCREF(__pyx_kp_s_Show_all_frames_before_the_first);
  __Pyx_GIVEREF(__pyx_kp_s_Show_all_frames_before_the_first);
  PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_kp_s_Show_all_frames_before_the_first);
  __pyx_t_2 = 0;

  /* "av/codec/context.pyx":129
 *     ('SHOW_ALL', lib.AV_CODEC_FLAG2_SHOW_ALL,
 *         """Show all frames before the first keyframe"""),
 *     ('EXPORT_MVS', lib.AV_CODEC_FLAG2_EXPORT_MVS,             # <<<<<<<<<<<<<<
 *         """Export motion vectors through frame side data"""),
 *     ('SKIP_MANUAL', lib.AV_CODEC_FLAG2_SKIP_MANUAL,
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(AV_CODEC_FLAG2_EXPORT_MVS); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_14 = PyTuple_New(3); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_INCREF(__pyx_n_s_EXPORT_MVS);
  __Pyx_GIVEREF(__pyx_n_s_EXPORT_MVS);
  PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_n_s_EXPORT_MVS);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_t_2);
  __Pyx_INCREF(__pyx_kp_s_Export_motion_vectors_through_fr);
  __Pyx_GIVEREF(__pyx_kp_s_Export_motion_vectors_through_fr);
  PyTuple_SET_ITEM(__pyx_t_14, 2, __pyx_kp_s_Export_motion_vectors_through_fr);
  __pyx_t_2 = 0;

  /* "av/codec/context.pyx":131
 *     ('EXPORT_MVS', lib.AV_CODEC_FLAG2_EXPORT_MVS,
 *         """Export motion vectors through frame side data"""),
 *     ('SKIP_MANUAL', lib.AV_CODEC_FLAG2_SKIP_MANUAL,             # <<<<<<<<<<<<<<
 *         """Do not skip samples and export skip information as frame side data"""),
 *     ('RO_FLUSH_NOOP', lib.AV_CODEC_FLAG2_RO_FLUSH_NOOP,
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(AV_CODEC_FLAG2_SKIP_MANUAL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_13 = PyTuple_New(3); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_INCREF(__pyx_n_s_SKIP_MANUAL);
  __Pyx_GIVEREF(__pyx_n_s_SKIP_MANUAL);
  PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_n_s_SKIP_MANUAL);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_t_2);
  __Pyx_INCREF(__pyx_kp_s_Do_not_skip_samples_and_export_s);
  __Pyx_GIVEREF(__pyx_kp_s_Do_not_skip_samples_and_export_s);
  PyTuple_SET_ITEM(__pyx_t_13, 2, __pyx_kp_s_Do_not_skip_samples_and_export_s);
  __pyx_t_2 = 0;

  /* "av/codec/context.pyx":133
 *     ('SKIP_MANUAL', lib.AV_CODEC_FLAG2_SKIP_MANUAL,
 *         """Do not skip samples and export skip information as frame side data"""),
 *     ('RO_FLUSH_NOOP', lib.AV_CODEC_FLAG2_RO_FLUSH_NOOP,             # <<<<<<<<<<<<<<
 *         """Do not reset ASS ReadOrder field on flush (subtitles decoding)"""),
 * ), is_flags=True)
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(AV_CODEC_FLAG2_RO_FLUSH_NOOP); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_12 = PyTuple_New(3); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_INCREF(__pyx_n_s_RO_FLUSH_NOOP);
  __Pyx_GIVEREF(__pyx_n_s_RO_FLUSH_NOOP);
  PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_n_s_RO_FLUSH_NOOP);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_t_2);
  __Pyx_INCREF(__pyx_kp_s_Do_not_reset_ASS_ReadOrder_field);
  __Pyx_GIVEREF(__pyx_kp_s_Do_not_reset_ASS_ReadOrder_field);
  PyTuple_SET_ITEM(__pyx_t_12, 2, __pyx_kp_s_Do_not_reset_ASS_ReadOrder_field);
  __pyx_t_2 = 0;

  /* "av/codec/context.pyx":115
 * 
 * Flags2 = define_enum('Flags2', __name__, (
 *     ('NONE', 0),             # <<<<<<<<<<<<<<
 *     ('FAST', lib.AV_CODEC_FLAG2_FAST,
 *         """Allow non spec compliant speedup tricks."""),
 */
  __pyx_t_2 = PyTuple_New(10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_19);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_19);
  __Pyx_GIVEREF(__pyx_t_18);
  PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_18);
  __Pyx_GIVEREF(__pyx_t_17);
  PyTuple_SET_ITEM(__pyx_t_2, 4, __pyx_t_17);
  __Pyx_GIVEREF(__pyx_t_16);
  PyTuple_SET_ITEM(__pyx_t_2, 5, __pyx_t_16);
  __Pyx_GIVEREF(__pyx_t_15);
  PyTuple_SET_ITEM(__pyx_t_2, 6, __pyx_t_15);
  __Pyx_GIVEREF(__pyx_t_14);
  PyTuple_SET_ITEM(__pyx_t_2, 7, __pyx_t_14);
  __Pyx_GIVEREF(__pyx_t_13);
  PyTuple_SET_ITEM(__pyx_t_2, 8, __pyx_t_13);
  __Pyx_GIVEREF(__pyx_t_12);
  PyTuple_SET_ITEM(__pyx_t_2, 9, __pyx_t_12);
  __pyx_t_10 = 0;
  __pyx_t_19 = 0;
  __pyx_t_18 = 0;
  __pyx_t_17 = 0;
  __pyx_t_16 = 0;
  __pyx_t_15 = 0;
  __pyx_t_14 = 0;
  __pyx_t_13 = 0;
  __pyx_t_12 = 0;

  /* "av/codec/context.pyx":114
 * ), is_flags=True)
 * 
 * Flags2 = define_enum('Flags2', __name__, (             # <<<<<<<<<<<<<<
 *     ('NONE', 0),
 *     ('FAST', lib.AV_CODEC_FLAG2_FAST,
 */
  __pyx_t_6.__pyx_n = 1;
  __pyx_t_6.is_flags = 1;
  __pyx_t_12 = __pyx_f_2av_4enum_define_enum(__pyx_n_s_Flags2, __pyx_t_20, __pyx_t_2, 0, &__pyx_t_6); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Flags2, __pyx_t_12) < 0) __PYX_ERR(0, 114, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

  /* "av/codec/context.pyx":141
 * 
 *     @staticmethod
 *     def create(codec, mode=None):             # <<<<<<<<<<<<<<
 *         cdef Codec cy_codec = codec if isinstance(codec, Codec) else Codec(codec, mode)
 *         cdef lib.AVCodecContext *c_ctx = lib.avcodec_alloc_context3(cy_codec.ptr)
 */
  __pyx_t_12 = PyCFunction_NewEx(&__pyx_mdef_2av_5codec_7context_12CodecContext_1create, NULL, __pyx_n_s_av_codec_context); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext->tp_dict, __pyx_n_s_create, __pyx_t_12) < 0) __PYX_ERR(0, 141, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  PyType_Modified(__pyx_ptype_2av_5codec_7context_CodecContext);

  /* "av/codec/context.pyx":140
 * cdef class CodecContext(object):
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     def create(codec, mode=None):
 *         cdef Codec cy_codec = codec if isinstance(codec, Codec) else Codec(codec, mode)
 */
  __Pyx_GetNameInClass(__pyx_t_12, (PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext, __pyx_n_s_create); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_staticmethod, __pyx_t_12); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext->tp_dict, __pyx_n_s_create, __pyx_t_2) < 0) __PYX_ERR(0, 141, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_2av_5codec_7context_CodecContext);

  /* "av/codec/context.pyx":172
 *         self.ptr.flags = value
 * 
 *     flags = Flags.property(             # <<<<<<<<<<<<<<
 *         _get_flags,
 *         _set_flags,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_Flags); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_20 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_property); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

  /* "av/codec/context.pyx":173
 * 
 *     flags = Flags.property(
 *         _get_flags,             # <<<<<<<<<<<<<<
 *         _set_flags,
 *         """Flag property of :class:`.Flags`."""
 */
  __Pyx_GetNameInClass(__pyx_t_12, (PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext, __pyx_n_s_get_flags); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);

  /* "av/codec/context.pyx":174
 *     flags = Flags.property(
 *         _get_flags,
 *         _set_flags,             # <<<<<<<<<<<<<<
 *         """Flag property of :class:`.Flags`."""
 *     )
 */
  __Pyx_GetNameInClass(__pyx_t_13, (PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext, __pyx_n_s_set_flags); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_14 = NULL;
  __pyx_t_21 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_20))) {
    __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_20);
    if (likely(__pyx_t_14)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_20);
      __Pyx_INCREF(__pyx_t_14);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_20, function);
      __pyx_t_21 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_20)) {
    PyObject *__pyx_temp[4] = {__pyx_t_14, __pyx_t_12, __pyx_t_13, __pyx_kp_s_Flag_property_of_class_Flags};
    __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_20, __pyx_temp+1-__pyx_t_21, 3+__pyx_t_21); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 172, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_20)) {
    PyObject *__pyx_temp[4] = {__pyx_t_14, __pyx_t_12, __pyx_t_13, __pyx_kp_s_Flag_property_of_class_Flags};
    __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_20, __pyx_temp+1-__pyx_t_21, 3+__pyx_t_21); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 172, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  } else
  #endif
  {
    __pyx_t_15 = PyTuple_New(3+__pyx_t_21); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 172, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    if (__pyx_t_14) {
      __Pyx_GIVEREF(__pyx_t_14); PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_14); __pyx_t_14 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_12);
    PyTuple_SET_ITEM(__pyx_t_15, 0+__pyx_t_21, __pyx_t_12);
    __Pyx_GIVEREF(__pyx_t_13);
    PyTuple_SET_ITEM(__pyx_t_15, 1+__pyx_t_21, __pyx_t_13);
    __Pyx_INCREF(__pyx_kp_s_Flag_property_of_class_Flags);
    __Pyx_GIVEREF(__pyx_kp_s_Flag_property_of_class_Flags);
    PyTuple_SET_ITEM(__pyx_t_15, 2+__pyx_t_21, __pyx_kp_s_Flag_property_of_class_Flags);
    __pyx_t_12 = 0;
    __pyx_t_13 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_20, __pyx_t_15, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 172, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  }
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext->tp_dict, __pyx_n_s_flags, __pyx_t_2) < 0) __PYX_ERR(0, 172, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_2av_5codec_7context_CodecContext);

  /* "av/codec/context.pyx":178
 *     )
 * 
 *     unaligned = flags.flag_property('UNALIGNED')             # <<<<<<<<<<<<<<
 *     qscale = flags.flag_property('QSCALE')
 *     four_mv = flags.flag_property('4MV')
 */
  __Pyx_GetNameInClass(__pyx_t_20, (PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext, __pyx_n_s_flags); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_20, __pyx_n_s_flag_property); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  __pyx_t_20 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_15))) {
    __pyx_t_20 = PyMethod_GET_SELF(__pyx_t_15);
    if (likely(__pyx_t_20)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_15);
      __Pyx_INCREF(__pyx_t_20);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_15, function);
    }
  }
  __pyx_t_2 = (__pyx_t_20) ? __Pyx_PyObject_Call2Args(__pyx_t_15, __pyx_t_20, __pyx_n_s_UNALIGNED) : __Pyx_PyObject_CallOneArg(__pyx_t_15, __pyx_n_s_UNALIGNED);
  __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext->tp_dict, __pyx_n_s_unaligned, __pyx_t_2) < 0) __PYX_ERR(0, 178, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_2av_5codec_7context_CodecContext);

  /* "av/codec/context.pyx":179
 * 
 *     unaligned = flags.flag_property('UNALIGNED')
 *     qscale = flags.flag_property('QSCALE')             # <<<<<<<<<<<<<<
 *     four_mv = flags.flag_property('4MV')
 *     output_corrupt = flags.flag_property('OUTPUT_CORRUPT')
 */
  __Pyx_GetNameInClass(__pyx_t_15, (PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext, __pyx_n_s_flags); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_20 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_flag_property); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_15 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_20))) {
    __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_20);
    if (likely(__pyx_t_15)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_20);
      __Pyx_INCREF(__pyx_t_15);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_20, function);
    }
  }
  __pyx_t_2 = (__pyx_t_15) ? __Pyx_PyObject_Call2Args(__pyx_t_20, __pyx_t_15, __pyx_n_s_QSCALE) : __Pyx_PyObject_CallOneArg(__pyx_t_20, __pyx_n_s_QSCALE);
  __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext->tp_dict, __pyx_n_s_qscale, __pyx_t_2) < 0) __PYX_ERR(0, 179, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_2av_5codec_7context_CodecContext);

  /* "av/codec/context.pyx":180
 *     unaligned = flags.flag_property('UNALIGNED')
 *     qscale = flags.flag_property('QSCALE')
 *     four_mv = flags.flag_property('4MV')             # <<<<<<<<<<<<<<
 *     output_corrupt = flags.flag_property('OUTPUT_CORRUPT')
 *     qpel = flags.flag_property('QPEL')
 */
  __Pyx_GetNameInClass(__pyx_t_20, (PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext, __pyx_n_s_flags); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_20, __pyx_n_s_flag_property); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  __pyx_t_20 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_15))) {
    __pyx_t_20 = PyMethod_GET_SELF(__pyx_t_15);
    if (likely(__pyx_t_20)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_15);
      __Pyx_INCREF(__pyx_t_20);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_15, function);
    }
  }
  __pyx_t_2 = (__pyx_t_20) ? __Pyx_PyObject_Call2Args(__pyx_t_15, __pyx_t_20, __pyx_kp_s_4MV) : __Pyx_PyObject_CallOneArg(__pyx_t_15, __pyx_kp_s_4MV);
  __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext->tp_dict, __pyx_n_s_four_mv, __pyx_t_2) < 0) __PYX_ERR(0, 180, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_2av_5codec_7context_CodecContext);

  /* "av/codec/context.pyx":181
 *     qscale = flags.flag_property('QSCALE')
 *     four_mv = flags.flag_property('4MV')
 *     output_corrupt = flags.flag_property('OUTPUT_CORRUPT')             # <<<<<<<<<<<<<<
 *     qpel = flags.flag_property('QPEL')
 *     drop_changed = flags.flag_property('DROPCHANGED')
 */
  __Pyx_GetNameInClass(__pyx_t_15, (PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext, __pyx_n_s_flags); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_20 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_flag_property); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_15 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_20))) {
    __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_20);
    if (likely(__pyx_t_15)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_20);
      __Pyx_INCREF(__pyx_t_15);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_20, function);
    }
  }
  __pyx_t_2 = (__pyx_t_15) ? __Pyx_PyObject_Call2Args(__pyx_t_20, __pyx_t_15, __pyx_n_s_OUTPUT_CORRUPT) : __Pyx_PyObject_CallOneArg(__pyx_t_20, __pyx_n_s_OUTPUT_CORRUPT);
  __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext->tp_dict, __pyx_n_s_output_corrupt, __pyx_t_2) < 0) __PYX_ERR(0, 181, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_2av_5codec_7context_CodecContext);

  /* "av/codec/context.pyx":182
 *     four_mv = flags.flag_property('4MV')
 *     output_corrupt = flags.flag_property('OUTPUT_CORRUPT')
 *     qpel = flags.flag_property('QPEL')             # <<<<<<<<<<<<<<
 *     drop_changed = flags.flag_property('DROPCHANGED')
 *     pass1 = flags.flag_property('PASS1')
 */
  __Pyx_GetNameInClass(__pyx_t_20, (PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext, __pyx_n_s_flags); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_20, __pyx_n_s_flag_property); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  __pyx_t_20 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_15))) {
    __pyx_t_20 = PyMethod_GET_SELF(__pyx_t_15);
    if (likely(__pyx_t_20)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_15);
      __Pyx_INCREF(__pyx_t_20);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_15, function);
    }
  }
  __pyx_t_2 = (__pyx_t_20) ? __Pyx_PyObject_Call2Args(__pyx_t_15, __pyx_t_20, __pyx_n_s_QPEL) : __Pyx_PyObject_CallOneArg(__pyx_t_15, __pyx_n_s_QPEL);
  __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext->tp_dict, __pyx_n_s_qpel, __pyx_t_2) < 0) __PYX_ERR(0, 182, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_2av_5codec_7context_CodecContext);

  /* "av/codec/context.pyx":183
 *     output_corrupt = flags.flag_property('OUTPUT_CORRUPT')
 *     qpel = flags.flag_property('QPEL')
 *     drop_changed = flags.flag_property('DROPCHANGED')             # <<<<<<<<<<<<<<
 *     pass1 = flags.flag_property('PASS1')
 *     pass2 = flags.flag_property('PASS2')
 */
  __Pyx_GetNameInClass(__pyx_t_15, (PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext, __pyx_n_s_flags); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_20 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_flag_property); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_15 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_20))) {
    __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_20);
    if (likely(__pyx_t_15)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_20);
      __Pyx_INCREF(__pyx_t_15);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_20, function);
    }
  }
  __pyx_t_2 = (__pyx_t_15) ? __Pyx_PyObject_Call2Args(__pyx_t_20, __pyx_t_15, __pyx_n_s_DROPCHANGED) : __Pyx_PyObject_CallOneArg(__pyx_t_20, __pyx_n_s_DROPCHANGED);
  __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext->tp_dict, __pyx_n_s_drop_changed, __pyx_t_2) < 0) __PYX_ERR(0, 183, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_2av_5codec_7context_CodecContext);

  /* "av/codec/context.pyx":184
 *     qpel = flags.flag_property('QPEL')
 *     drop_changed = flags.flag_property('DROPCHANGED')
 *     pass1 = flags.flag_property('PASS1')             # <<<<<<<<<<<<<<
 *     pass2 = flags.flag_property('PASS2')
 *     loop_filter = flags.flag_property('LOOP_FILTER')
 */
  __Pyx_GetNameInClass(__pyx_t_20, (PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext, __pyx_n_s_flags); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_20, __pyx_n_s_flag_property); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  __pyx_t_20 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_15))) {
    __pyx_t_20 = PyMethod_GET_SELF(__pyx_t_15);
    if (likely(__pyx_t_20)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_15);
      __Pyx_INCREF(__pyx_t_20);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_15, function);
    }
  }
  __pyx_t_2 = (__pyx_t_20) ? __Pyx_PyObject_Call2Args(__pyx_t_15, __pyx_t_20, __pyx_n_s_PASS1) : __Pyx_PyObject_CallOneArg(__pyx_t_15, __pyx_n_s_PASS1);
  __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext->tp_dict, __pyx_n_s_pass1, __pyx_t_2) < 0) __PYX_ERR(0, 184, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_2av_5codec_7context_CodecContext);

  /* "av/codec/context.pyx":185
 *     drop_changed = flags.flag_property('DROPCHANGED')
 *     pass1 = flags.flag_property('PASS1')
 *     pass2 = flags.flag_property('PASS2')             # <<<<<<<<<<<<<<
 *     loop_filter = flags.flag_property('LOOP_FILTER')
 *     gray = flags.flag_property('GRAY')
 */
  __Pyx_GetNameInClass(__pyx_t_15, (PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext, __pyx_n_s_flags); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_20 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_flag_property); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_15 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_20))) {
    __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_20);
    if (likely(__pyx_t_15)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_20);
      __Pyx_INCREF(__pyx_t_15);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_20, function);
    }
  }
  __pyx_t_2 = (__pyx_t_15) ? __Pyx_PyObject_Call2Args(__pyx_t_20, __pyx_t_15, __pyx_n_s_PASS2) : __Pyx_PyObject_CallOneArg(__pyx_t_20, __pyx_n_s_PASS2);
  __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext->tp_dict, __pyx_n_s_pass2, __pyx_t_2) < 0) __PYX_ERR(0, 185, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_2av_5codec_7context_CodecContext);

  /* "av/codec/context.pyx":186
 *     pass1 = flags.flag_property('PASS1')
 *     pass2 = flags.flag_property('PASS2')
 *     loop_filter = flags.flag_property('LOOP_FILTER')             # <<<<<<<<<<<<<<
 *     gray = flags.flag_property('GRAY')
 *     psnr = flags.flag_property('PSNR')
 */
  __Pyx_GetNameInClass(__pyx_t_20, (PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext, __pyx_n_s_flags); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_20, __pyx_n_s_flag_property); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  __pyx_t_20 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_15))) {
    __pyx_t_20 = PyMethod_GET_SELF(__pyx_t_15);
    if (likely(__pyx_t_20)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_15);
      __Pyx_INCREF(__pyx_t_20);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_15, function);
    }
  }
  __pyx_t_2 = (__pyx_t_20) ? __Pyx_PyObject_Call2Args(__pyx_t_15, __pyx_t_20, __pyx_n_s_LOOP_FILTER) : __Pyx_PyObject_CallOneArg(__pyx_t_15, __pyx_n_s_LOOP_FILTER);
  __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext->tp_dict, __pyx_n_s_loop_filter_2, __pyx_t_2) < 0) __PYX_ERR(0, 186, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_2av_5codec_7context_CodecContext);

  /* "av/codec/context.pyx":187
 *     pass2 = flags.flag_property('PASS2')
 *     loop_filter = flags.flag_property('LOOP_FILTER')
 *     gray = flags.flag_property('GRAY')             # <<<<<<<<<<<<<<
 *     psnr = flags.flag_property('PSNR')
 *     interlaced_dct = flags.flag_property('INTERLACED_DCT')
 */
  __Pyx_GetNameInClass(__pyx_t_15, (PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext, __pyx_n_s_flags); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_20 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_flag_property); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_15 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_20))) {
    __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_20);
    if (likely(__pyx_t_15)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_20);
      __Pyx_INCREF(__pyx_t_15);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_20, function);
    }
  }
  __pyx_t_2 = (__pyx_t_15) ? __Pyx_PyObject_Call2Args(__pyx_t_20, __pyx_t_15, __pyx_n_s_GRAY) : __Pyx_PyObject_CallOneArg(__pyx_t_20, __pyx_n_s_GRAY);
  __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext->tp_dict, __pyx_n_s_gray, __pyx_t_2) < 0) __PYX_ERR(0, 187, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_2av_5codec_7context_CodecContext);

  /* "av/codec/context.pyx":188
 *     loop_filter = flags.flag_property('LOOP_FILTER')
 *     gray = flags.flag_property('GRAY')
 *     psnr = flags.flag_property('PSNR')             # <<<<<<<<<<<<<<
 *     interlaced_dct = flags.flag_property('INTERLACED_DCT')
 *     low_delay = flags.flag_property('LOW_DELAY')
 */
  __Pyx_GetNameInClass(__pyx_t_20, (PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext, __pyx_n_s_flags); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_20, __pyx_n_s_flag_property); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  __pyx_t_20 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_15))) {
    __pyx_t_20 = PyMethod_GET_SELF(__pyx_t_15);
    if (likely(__pyx_t_20)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_15);
      __Pyx_INCREF(__pyx_t_20);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_15, function);
    }
  }
  __pyx_t_2 = (__pyx_t_20) ? __Pyx_PyObject_Call2Args(__pyx_t_15, __pyx_t_20, __pyx_n_s_PSNR) : __Pyx_PyObject_CallOneArg(__pyx_t_15, __pyx_n_s_PSNR);
  __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext->tp_dict, __pyx_n_s_psnr, __pyx_t_2) < 0) __PYX_ERR(0, 188, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_2av_5codec_7context_CodecContext);

  /* "av/codec/context.pyx":189
 *     gray = flags.flag_property('GRAY')
 *     psnr = flags.flag_property('PSNR')
 *     interlaced_dct = flags.flag_property('INTERLACED_DCT')             # <<<<<<<<<<<<<<
 *     low_delay = flags.flag_property('LOW_DELAY')
 *     global_header = flags.flag_property('GLOBAL_HEADER')
 */
  __Pyx_GetNameInClass(__pyx_t_15, (PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext, __pyx_n_s_flags); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_20 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_flag_property); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_15 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_20))) {
    __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_20);
    if (likely(__pyx_t_15)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_20);
      __Pyx_INCREF(__pyx_t_15);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_20, function);
    }
  }
  __pyx_t_2 = (__pyx_t_15) ? __Pyx_PyObject_Call2Args(__pyx_t_20, __pyx_t_15, __pyx_n_s_INTERLACED_DCT) : __Pyx_PyObject_CallOneArg(__pyx_t_20, __pyx_n_s_INTERLACED_DCT);
  __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext->tp_dict, __pyx_n_s_interlaced_dct, __pyx_t_2) < 0) __PYX_ERR(0, 189, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_2av_5codec_7context_CodecContext);

  /* "av/codec/context.pyx":190
 *     psnr = flags.flag_property('PSNR')
 *     interlaced_dct = flags.flag_property('INTERLACED_DCT')
 *     low_delay = flags.flag_property('LOW_DELAY')             # <<<<<<<<<<<<<<
 *     global_header = flags.flag_property('GLOBAL_HEADER')
 *     bitexact = flags.flag_property('BITEXACT')
 */
  __Pyx_GetNameInClass(__pyx_t_20, (PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext, __pyx_n_s_flags); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_20, __pyx_n_s_flag_property); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  __pyx_t_20 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_15))) {
    __pyx_t_20 = PyMethod_GET_SELF(__pyx_t_15);
    if (likely(__pyx_t_20)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_15);
      __Pyx_INCREF(__pyx_t_20);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_15, function);
    }
  }
  __pyx_t_2 = (__pyx_t_20) ? __Pyx_PyObject_Call2Args(__pyx_t_15, __pyx_t_20, __pyx_n_s_LOW_DELAY) : __Pyx_PyObject_CallOneArg(__pyx_t_15, __pyx_n_s_LOW_DELAY);
  __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext->tp_dict, __pyx_n_s_low_delay, __pyx_t_2) < 0) __PYX_ERR(0, 190, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_2av_5codec_7context_CodecContext);

  /* "av/codec/context.pyx":191
 *     interlaced_dct = flags.flag_property('INTERLACED_DCT')
 *     low_delay = flags.flag_property('LOW_DELAY')
 *     global_header = flags.flag_property('GLOBAL_HEADER')             # <<<<<<<<<<<<<<
 *     bitexact = flags.flag_property('BITEXACT')
 *     ac_pred = flags.flag_property('AC_PRED')
 */
  __Pyx_GetNameInClass(__pyx_t_15, (PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext, __pyx_n_s_flags); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_20 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_flag_property); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_15 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_20))) {
    __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_20);
    if (likely(__pyx_t_15)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_20);
      __Pyx_INCREF(__pyx_t_15);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_20, function);
    }
  }
  __pyx_t_2 = (__pyx_t_15) ? __Pyx_PyObject_Call2Args(__pyx_t_20, __pyx_t_15, __pyx_n_s_GLOBAL_HEADER) : __Pyx_PyObject_CallOneArg(__pyx_t_20, __pyx_n_s_GLOBAL_HEADER);
  __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext->tp_dict, __pyx_n_s_global_header, __pyx_t_2) < 0) __PYX_ERR(0, 191, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_2av_5codec_7context_CodecContext);

  /* "av/codec/context.pyx":192
 *     low_delay = flags.flag_property('LOW_DELAY')
 *     global_header = flags.flag_property('GLOBAL_HEADER')
 *     bitexact = flags.flag_property('BITEXACT')             # <<<<<<<<<<<<<<
 *     ac_pred = flags.flag_property('AC_PRED')
 *     interlaced_me = flags.flag_property('INTERLACED_ME')
 */
  __Pyx_GetNameInClass(__pyx_t_20, (PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext, __pyx_n_s_flags); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_20, __pyx_n_s_flag_property); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  __pyx_t_20 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_15))) {
    __pyx_t_20 = PyMethod_GET_SELF(__pyx_t_15);
    if (likely(__pyx_t_20)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_15);
      __Pyx_INCREF(__pyx_t_20);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_15, function);
    }
  }
  __pyx_t_2 = (__pyx_t_20) ? __Pyx_PyObject_Call2Args(__pyx_t_15, __pyx_t_20, __pyx_n_s_BITEXACT) : __Pyx_PyObject_CallOneArg(__pyx_t_15, __pyx_n_s_BITEXACT);
  __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext->tp_dict, __pyx_n_s_bitexact, __pyx_t_2) < 0) __PYX_ERR(0, 192, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_2av_5codec_7context_CodecContext);

  /* "av/codec/context.pyx":193
 *     global_header = flags.flag_property('GLOBAL_HEADER')
 *     bitexact = flags.flag_property('BITEXACT')
 *     ac_pred = flags.flag_property('AC_PRED')             # <<<<<<<<<<<<<<
 *     interlaced_me = flags.flag_property('INTERLACED_ME')
 *     closed_gop = flags.flag_property('CLOSED_GOP')
 */
  __Pyx_GetNameInClass(__pyx_t_15, (PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext, __pyx_n_s_flags); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_20 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_flag_property); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_15 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_20))) {
    __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_20);
    if (likely(__pyx_t_15)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_20);
      __Pyx_INCREF(__pyx_t_15);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_20, function);
    }
  }
  __pyx_t_2 = (__pyx_t_15) ? __Pyx_PyObject_Call2Args(__pyx_t_20, __pyx_t_15, __pyx_n_s_AC_PRED) : __Pyx_PyObject_CallOneArg(__pyx_t_20, __pyx_n_s_AC_PRED);
  __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext->tp_dict, __pyx_n_s_ac_pred, __pyx_t_2) < 0) __PYX_ERR(0, 193, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_2av_5codec_7context_CodecContext);

  /* "av/codec/context.pyx":194
 *     bitexact = flags.flag_property('BITEXACT')
 *     ac_pred = flags.flag_property('AC_PRED')
 *     interlaced_me = flags.flag_property('INTERLACED_ME')             # <<<<<<<<<<<<<<
 *     closed_gop = flags.flag_property('CLOSED_GOP')
 * 
 */
  __Pyx_GetNameInClass(__pyx_t_20, (PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext, __pyx_n_s_flags); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_20, __pyx_n_s_flag_property); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  __pyx_t_20 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_15))) {
    __pyx_t_20 = PyMethod_GET_SELF(__pyx_t_15);
    if (likely(__pyx_t_20)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_15);
      __Pyx_INCREF(__pyx_t_20);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_15, function);
    }
  }
  __pyx_t_2 = (__pyx_t_20) ? __Pyx_PyObject_Call2Args(__pyx_t_15, __pyx_t_20, __pyx_n_s_INTERLACED_ME) : __Pyx_PyObject_CallOneArg(__pyx_t_15, __pyx_n_s_INTERLACED_ME);
  __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext->tp_dict, __pyx_n_s_interlaced_me, __pyx_t_2) < 0) __PYX_ERR(0, 194, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_2av_5codec_7context_CodecContext);

  /* "av/codec/context.pyx":195
 *     ac_pred = flags.flag_property('AC_PRED')
 *     interlaced_me = flags.flag_property('INTERLACED_ME')
 *     closed_gop = flags.flag_property('CLOSED_GOP')             # <<<<<<<<<<<<<<
 * 
 *     def _get_flags2(self):
 */
  __Pyx_GetNameInClass(__pyx_t_15, (PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext, __pyx_n_s_flags); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_20 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_flag_property); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_15 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_20))) {
    __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_20);
    if (likely(__pyx_t_15)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_20);
      __Pyx_INCREF(__pyx_t_15);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_20, function);
    }
  }
  __pyx_t_2 = (__pyx_t_15) ? __Pyx_PyObject_Call2Args(__pyx_t_20, __pyx_t_15, __pyx_n_s_CLOSED_GOP) : __Pyx_PyObject_CallOneArg(__pyx_t_20, __pyx_n_s_CLOSED_GOP);
  __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext->tp_dict, __pyx_n_s_closed_gop, __pyx_t_2) < 0) __PYX_ERR(0, 195, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_2av_5codec_7context_CodecContext);

  /* "av/codec/context.pyx":203
 *         self.ptr.flags2 = value
 * 
 *     flags2 = Flags2.property(             # <<<<<<<<<<<<<<
 *         _get_flags2,
 *         _set_flags2,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_20, __pyx_n_s_Flags2); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_20, __pyx_n_s_property); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "av/codec/context.pyx":204
 * 
 *     flags2 = Flags2.property(
 *         _get_flags2,             # <<<<<<<<<<<<<<
 *         _set_flags2,
 *         """Flag property of :class:`.Flags2`."""
 */
  __Pyx_GetNameInClass(__pyx_t_20, (PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext, __pyx_n_s_get_flags2); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);

  /* "av/codec/context.pyx":205
 *     flags2 = Flags2.property(
 *         _get_flags2,
 *         _set_flags2,             # <<<<<<<<<<<<<<
 *         """Flag property of :class:`.Flags2`."""
 *     )
 */
  __Pyx_GetNameInClass(__pyx_t_13, (PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext, __pyx_n_s_set_flags2); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 205, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_12 = NULL;
  __pyx_t_21 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_15))) {
    __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_15);
    if (likely(__pyx_t_12)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_15);
      __Pyx_INCREF(__pyx_t_12);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_15, function);
      __pyx_t_21 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_15)) {
    PyObject *__pyx_temp[4] = {__pyx_t_12, __pyx_t_20, __pyx_t_13, __pyx_kp_s_Flag_property_of_class_Flags2};
    __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_15, __pyx_temp+1-__pyx_t_21, 3+__pyx_t_21); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 203, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_15)) {
    PyObject *__pyx_temp[4] = {__pyx_t_12, __pyx_t_20, __pyx_t_13, __pyx_kp_s_Flag_property_of_class_Flags2};
    __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_15, __pyx_temp+1-__pyx_t_21, 3+__pyx_t_21); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 203, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  } else
  #endif
  {
    __pyx_t_14 = PyTuple_New(3+__pyx_t_21); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 203, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    if (__pyx_t_12) {
      __Pyx_GIVEREF(__pyx_t_12); PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_12); __pyx_t_12 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_20);
    PyTuple_SET_ITEM(__pyx_t_14, 0+__pyx_t_21, __pyx_t_20);
    __Pyx_GIVEREF(__pyx_t_13);
    PyTuple_SET_ITEM(__pyx_t_14, 1+__pyx_t_21, __pyx_t_13);
    __Pyx_INCREF(__pyx_kp_s_Flag_property_of_class_Flags2);
    __Pyx_GIVEREF(__pyx_kp_s_Flag_property_of_class_Flags2);
    PyTuple_SET_ITEM(__pyx_t_14, 2+__pyx_t_21, __pyx_kp_s_Flag_property_of_class_Flags2);
    __pyx_t_20 = 0;
    __pyx_t_13 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_15, __pyx_t_14, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 203, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  }
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext->tp_dict, __pyx_n_s_flags2, __pyx_t_2) < 0) __PYX_ERR(0, 203, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_2av_5codec_7context_CodecContext);

  /* "av/codec/context.pyx":209
 *     )
 * 
 *     fast = flags2.flag_property('FAST')             # <<<<<<<<<<<<<<
 *     no_output = flags2.flag_property('NO_OUTPUT')
 *     local_header = flags2.flag_property('LOCAL_HEADER')
 */
  __Pyx_GetNameInClass(__pyx_t_15, (PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext, __pyx_n_s_flags2); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_flag_property); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_15 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_14))) {
    __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_14);
    if (likely(__pyx_t_15)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
      __Pyx_INCREF(__pyx_t_15);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_14, function);
    }
  }
  __pyx_t_2 = (__pyx_t_15) ? __Pyx_PyObject_Call2Args(__pyx_t_14, __pyx_t_15, __pyx_n_s_FAST) : __Pyx_PyObject_CallOneArg(__pyx_t_14, __pyx_n_s_FAST);
  __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext->tp_dict, __pyx_n_s_fast, __pyx_t_2) < 0) __PYX_ERR(0, 209, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_2av_5codec_7context_CodecContext);

  /* "av/codec/context.pyx":210
 * 
 *     fast = flags2.flag_property('FAST')
 *     no_output = flags2.flag_property('NO_OUTPUT')             # <<<<<<<<<<<<<<
 *     local_header = flags2.flag_property('LOCAL_HEADER')
 *     chunks = flags2.flag_property('CHUNKS')
 */
  __Pyx_GetNameInClass(__pyx_t_14, (PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext, __pyx_n_s_flags2); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_14, __pyx_n_s_flag_property); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __pyx_t_14 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_15))) {
    __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_15);
    if (likely(__pyx_t_14)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_15);
      __Pyx_INCREF(__pyx_t_14);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_15, function);
    }
  }
  __pyx_t_2 = (__pyx_t_14) ? __Pyx_PyObject_Call2Args(__pyx_t_15, __pyx_t_14, __pyx_n_s_NO_OUTPUT) : __Pyx_PyObject_CallOneArg(__pyx_t_15, __pyx_n_s_NO_OUTPUT);
  __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext->tp_dict, __pyx_n_s_no_output, __pyx_t_2) < 0) __PYX_ERR(0, 210, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_2av_5codec_7context_CodecContext);

  /* "av/codec/context.pyx":211
 *     fast = flags2.flag_property('FAST')
 *     no_output = flags2.flag_property('NO_OUTPUT')
 *     local_header = flags2.flag_property('LOCAL_HEADER')             # <<<<<<<<<<<<<<
 *     chunks = flags2.flag_property('CHUNKS')
 *     ignore_crop = flags2.flag_property('IGNORE_CROP')
 */
  __Pyx_GetNameInClass(__pyx_t_15, (PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext, __pyx_n_s_flags2); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_flag_property); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_15 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_14))) {
    __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_14);
    if (likely(__pyx_t_15)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
      __Pyx_INCREF(__pyx_t_15);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_14, function);
    }
  }
  __pyx_t_2 = (__pyx_t_15) ? __Pyx_PyObject_Call2Args(__pyx_t_14, __pyx_t_15, __pyx_n_s_LOCAL_HEADER) : __Pyx_PyObject_CallOneArg(__pyx_t_14, __pyx_n_s_LOCAL_HEADER);
  __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext->tp_dict, __pyx_n_s_local_header, __pyx_t_2) < 0) __PYX_ERR(0, 211, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_2av_5codec_7context_CodecContext);

  /* "av/codec/context.pyx":212
 *     no_output = flags2.flag_property('NO_OUTPUT')
 *     local_header = flags2.flag_property('LOCAL_HEADER')
 *     chunks = flags2.flag_property('CHUNKS')             # <<<<<<<<<<<<<<
 *     ignore_crop = flags2.flag_property('IGNORE_CROP')
 *     show_all = flags2.flag_property('SHOW_ALL')
 */
  __Pyx_GetNameInClass(__pyx_t_14, (PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext, __pyx_n_s_flags2); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_14, __pyx_n_s_flag_property); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __pyx_t_14 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_15))) {
    __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_15);
    if (likely(__pyx_t_14)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_15);
      __Pyx_INCREF(__pyx_t_14);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_15, function);
    }
  }
  __pyx_t_2 = (__pyx_t_14) ? __Pyx_PyObject_Call2Args(__pyx_t_15, __pyx_t_14, __pyx_n_s_CHUNKS) : __Pyx_PyObject_CallOneArg(__pyx_t_15, __pyx_n_s_CHUNKS);
  __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext->tp_dict, __pyx_n_s_chunks, __pyx_t_2) < 0) __PYX_ERR(0, 212, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_2av_5codec_7context_CodecContext);

  /* "av/codec/context.pyx":213
 *     local_header = flags2.flag_property('LOCAL_HEADER')
 *     chunks = flags2.flag_property('CHUNKS')
 *     ignore_crop = flags2.flag_property('IGNORE_CROP')             # <<<<<<<<<<<<<<
 *     show_all = flags2.flag_property('SHOW_ALL')
 *     export_mvs = flags2.flag_property('EXPORT_MVS')
 */
  __Pyx_GetNameInClass(__pyx_t_15, (PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext, __pyx_n_s_flags2); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_flag_property); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_15 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_14))) {
    __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_14);
    if (likely(__pyx_t_15)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
      __Pyx_INCREF(__pyx_t_15);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_14, function);
    }
  }
  __pyx_t_2 = (__pyx_t_15) ? __Pyx_PyObject_Call2Args(__pyx_t_14, __pyx_t_15, __pyx_n_s_IGNORE_CROP) : __Pyx_PyObject_CallOneArg(__pyx_t_14, __pyx_n_s_IGNORE_CROP);
  __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext->tp_dict, __pyx_n_s_ignore_crop, __pyx_t_2) < 0) __PYX_ERR(0, 213, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_2av_5codec_7context_CodecContext);

  /* "av/codec/context.pyx":214
 *     chunks = flags2.flag_property('CHUNKS')
 *     ignore_crop = flags2.flag_property('IGNORE_CROP')
 *     show_all = flags2.flag_property('SHOW_ALL')             # <<<<<<<<<<<<<<
 *     export_mvs = flags2.flag_property('EXPORT_MVS')
 *     skip_manual = flags2.flag_property('SKIP_MANUAL')
 */
  __Pyx_GetNameInClass(__pyx_t_14, (PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext, __pyx_n_s_flags2); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_14, __pyx_n_s_flag_property); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __pyx_t_14 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_15))) {
    __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_15);
    if (likely(__pyx_t_14)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_15);
      __Pyx_INCREF(__pyx_t_14);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_15, function);
    }
  }
  __pyx_t_2 = (__pyx_t_14) ? __Pyx_PyObject_Call2Args(__pyx_t_15, __pyx_t_14, __pyx_n_s_SHOW_ALL) : __Pyx_PyObject_CallOneArg(__pyx_t_15, __pyx_n_s_SHOW_ALL);
  __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext->tp_dict, __pyx_n_s_show_all, __pyx_t_2) < 0) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_2av_5codec_7context_CodecContext);

  /* "av/codec/context.pyx":215
 *     ignore_crop = flags2.flag_property('IGNORE_CROP')
 *     show_all = flags2.flag_property('SHOW_ALL')
 *     export_mvs = flags2.flag_property('EXPORT_MVS')             # <<<<<<<<<<<<<<
 *     skip_manual = flags2.flag_property('SKIP_MANUAL')
 *     ro_flush_noop = flags2.flag_property('RO_FLUSH_NOOP')
 */
  __Pyx_GetNameInClass(__pyx_t_15, (PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext, __pyx_n_s_flags2); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_flag_property); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_15 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_14))) {
    __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_14);
    if (likely(__pyx_t_15)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
      __Pyx_INCREF(__pyx_t_15);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_14, function);
    }
  }
  __pyx_t_2 = (__pyx_t_15) ? __Pyx_PyObject_Call2Args(__pyx_t_14, __pyx_t_15, __pyx_n_s_EXPORT_MVS) : __Pyx_PyObject_CallOneArg(__pyx_t_14, __pyx_n_s_EXPORT_MVS);
  __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext->tp_dict, __pyx_n_s_export_mvs, __pyx_t_2) < 0) __PYX_ERR(0, 215, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_2av_5codec_7context_CodecContext);

  /* "av/codec/context.pyx":216
 *     show_all = flags2.flag_property('SHOW_ALL')
 *     export_mvs = flags2.flag_property('EXPORT_MVS')
 *     skip_manual = flags2.flag_property('SKIP_MANUAL')             # <<<<<<<<<<<<<<
 *     ro_flush_noop = flags2.flag_property('RO_FLUSH_NOOP')
 * 
 */
  __Pyx_GetNameInClass(__pyx_t_14, (PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext, __pyx_n_s_flags2); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_14, __pyx_n_s_flag_property); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __pyx_t_14 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_15))) {
    __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_15);
    if (likely(__pyx_t_14)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_15);
      __Pyx_INCREF(__pyx_t_14);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_15, function);
    }
  }
  __pyx_t_2 = (__pyx_t_14) ? __Pyx_PyObject_Call2Args(__pyx_t_15, __pyx_t_14, __pyx_n_s_SKIP_MANUAL) : __Pyx_PyObject_CallOneArg(__pyx_t_15, __pyx_n_s_SKIP_MANUAL);
  __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext->tp_dict, __pyx_n_s_skip_manual, __pyx_t_2) < 0) __PYX_ERR(0, 216, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_2av_5codec_7context_CodecContext);

  /* "av/codec/context.pyx":217
 *     export_mvs = flags2.flag_property('EXPORT_MVS')
 *     skip_manual = flags2.flag_property('SKIP_MANUAL')
 *     ro_flush_noop = flags2.flag_property('RO_FLUSH_NOOP')             # <<<<<<<<<<<<<<
 * 
 *     property extradata:
 */
  __Pyx_GetNameInClass(__pyx_t_15, (PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext, __pyx_n_s_flags2); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_flag_property); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_15 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_14))) {
    __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_14);
    if (likely(__pyx_t_15)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
      __Pyx_INCREF(__pyx_t_15);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_14, function);
    }
  }
  __pyx_t_2 = (__pyx_t_15) ? __Pyx_PyObject_Call2Args(__pyx_t_14, __pyx_t_15, __pyx_n_s_RO_FLUSH_NOOP) : __Pyx_PyObject_CallOneArg(__pyx_t_14, __pyx_n_s_RO_FLUSH_NOOP);
  __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_2av_5codec_7context_CodecContext->tp_dict, __pyx_n_s_ro_flush_noop, __pyx_t_2) < 0) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_2av_5codec_7context_CodecContext);

  /* "av/codec/context.pyx":1
 * import warnings             # <<<<<<<<<<<<<<
 * 
 * from libc.errno cimport EAGAIN
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_XDECREF(__pyx_t_20);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init av.codec.context", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init av.codec.context");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (__Pyx_PyFastCFunction_Check(func)) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_FAST_THREAD_STATE
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#else
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#if defined(__Pyx_CyFunction_USED) && defined(NDEBUG)
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr = NULL;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* PyObjectGetAttrStrNoError */
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_n_s_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
    Py_ssize_t itemsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
    PyObject *py_itemsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
    itemsize = ((PyTypeObject *)result)->tp_itemsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
    py_itemsize = PyObject_GetAttrString(result, "__itemsize__");
    if (!py_itemsize)
        goto bad;
    itemsize = PyLong_AsSsize_t(py_itemsize);
    Py_DECREF(py_itemsize);
    py_itemsize = 0;
    if (itemsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (itemsize) {
        if (size % alignment) {
            alignment = size % alignment;
        }
        if (itemsize < (Py_ssize_t)alignment)
            itemsize = (Py_ssize_t)alignment;
    }
    if ((size_t)(basicsize + itemsize) < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* GetVTable */
static void* __Pyx_GetVtable(PyObject *dict) {
    void* ptr;
    PyObject *ob = PyObject_GetItem(dict, __pyx_n_s_pyx_vtable);
    if (!ob)
        goto bad;
#if PY_VERSION_HEX >= 0x02070000
    ptr = PyCapsule_GetPointer(ob, 0);
#else
    ptr = PyCObject_AsVoidPtr(ob);
#endif
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* GetNameInClass */
static PyObject *__Pyx_GetGlobalNameAfterAttributeLookup(PyObject *name) {
    PyObject *result;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    __Pyx_GetModuleGlobalNameUncached(result, name);
    return result;
}
static PyObject *__Pyx__GetNameInClass(PyObject *nmspace, PyObject *name) {
    PyObject *result;
    result = __Pyx_PyObject_GetAttrStr(nmspace, name);
    if (!result) {
        result = __Pyx_GetGlobalNameAfterAttributeLookup(name);
    }
    return result;
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);  // XDECREF since it's only set on Py3 if cline
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__AVDiscard(enum AVDiscard value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum AVDiscard neg_one = (enum AVDiscard) -1, const_zero = (enum AVDiscard) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum AVDiscard) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum AVDiscard) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum AVDiscard) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum AVDiscard) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum AVDiscard) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum AVDiscard),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_int(unsigned int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned int neg_one = (unsigned int) -1, const_zero = (unsigned int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(unsigned int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(unsigned int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned int),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE unsigned int __Pyx_PyInt_As_unsigned_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned int neg_one = (unsigned int) -1, const_zero = (unsigned int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(unsigned int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(unsigned int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (unsigned int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (unsigned int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(unsigned int, digit, digits[0])
                case 2:
                    if (8 * sizeof(unsigned int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) >= 2 * PyLong_SHIFT) {
                            return (unsigned int) (((((unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(unsigned int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) >= 3 * PyLong_SHIFT) {
                            return (unsigned int) (((((((unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(unsigned int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) >= 4 * PyLong_SHIFT) {
                            return (unsigned int) (((((((((unsigned int)digits[3]) << PyLong_SHIFT) | (unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (unsigned int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(unsigned int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(unsigned int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (unsigned int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(unsigned int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(unsigned int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(unsigned int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) - 1 > 2 * PyLong_SHIFT) {
                            return (unsigned int) (((unsigned int)-1)*(((((unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(unsigned int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) - 1 > 2 * PyLong_SHIFT) {
                            return (unsigned int) ((((((unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(unsigned int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) - 1 > 3 * PyLong_SHIFT) {
                            return (unsigned int) (((unsigned int)-1)*(((((((unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(unsigned int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) - 1 > 3 * PyLong_SHIFT) {
                            return (unsigned int) ((((((((unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(unsigned int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) - 1 > 4 * PyLong_SHIFT) {
                            return (unsigned int) (((unsigned int)-1)*(((((((((unsigned int)digits[3]) << PyLong_SHIFT) | (unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(unsigned int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) - 1 > 4 * PyLong_SHIFT) {
                            return (unsigned int) ((((((((((unsigned int)digits[3]) << PyLong_SHIFT) | (unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(unsigned int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(unsigned int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            unsigned int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (unsigned int) -1;
        }
    } else {
        unsigned int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (unsigned int) -1;
        val = __Pyx_PyInt_As_unsigned_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to unsigned int");
    return (unsigned int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to unsigned int");
    return (unsigned int) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int64_t(int64_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int64_t neg_one = (int64_t) -1, const_zero = (int64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int64_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int64_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int64_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int64_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int64_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int64_t),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE enum AVDiscard __Pyx_PyInt_As_enum__AVDiscard(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum AVDiscard neg_one = (enum AVDiscard) -1, const_zero = (enum AVDiscard) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(enum AVDiscard) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(enum AVDiscard, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (enum AVDiscard) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (enum AVDiscard) 0;
                case  1: __PYX_VERIFY_RETURN_INT(enum AVDiscard, digit, digits[0])
                case 2:
                    if (8 * sizeof(enum AVDiscard) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum AVDiscard, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum AVDiscard) >= 2 * PyLong_SHIFT) {
                            return (enum AVDiscard) (((((enum AVDiscard)digits[1]) << PyLong_SHIFT) | (enum AVDiscard)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(enum AVDiscard) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum AVDiscard, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum AVDiscard) >= 3 * PyLong_SHIFT) {
                            return (enum AVDiscard) (((((((enum AVDiscard)digits[2]) << PyLong_SHIFT) | (enum AVDiscard)digits[1]) << PyLong_SHIFT) | (enum AVDiscard)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(enum AVDiscard) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum AVDiscard, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum AVDiscard) >= 4 * PyLong_SHIFT) {
                            return (enum AVDiscard) (((((((((enum AVDiscard)digits[3]) << PyLong_SHIFT) | (enum AVDiscard)digits[2]) << PyLong_SHIFT) | (enum AVDiscard)digits[1]) << PyLong_SHIFT) | (enum AVDiscard)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (enum AVDiscard) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(enum AVDiscard) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum AVDiscard, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(enum AVDiscard) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum AVDiscard, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (enum AVDiscard) 0;
                case -1: __PYX_VERIFY_RETURN_INT(enum AVDiscard, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(enum AVDiscard,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(enum AVDiscard) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum AVDiscard, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum AVDiscard) - 1 > 2 * PyLong_SHIFT) {
                            return (enum AVDiscard) (((enum AVDiscard)-1)*(((((enum AVDiscard)digits[1]) << PyLong_SHIFT) | (enum AVDiscard)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(enum AVDiscard) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum AVDiscard, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum AVDiscard) - 1 > 2 * PyLong_SHIFT) {
                            return (enum AVDiscard) ((((((enum AVDiscard)digits[1]) << PyLong_SHIFT) | (enum AVDiscard)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(enum AVDiscard) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum AVDiscard, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum AVDiscard) - 1 > 3 * PyLong_SHIFT) {
                            return (enum AVDiscard) (((enum AVDiscard)-1)*(((((((enum AVDiscard)digits[2]) << PyLong_SHIFT) | (enum AVDiscard)digits[1]) << PyLong_SHIFT) | (enum AVDiscard)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(enum AVDiscard) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum AVDiscard, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum AVDiscard) - 1 > 3 * PyLong_SHIFT) {
                            return (enum AVDiscard) ((((((((enum AVDiscard)digits[2]) << PyLong_SHIFT) | (enum AVDiscard)digits[1]) << PyLong_SHIFT) | (enum AVDiscard)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(enum AVDiscard) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum AVDiscard, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum AVDiscard) - 1 > 4 * PyLong_SHIFT) {
                            return (enum AVDiscard) (((enum AVDiscard)-1)*(((((((((enum AVDiscard)digits[3]) << PyLong_SHIFT) | (enum AVDiscard)digits[2]) << PyLong_SHIFT) | (enum AVDiscard)digits[1]) << PyLong_SHIFT) | (enum AVDiscard)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(enum AVDiscard) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum AVDiscard, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum AVDiscard) - 1 > 4 * PyLong_SHIFT) {
                            return (enum AVDiscard) ((((((((((enum AVDiscard)digits[3]) << PyLong_SHIFT) | (enum AVDiscard)digits[2]) << PyLong_SHIFT) | (enum AVDiscard)digits[1]) << PyLong_SHIFT) | (enum AVDiscard)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(enum AVDiscard) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum AVDiscard, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(enum AVDiscard) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum AVDiscard, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            enum AVDiscard val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (enum AVDiscard) -1;
        }
    } else {
        enum AVDiscard val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (enum AVDiscard) -1;
        val = __Pyx_PyInt_As_enum__AVDiscard(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to enum AVDiscard");
    return (enum AVDiscard) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to enum AVDiscard");
    return (enum AVDiscard) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[5];
    int same=1, i, found_dot;
    const char* rt_from_call = Py_GetVersion();
    PyOS_snprintf(ctversion, 5, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    found_dot = 0;
    for (i = 0; i < 4; i++) {
        if (!ctversion[i]) {
            same = (rt_from_call[i] < '0' || rt_from_call[i] > '9');
            break;
        }
        if (rt_from_call[i] != ctversion[i]) {
            same = 0;
            break;
        }
    }
    if (!same) {
        char rtversion[5] = {'\0'};
        char message[200];
        for (i=0; i<4; ++i) {
            if (rt_from_call[i] == '.') {
                if (found_dot) break;
                found_dot = 1;
            } else if (rt_from_call[i] < '0' || rt_from_call[i] > '9') {
                break;
            }
            rtversion[i] = rt_from_call[i];
        }
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* FunctionExport */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(__pyx_m, (char *)"__pyx_capi__");
    if (!d) {
        PyErr_Clear();
        d = PyDict_New();
        if (!d)
            goto bad;
        Py_INCREF(d);
        if (PyModule_AddObject(__pyx_m, (char *)"__pyx_capi__", d) < 0)
            goto bad;
    }
    tmp.fp = f;
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(tmp.p, sig, 0);
#else
    cobj = PyCObject_FromVoidPtrAndDesc(tmp.p, (void *)sig, 0);
#endif
    if (!cobj)
        goto bad;
    if (PyDict_SetItemString(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* FunctionImport */
#ifndef __PYX_HAVE_RT_ImportFunction
#define __PYX_HAVE_RT_ImportFunction
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, funcname);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C function %.200s",
                PyModule_GetName(module), funcname);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    tmp.p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, desc);
        goto bad;
    }
    tmp.p = PyCObject_AsVoidPtr(cobj);}
#endif
    *f = tmp.fp;
    if (!(*f))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
